{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit ICQv7;

interface

uses
  sysutils, classes, extctrls, windows, wsocket, flap, contacts, strutils,
  dateUtils;

const
  maxRefs=2000;
  directProtoVersion=7;
  maxPwdLength=8;

type
  TicqError=(
    EC_rateExceeded,
    EC_cantConnect,
    EC_socket,
    EC_other,
    EC_badUIN,           // at login-time, referred to my own uin
    EC_missingLogin,
    EC_anotherLogin,
    EC_serverDisconnected,
    EC_badPwd,
    EC_cantChangePwd,
    EC_loginDelay,
    EC_cantCreateUIN,
    EC_invalidFlap,
    EC_badContact,
    EC_cantConnect_dc,
    EC_proxy_error,
    EC_proxy_badPwd,
    EC_proxy_unk         // unknown reply
  );

  TwpResult=record
    uin:integer;
    nick,first,last,email:string;
    authRequired:boolean;
    status:byte;  // 0=offline 1=online 2=don't know
    end; // TwpResult

  TwpSearch=record
    uin:integer;
    nick,first,last,email,city,state:string;
    gender, lang:byte;
    country:word;
    age:integer;
    onlineOnly:boolean;
    end; // TwpSearch

  TmsgID = int64;

  TicqEvent=(
    IE_error,
    IE_online,
    IE_offline,
    IE_oncoming,
    IE_offgoing,
    IE_userinfo,
    IE_msg,
    IE_email,
    IE_webpager,
    IE_fromMirabilis,
    IE_contacts,
    IE_numOfContactsChanged,
		IE_wpEnd,
    IE_statusChanged,
    IE_authReq,
    IE_auth,
    IE_authDenied,
    IE_url,
    IE_gcard,
    IE_wpResult,
    IE_addedYou,
    IE_visibilityChanged,
    IE_toofast,

    IE_connecting,
    IE_connected,
    IE_loggin,
    IE_redirecting,
    IE_redirected,
    IE_almostOnline,

    IE_serverConnected,
    IE_serverDisconnected,
    IE_serverSent,
    IE_serverGot,
    IE_dcConnected,
    IE_dcDisconnected,
    IE_dcSent,
    IE_dcGot,
    IE_dcError,

    IE_creatingUIN,
    IE_newUin,
    IE_uinDeleted,
    IE_myinfoACK,
    IE_pwdChanged,
    IE_pause,
    IE_ack,
    IE_automsgreq,
    IE_sendingAutomsg,
    IE_endOfOfflineMsgs,
    IE_serverAck,
    IE_msgError,

    IE_fileReq,
    IE_fileOk,
    IE_fileDenied,
    IE_fileack,
    IE_fileabort
  );

  TicqPhase=(
    null_,               // offline
    connecting_,         // trying to reach the login server
    login_,              // performing login on login server
    reconnecting_,       // trying to reach the service server
    relogin_,               // performing login on service server
    settingup_,          // setting up things
    online_,

    creating_uin_      // asking for a new uin
  );

  TicqAccept=( AC_OK, AC_DENIED, AC_AWAY );

  TicqSession=class;

  TicqNotify=procedure (Sender:TicqSession; event:TicqEvent) of object;

  TicqDCmode=( DC_NONE, DC_UPONAUTH, DC_ROASTER, DC_EVERYONE );

  TrefKind=(
    REF_null, 
    REF_wp,
    REF_query,
    REF_simplequery,
    REF_savemyinfo,
    REF_file,
    REF_status,
    REF_msg,
    REF_contacts,
    REF_auth,
    REF_sms
  );

  Tdirect=class;

  Tdirects=class(Tlist)
    icq :TicqSession;
    constructor create(icq_:TicqSession);
    destructor destroy; override;
    function  newFor(c:Tcontact):Tdirect;
    end; // Tdirects

  Tdirect=class
  private
    P_host, P_port:string;
  public
    sock :Twsocket;
    contact :Tcontact;
    imserver :boolean;
    directs :Tdirects;
    kind :(DK_none, DK_file);
    fileName :string;
    fileReceived, fileTotal :integer;
    fileData :string;
    buf :string;
    myspeed :integer;
    hisVer :integer;
    data :pointer;
    constructor create;
    destructor destroy; override;
    procedure connected(Sender: TObject; Error: Word);
    procedure received(Sender: TObject; Error: Word);
    procedure disconnected(Sender: TObject; Error: Word);
    property  host:string read P_host;
    property  port:string read P_port;
    function  myPort:integer;
    function  myinfo:Tcontact;
    procedure connect;
    procedure listen;
    procedure close;
    procedure sendPkt(s:string);

    procedure sendACK1;
    procedure sendACK2;
    procedure sendACK3;
    procedure sendVcard;
    procedure sendSpeed;

    procedure parseVcard(s:string);
    end; // Tdirect

  TicqSession=class
   private
    startingStatus     :Tstatus;
    startingInvisible  :boolean;
    phase              :TicqPhase;
    Q                  :TflapQueue;
    cookie             :string;
    FLAPseq            :word;
    SNACref            :integer;
    waitingNewPwd      :string;
    P_DCmode           :TicqDCmode;
    refs               :array [1..maxRefs] of record
                          kind:TrefKind;
                          uin:integer;
                          end;
    wasUINwp           :boolean;  // trigger a last result at first result
    savingMyInfo       :record
                          running:boolean;
                          ACKcount:integer;
                          c:Tcontact;
                          end;
    creatingUIN        :boolean;  // this is a special session, to create uin
    previousInvisible  :boolean;
    serviceServerAddr  :string;
    serviceServerPort  :string;

    roaster            :TcontactList;
    visibleList        :TcontactList;
    invisibleList      :TcontactList;
    tempVisibleList    :TcontactList;

    P_webaware         :boolean;
    P_authneeded       :boolean;
    P_pwd              :string;
    procedure setWebaware(value:boolean);
    procedure setAuthNeeded(value:boolean);
    procedure setPwd(value:string);
    procedure setDCmode(v:TicqDCmode);
    procedure proxy_connected;

   public
    listener            :TicqNotify;
    sock                :Twsocket;
    server              :Twsocket;
    directs             :Tdirects;
    loginServerAddr     :string;
    loginServerPort     :string;
    myInfo              :Tcontact;
    publicEmail         :boolean;
    birthdayFlag        :boolean;
    http                :record
                           enabled:boolean;
                           addr, port, user, pwd:string;
                           end;
    // used to pass valors to listeners
    eventError          :TicqError;
    eventContact        :Tcontact;
    eventContacts       :TcontactList;
    eventMsg            :string;
    eventWP             :TwpResult;
    eventAddress        :string;
    eventInt            :integer;    // multi-purpose
    eventTime           :TdateTime;  // in local time
    eventName           :string;
    eventFlags          :dword;
    eventOldStatus      :Tstatus;
    eventOldInvisible   :boolean;
    eventData           :string;
    eventUrgent         :boolean;
    eventAccept         :TicqAccept;
    eventFilename       :string;
    eventDirect         :Tdirect;
    eventMsgID          :TmsgID;
    property webaware:boolean  read P_webaware write setWebaware;
    property authNeeded:boolean  read P_authNeeded write setAuthNeeded;
    property pwd:string read P_pwd write setPwd;
    property DCmode:TicqDCmode read P_dcmode write setDCmode;

    constructor create;
    destructor destroy; override;
    procedure connect; overload;
    procedure connect(createUIN:boolean); overload;
    procedure disconnect;
    procedure setStatus(s:Tstatus; inv:boolean);
    function  getStatus:Tstatus;
    function  isOnline:boolean;
    function  isOffline:boolean;
    function  isReady:boolean;     // we can send commands
    function  isConnecting:boolean;
    function  imVisibleTo(c:Tcontact):boolean;
    function  maxCharsFor(c:Tcontact):integer;
    procedure notificationForMsg(msgtype:byte; flags:byte; urgent:boolean; msg:string; offline:boolean);
    function  getLocalIPstr:string;
    function  getLocalIP:integer;
    function  directTo(c:Tcontact):Tdirect;
    function  serverPort:word;
    function  serverStart:word;
    // manage contact lists
    function  readInvisible:TcontactList;
    function  readRoaster:TcontactList;
    function  readVisible:TcontactList;
    function  readTemporaryVisible:TcontactList;
    // manage contacts
    procedure clearTemporaryVisible;
    function  addTemporaryVisible(c:Tcontact):boolean; overload;
    function  addTemporaryVisible(cl:Tcontactlist):boolean; overload;
    function  removeTemporaryVisible(c:Tcontact):boolean; overload;
    function  removeTemporaryVisible(cl:TcontactList):boolean; overload;
    function  add2visible(c:Tcontact):boolean; overload;
    procedure add2visible(cl:TcontactList); overload;
    function  removeFromVisible(c:Tcontact):boolean; overload;
    procedure removeFromVisible(cl:tcontactlist); overload;
    function  add2invisible(c:Tcontact):boolean; overload;
    procedure add2invisible(cl:TcontactList); overload;
    function  removeFromInvisible(c:Tcontact):boolean; overload;
    procedure removeFromInvisible(cl:TcontactList); overload;
    function  addContact(c:Tcontact):boolean; overload;
    procedure addContact(cl:TcontactList); overload;
    function  removeContact(c:Tcontact):boolean;
    procedure setVisibleList(cl:TcontactList);
    procedure setInvisibleList(cl:TcontactList);

    // event managing
    procedure notifyListeners(ev:TicqEvent);
    // send packets
    function  sendFLAP(ch:word; data:string):boolean;
    function  sendSNAC(fam,sub:word; data:string):boolean;
    procedure sendKeepalive;
    function  sendMsg(uin,flags:dword; msg:string; var requiredACK:boolean):integer; // returns handle
    procedure sendSMS(dest, msg:string; ack:boolean);
    function  sendAutoMsgReq(uin:integer):integer;
    procedure sendContacts(uin,flags:dword; cl:TcontactList);
    procedure sendQueryInfo(uin:integer);
    procedure sendSimpleQueryInfo(uin:integer);
    procedure sendAddedYou(uin:integer);
    function  sendFileReq(uin:integer; msg,fn:string; size:integer):integer; // returns handle
    procedure sendFileOk(msgID:TmsgID; c:Tcontact);
    procedure sendFileAck(msgID:TmsgID);
    procedure sendFileAbort(msgID:TmsgID);
    procedure sendPermissions;

    procedure sendAddContact(cl:Tcontactlist); overload;
    procedure sendRemoveContact(cl:Tcontactlist); overload;
    procedure sendAddVisible(cl:Tcontactlist); overload;
    procedure sendRemoveVisible(cl:Tcontactlist); overload;
    procedure sendAddInvisible(cl:Tcontactlist); overload;
    procedure sendRemoveInvisible(cl:Tcontactlist); overload;

    procedure sendAddContact(buinlist:string); overload;
    procedure sendRemoveContact(buinlist:string); overload;
    procedure sendAddVisible(buinlist:string); overload;
    procedure sendRemoveVisible(buinlist:string); overload;
    procedure sendAddInvisible(buinlist:string); overload;
    procedure sendRemoveInvisible(buinlist:string); overload;

    procedure sendACK(status:integer; msg, snac:string);
    procedure sendWPsearch(wp:TwpSearch);
    procedure sendAuthReq(uin:integer; msg:string);
    procedure sendAuth(uin:integer);
    procedure sendAuthDenied(uin:integer; msg:string='');
    procedure sendStatusCode;
    procedure sendCreateUIN;
    procedure sendDeleteUIN;
    procedure sendSaveMyInfoAs(c:Tcontact);
    procedure sendReqOfflineMsgs;
    procedure sendDeleteOfflineMsgs;

    procedure saveMyInfo1(c:Tcontact);
    procedure saveMyInfo2(c:Tcontact);
    procedure saveMyInfo3(c:Tcontact);
    procedure saveMyInfo4(c:Tcontact);
   private
    function  useMsgType2for(c:Tcontact):boolean;
    procedure sendChangePwd(newPwd:string);
    procedure parseGCdata(snac:string; offline:boolean=FALSE);
    procedure parseStatus(snac:string; ofs:integer);
    procedure parseNewUIN(snac:string);
    procedure parseCookie(flap:string);
    procedure parseOncomingUser(snac:string);
    procedure parseOffgoingUser(snac:string);
    procedure parseMsgError(snac:string; ref:integer);
    procedure parseServerAck(snac:string; ref:integer);
    procedure parse1503(snac:string; ref:integer);
    procedure parse040B(snac:string);
    procedure parse010F(snac:string);
    procedure parseIncomingMsg(snac:string);
    procedure goneOffline; // called going offline
    procedure dc_connected(Sender: TObject; Error: Word);
    procedure connected(Sender: TObject; Error: Word);
    procedure received(Sender: TObject; Error: Word);
    procedure disconnected(Sender: TObject; Error: Word);
    procedure parseContactsString(s:string);
    procedure parseAuthString(s:string);
    procedure parsePagerString(s:string);

    procedure sendLogin;
    procedure sendImICQ;
    procedure sendCookie;

    procedure sendIMparameter;
    procedure sendCapabilities;
    procedure sendClientReady;
    procedure sendAckTo107;
    function  addRef(k:TrefKind; uin:integer):integer;
    function  dontBotherStatus:boolean;
    function  myUINle:string;
    function  getFullStatusCode:dword;
    end; // TicqSession

const
  IF_multiple = 1 shl 0;      // multiple recipients
  IF_offline  = 1 shl 1;      // sent while you were offline
  IF_urgent   = 1 shl 2;      // send msg urgent
  IF_noblink  = 1 shl 3;      // send to contact list
  last_IF     = 1 shl 3;      // useful for external additional flags
  Z=#0#0#0#0;
  capability:array [1..7] of string[16]=(
  	#$09#$46#$13#$49#$4C#$7F#$11#$D1#$82#$22#$44#$45#$53#$54#$00#$00,
  	#$09#$46#$13#$44#$4C#$7F#$11#$D1#$82#$22#$44#$45#$53#$54#$00#$00,
		#$97#$B1#$27#$51#$24#$3C#$43#$34#$AD#$22#$D6#$AB#$F7#$3F#$14#$92, // RTF
    #$2E#$7A#$64#$75#$FA#$DF#$4D#$C8#$88#$6F#$EA#$35#$95#$FD#$B6#$DF,
  	#$09#$46#$13#$4E#$4C#$7F#$11#$D1#$82#$22#$44#$45#$53#$54#$00#$00,
  	#$56#$3F#$C8#$09#$0B#$6F#$41#$BD#$9F#$79#$42#$26#$09#$DF#$A2#$F3,
  	#$F2#$E7#$C7#$F4#$FE#$AD#$4D#$FB#$B2#$35#$36#$79#$8B#$DF#$00#$00 // trillian
	);
  header2711=#$1B#0+char(directprotoversion)+Z+Z+Z+Z+#0#0#0#3#0#0#0
            +#4#$FF#$FF#$E#0#$FF#$FF+Z+Z+Z;
  status2code:array [Tstatus] of dword=(0,$11,$13,5,1,$20,0,0);
  overrideStatus2code:array [Tstatus] of dword=(0,9,$A,$E,4,0,0,0);
  publicEmailTab:array [boolean] of byte=(1,0);
  MSG_MSG=1;
  MSG_FILE=3;
  MSG_URL=4;
  MSG_AUTHREQ=6;
  MSG_AUTHDENIED=7;
  MSG_AUTHGIVEN=8;
  MSG_MIRABILIS=9;
  MSG_ADDEDYOU=$C;
  MSG_WEBPAGER=$D;
  MSG_EMAIL=$E;
  MSG_CONTACTS=$13;
  MSG_GCARD=$1A;
  MSG_AUTOMSG_AWAY=$E8;
  MSG_AUTOMSG_OCCUPIED=$E9;
  MSG_AUTOMSG_NA=$EA;
  MSG_AUTOMSG_DND=$EB;
  MSG_AUTOMSG_F4C=$EC;
  MSG_AUTOMSGS =[ MSG_AUTOMSG_AWAY, MSG_AUTOMSG_OCCUPIED, MSG_AUTOMSG_NA,
  								MSG_AUTOMSG_DND, MSG_AUTOMSG_F4C ];

  ACK_OK=0;
  ACK_FILEDENY=1;
  ACK_OCCUPIED=9;
  ACK_DND=$A;
  ACK_AWAY=4;
  ACK_NA=$E;
  ACK_NOBLINK=$C;

  flag_invisible=$00000100;
  flag_webaware =$00010000;
  flag_showip   =$00020000;
  flag_birthday =$00080000;
  flag_unknown  =$00100000;
  // if none of the following direct connection is for everyone
  flag_dcByRequest=$10000000;
  flag_dcForRoaster=$20000000;

  CRLF=#13#10;
  CRLFCRLF=CRLF+CRLF;

var
  GMToffset:TdateTime;  // add it to a GMT time, subtract it from your local time
  contactsDB:TcontactList;

function BUIN(uin:integer):string;

IMPLEMENTATION

type
  TsplitProc=procedure(s:string) of object;

function base64encode(s:string):string;
const
  TABLE='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
type
  Ttriple=array [0..2] of byte;
var
  p:^Ttriple;
  i:integer;
begin
result:='';
p:=@s[1];
for i:=1 to length(s) div 3 do
  begin
  result:=result+TABLE[1+p[0] shr 2]
  	+TABLE[1+(p[0] and 3) shl 4+p[1] shr 4]
    +TABLE[1+(p[1] and 15) shl 2+p[2] shr 6]
    +TABLE[1+(p[2] and 63)];
  inc(p);
  end;
if length(s) mod 3 > 0 then
	result:=result+TABLE[1+p[0] shr 2]+TABLE[1+(p[0] and 3) shl 4+p[1] shr 4];
case length(s) mod 3 of
	1:result:=result+'==';
	2:result:=result+TABLE[1+(p[1] and 15) shl 2+p[2] shr 6]+'=';
  end;
end; // base64encode

procedure splitCL(proc:TsplitProc; cl:TcontactList);
var
  i,cnt:integer;
  s:string;
begin
if cl.count=0 then
  begin
  proc('');
  exit;
  end;
i:=0;
while (i<cl.count) do
  begin
  if i > 0 then
    sleep(1000);
  cnt:=5000;
  s:='';
  while (i<cl.count) and (cnt>0) do
    begin
    s:=s+cl.getAt(i).buin;
    inc(i);
    dec(cnt);
    end;
  proc(s);
  end;
end;

function BUIN(uin:integer):string;
begin
result:=intToStr(uin);
result:=char(length(result))+result;
end;

function chop(ss:string; var s:string):string; overload;
var
  i:integer;
begin
i:=pos(ss,s);
if i=0 then i:=length(s)+1;
result:=copy(s,1,i-1);
delete(s,1,i-1+length(ss));
end; // chop

function code2status(code:dword):Tstatus;
begin
code:=code and ($FFFF-8-flag_invisible);
case code of
  $10: begin result:=SC_OCCUPIED; exit end;
  4: begin result:=SC_NA; exit end;
  2: begin result:=SC_DND; exit end;
  end;
for result:=low(result) to high(result) do
  if status2code[result] = code then
    exit;
result:=SC_ONLINE;
end; // code2status

function sameMethods(a,b:TicqNotify):boolean;
begin result:= double((@a)^) = double((@b)^) end;

function encrypted(s:string):string;
const
  cryptData:array [1..16] of byte=($F3,$26,$81,$C4,$39,$86,$DB,$92,$71,$A3,$B9,$E6,$53,$7A,$95,$7C);
var
  i:integer;
begin
i:=length(s);
setLength(result,i);
while i > 0 do
  begin
  byte(result[i]):=byte(s[i]) xor cryptData[i];
  dec(i);
  end;
end; // encrypted

function str2url(s:string):string;
var
  i:integer;
  ss:string;
begin
result:='';
for i:=1 to length(s) do
  begin
  case s[i] of
    ' ':ss:='%20';
    'A'..'Z','a'..'z','0'..'9':ss:=s[i];
    else ss:='%'+intToHex(ord(s[i]),2);
    end;
  result:=result+ss;
  end;
end; // str2html

function str2html(s:string):string;
var
  i:integer;
  ss:string;
begin
result:='';
for i:=1 to length(s) do
  begin
  case s[i] of
    '':ss:='&egrave;';
    '':ss:='&egrave;';
    '':ss:='&eacute;';
    '':ss:='&igrave;';
    '':ss:='&ograve;';
    '':ss:='&ugrave;';
    '':ss:='&Eacute;';
    '"':ss:='&quot;';
    '&':ss:='&amp;';
    else ss:=s[i];
    end;
  result:=result+ss;
  end;
end; // str2html

function xml_sms(me:Tcontact; dest,msg:string; ack:boolean):string;
const
  yesno:array [boolean] of string=('No','Yes');
begin
result:=
 '<icq_sms_message>'+
 '<destination>'+dest+'</destination>'+
 '<text>'+str2html(msg)+'</text>'+
 '<codepage>1252</codepage>'+
 '<senders_UIN>'+intTostr(me.uin)+'</senders_UIN>'+
 '<senders_name>'+me.displayed+'</senders_name>'+
 '<delivery_receipt>'+yesno[ack]+'</delivery_receipt>'+
 '<time>'+formatDatetime('ddd, dd mmm yyyy hh:nn:ss GMT',now-gmtoffset)+'</time>'+
 '</icq_sms_message>';
end; // xml_sms

/////////////////////////////////////////////////////////

constructor Tdirect.create;
begin
sock:=Twsocket.create(NIL);
sock.tag:=integer(@self);
sock.OnDataAvailable:=received;
sock.OnSessionClosed:=disconnected;
imserver:=TRUE;
kind:=DK_none;
myspeed:=100;
end; // create

destructor Tdirect.destroy;
begin
sock.free;
directs.remove(self);
end; // destroy

procedure Tdirect.listen;
begin
sock.OnSessionAvailable:=connected;
imserver:=TRUE;
sock.addr:='0.0.0.0';
sock.port:='0';
sock.listen;
end; // listen

procedure Tdirect.sendPkt(s:string);
begin
s:=word_LEasStr(length(s))+s;
sock.sendStr(s);
with directs.icq do
  begin
  eventData:=s;
  eventDirect:=self;
  notifyListeners(IE_dcSent);
  end;
end; // sendPkt

procedure Tdirect.connect;
begin
sock.OnSessionConnected:=connected;
imserver:=FALSE;
sock.addr:=dword_LE2ip(contact.ip);
sock.port:=intToStr(contact.ft_port);
sock.connect;
end; // connect

procedure Tdirect.close;
begin sock.close end;

procedure Tdirect.connected(Sender: TObject; Error: Word);
var
  icq:TicqSession;
begin
icq:=directs.icq;
icq.eventDirect:=self;
if error<>0 then
  begin
  if WSocket_WSAGetLastError <> 0 then
    error:=WSocket_WSAGetLastError;
  icq.eventInt:=error;
  icq.eventError:=EC_cantconnect_dc;
  icq.notifyListeners(IE_error);
  exit;
  end;
if imserver then
  sock.dup(sock.accept);
P_host:=sock.GetPeerAddr;
P_port:=sock.GetPeerPort;
icq.notifyListeners(IE_dcConnected);
if not imserver and (kind = DK_file) then
  sendVcard;
end; // connected

procedure Tdirect.disconnected(Sender: TObject; Error: Word);
begin
with directs.icq do
  begin
  eventDirect:=self;
  notifyListeners(IE_dcDisconnected);
  end;
end; // disconnected

procedure Tdirect.received(Sender: TObject; Error: Word);
const
  Z=#0#0#0#0;
var
  s:string;
  l:integer;
begin
// queue in buf
buf:=buf+sock.receiveStr;
// extract the packet from buf
if length(buf) < 2 then exit;
l:=word_LEat(@buf[1]);
if length(buf) < l+2 then exit;
s:=copy(buf,1,l+2);
delete(buf,1,l+2);
// log
with directs.icq do
  begin
  eventData:=s;
  eventDirect:=self;
  notifyListeners(IE_dcGot);
  end;
delete(s,1,2);
// reply
if imserver then
  case s[1] of
    #$FF:
      begin
      parseVcard(s);
      sendACK1;
      sendVcard;
      end;
    #3: sendPkt(#3#10#0#0#0#1#0#0#0#1#0#0#0+Z+Z+#1#0#4#0+Z+Z);
    #0:
      begin
      sendSpeed;
      sendACK2;
      end;
    #2:
      begin
      sendPkt(
        #2+Z+#$DA#7#14#0+dword_LEasStr(contact.dc_cookie)+
        Z+Z+Z+#1#0#0+#0#0+WNTS('')+Z+Z
        );
      end;
    #6:
      begin
      fileData:=fileData+copy(s,2,length(s));
      if fileReceived = fileTotal then
        close;
      end;
    end;
if not imserver then
  case s[1] of
    #$FF:sendACK1;
    end;
s:='';
end; // received

procedure Tdirect.parseVcard(s:string);
begin
hisVer:=ord(s[2]);
if contact=NIL then
  begin
  contact:=contactsDB.get(dword_LEat(@s[16]));
  contact.port:=word_LEat(@s[29]);
  end;
end; // parseVcard

function Tdirect.myPort:integer;
begin result:=strToInt(sock.getxPort) end;

function Tdirect.myinfo:Tcontact;
begin result:=directs.icq.myinfo end;

procedure Tdirect.sendACK1;
begin sendPkt(#1#0#0#0) end;

procedure Tdirect.sendACK2;
begin sendPkt(#1+dword_LEasStr(myspeed)+WNTS(myinfo.displayed)) end;

procedure Tdirect.sendACK3;
begin sendPkt(#3+Z+Z+dword_LEasStr(myspeed)+#1#0#0) end;

procedure Tdirect.sendVcard;
begin
sendPkt(
  #$FF+char(6)+#0#$2B#0+dword_LEasStr(contact.uin)+#0#0+
  dword_LEasStr(directs.icq.serverPort)+directs.icq.myUINle+
  dword_LEasStr(directs.icq.getLocalIP)+#127#0#0#1+#4+
  dword_LEasStr(directs.icq.serverPort)+dword_LEasStr(contact.dc_cookie)+
  #$50#0#0#0#3#0#0#0#0#0#0#0
);
end;

procedure Tdirect.sendSpeed;
begin sendPkt(#5+dword_LEasStr(myspeed)) end;

///////////////////////////////////////////////////////////////////////

constructor Tdirects.create(icq_:TicqSession);
begin
icq:=icq_;
end; // create

destructor Tdirects.destroy;
var
  i:Integer;
begin
for i:=count-1 downto 0 do
  Tdirect(items[i]).free;
inherited;
end; // destroy

function Tdirects.newFor(c:Tcontact):Tdirect;
begin
result:=Tdirect.create;
result.contact:=c;
result.directs:=self;
add(result);
end; // newFor

///////////////////////////////////////////////////////////////////////

constructor TicqSession.create;
begin
inherited create;
phase:=null_;

listener:=NIL;

myinfo:=NIL;
P_pwd:='';
SNACref:=1;
FLAPseq:=$6700+random($100);
startingStatus:=SC_ONLINE;
directs:=Tdirects.create(self);
DCmode:=DC_none;

http.enabled:=FALSE;

loginServerAddr:='login.icq.com';
loginServerPort:='5190';
Q:=TflapQueue.create;
roaster:=TcontactList.create;
visibleList:=TcontactList.create;
invisibleList:=TcontactList.create;
tempVisibleList:=TcontactList.create;

savingmyinfo.running:=FALSE;

sock:=Twsocket.create(NIL);
sock.OnSessionConnected:=connected;
sock.OnDataAvailable:=received;
sock.OnSessionClosed:=disconnected;

server:=Twsocket.create(NIL);
server.OnSessionAvailable:=dc_connected;
end; // create

destructor TicqSession.destroy;
begin
directs.free;
Q.free;
sock.free;
roaster.free;
visibleList.free;
invisibleList.free;
tempvisibleList.free;
inherited destroy;
end; // destroy

function TicqSession.myUINle:string;
begin result:=dword_LEasStr(myinfo.uin) end;

procedure TicqSession.setPwd(value:string);
begin
if (value<>pwd) and (length(value) <= maxPwdLength) then
  if isOnline then
    sendChangePwd(value)
  else
    P_pwd:=value;
end; // setPwd

function TicqSession.sendFLAP(ch:word; data:string):boolean;
var
  s:string;
begin
result:=FALSE;
if sock.State <> wsConnected then exit;
s:='*'
  +char(ch)
  +word_BEasStr(FLAPseq)
  +word_BEasStr(length(data))
  +data;
 try
  sock.sendStr(s);
  eventData:=s;
  notifyListeners(IE_serverGot);
  inc(FLAPseq);
  if FLAPseq = $8000 then FLAPseq:=0;
 except
 end;
s:='';
result:=TRUE;
end; // sendFLAP

function TicqSession.sendSNAC(fam,sub:word; data:string):boolean;
begin result:=sendFLAP(SNAC_CHANNEL, SNAC(fam,sub, SNACref)+data) end;

procedure TicqSession.sendKeepalive;
begin sendFLAP(KEEPALIVE_CHANNEL,'') end; 

procedure TicqSession.notifyListeners(ev:TicqEvent);
begin
if assigned(listener) then
  listener(self,ev);
end; // notifyListeners

function TicqSession.isOffline:boolean;
begin result:= phase=null_ end;

function TicqSession.isOnline:boolean;
begin result:= phase=online_ end;

function TicqSession.isConnecting:boolean;
begin result:=not (isOffline or isOnline) end;

procedure TicqSession.dc_connected(Sender: TObject; Error: Word);
begin
if error<>0 then
  begin
  if WSocket_WSAGetLastError <> 0 then
    error:=WSocket_WSAGetLastError;
  eventInt:=error;
  eventError:=EC_cantconnect_dc;
  notifyListeners(IE_error);
  exit;
  end;
eventDirect:=directs.newFor(NIL);
eventDirect.sock.dup(server.accept);
notifyListeners(IE_dcConnected);
end; // dc_connected

procedure TicqSession.goneOffline;
var
  i:integer;
begin
if phase=null_ then exit;
phase:=null_;
tempvisibleList.clear;
if DCmode <> DC_none then
  server.close;
if assigned(myinfo) then
  begin
  myinfo.status:=SC_OFFLINE;
  myinfo.invisible:=FALSE;
  end;
with roaster do
  for i:=0 to count-1 do
    with getAt(i) do
      begin
      status:=SC_UNK;
      invisible:=FALSE;
      end;
notifyListeners(IE_offline);
end; // goneOffline

procedure TicqSession.disconnect;
begin
q.reset;
sock.close;
goneOffline;
end;

procedure TicqSession.connected(Sender: TObject; Error: Word);
var
  s:string;
begin
if error <> 0 then
  begin
  goneOffline;
  eventInt:=WSocket_WSAGetLastError;
  if eventInt=0 then eventInt:=error;
  eventError:=EC_cantconnect;
  notifyListeners(IE_error);
  exit;
  end;
eventAddress:=sock.Addr;
notifyListeners(IE_serverConnected);
if http.enabled then
  begin
  if phase = CONNECTING_ then eventData:=loginServerAddr+':'+loginServerPort
  else eventData:=serviceServerAddr+':'+serviceServerPort;
  if http.user > '' then
    begin
    s:=base64encode(http.user+':'+http.pwd);
    s:=
      'Authorization: Basic '+s+CRLF+
      'Proxy-authorization: Basic '+s+CRLF;
    end;
  eventData:=
    'CONNECT '+eventData+' HTTP/1.0'+CRLF+
    'User-agent: ICQ/2000b (Mozilla 1.24b; Windows; I; 32-bit)'+CRLF+
    s+    // eventually empty
    CRLF;
  sock.sendStr(eventData);
  notifyListeners(IE_serverGot);
  end
else
  proxy_connected;
end; // connected

procedure TicqSession.proxy_connected;
begin
if creatingUIN then
  begin
  phase:=creating_uin_;
  notifyListeners(IE_connected);
  end
else
  case phase of
    connecting_:
      begin
      phase:=login_;
      notifyListeners(IE_connected);
      end;
    reconnecting_:
      begin
      phase:=relogin_;
      notifyListeners(IE_redirected);
      end;
    end
end; // proxy_connected

procedure TicqSession.disconnected(Sender: TObject; Error: Word);
begin
q.reset;
eventAddress:=sock.addr;
notifyListeners(IE_serverDisconnected);
if error <> 0 then
  begin
  goneOffline;
  eventInt:=WSocket_WSAGetLastError;
  if eventInt=0 then eventInt:=error;
  eventError:=EC_socket;
  notifyListeners(IE_error);
  exit;
  end;
if phase<>login_ then
  goneOffline;
end; // disconnected

function TicqSession.isReady:boolean;
begin result:=phase in [SETTINGUP_,ONLINE_] end;

procedure TicqSession.sendStatusCode;
const
  cookie=#1#2#3#4;
var
  dc:string;
begin
if not isReady then exit;

if DCmode = DC_none then
  dc:=Z+dword_BEasStr(0)
else
  dc:=#127#0#0#1+dword_BEasStr(serverPort);
dc:=dc+#4#0+char(directProtoversion)+cookie+#0#0#0#$50#0#0#0#3+dword_BEasStr(myinfo.lastUpdate_dw)+dword_BEasStr(myinfo.lastInfoUpdate_dw)+Z+#0#0;

if previousInvisible<>myinfo.invisible then
  sendAddVisible(visibleList);

sendSNAC(1,$1E, TLV(6, getFullStatusCode)
  //+TLV(8, #0#0)
  +TLV($C, dc)
  +TLV($11, #1#$2C#$35#$FB#$3B)
  +TLV($12, #0#0)
);

{pkt.createSNAC(1,$11,0);
pkt.addDword_BE(0);
pkt.send(sock);
addRef(REF_null,0);}

if previousInvisible<>myinfo.invisible then
  begin
  if not myinfo.invisible then
    sendAddInvisible(invisibleList);

  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
previousInvisible:=myinfo.invisible;
end; // sendStatusCode

procedure TicqSession.sendAddVisible(buinlist:string);
begin
if not isReady or not myinfo.invisible then exit;
sendSNAC(9,5, buinlist);
end; // sendAddVisible

procedure TicqSession.sendRemoveVisible(buinlist:string);
begin
if not isReady or not myinfo.invisible then exit;
sendSNAC(9,6, buinlist);
end; // sendRemoveVisible

procedure TicqSession.sendAddInvisible(buinlist:string);
begin
if not isReady or myinfo.invisible then exit;
sendSNAC(9,7, buinlist);
end; // sendAddInvisible

procedure TicqSession.sendRemoveInvisible(buinlist:string);
begin
if not isReady or myinfo.invisible then exit;
sendSNAC(9,8, buinlist);
end; // sendRemoveInvisible

procedure TicqSession.sendAddContact(cl:Tcontactlist);
begin splitCL(sendAddContact,cl) end;

procedure TicqSession.sendRemoveContact(cl:Tcontactlist);
begin splitCL(sendRemoveContact,cl) end;

procedure TicqSession.sendAddInvisible(cl:Tcontactlist);
begin splitCL(sendAddInvisible,cl) end;

procedure TicqSession.sendRemoveInvisible(cl:Tcontactlist);
begin splitCL(sendRemoveInvisible,cl) end;

procedure TicqSession.sendAddVisible(cl:Tcontactlist);
begin splitCL(sendAddVisible,cl) end;

procedure TicqSession.sendRemoveVisible(cl:Tcontactlist);
begin splitCL(sendRemoveVisible,cl) end;

procedure TicqSession.sendAddContact(buinlist:string);
begin
if (buinlist='') or not isReady then exit;
sendSNAC(3,4, buinlist);
end; // sendAddContact

procedure TicqSession.sendRemoveContact(buinlist:string);
begin
if (buinlist='') or not isReady then exit;
sendSNAC(3,5, buinlist);
end; // sendRemoveContact

function TicqSession.sendFileReq(uin:integer; msg,fn:string; size:integer):integer;
var
  c:Tcontact;
begin
result:=-1;
if not isReady then exit;

c:=contactsDB.get(uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

sendSNAC(4,6, qword_LEasStr(SNACref)+#0#2
  +BUIN(uin)
  +TLV(5, #0#0
    +qword_LEasStr(SNACref)+capability[1]+TLV($A,#0#1)
    +TLV(5,word_BEasStr(serverPort))+TLV(3,#10#0#0#1)+TLV($F,'')
    +TLV($2711, header2711+char(MSG_FILE)+#0
      +word_LEasStr(word(status2code[myinfo.status]))+#1#0+WNTS(msg)
      +Z+WNTS(fn)+dword_LEasStr(size)+Z )
  )
  +TLV(3,'')
);
result:=addRef(REF_file,uin);
end; // sendFileReq

procedure TicqSession.sendFileOk(msgID:TmsgID; c:Tcontact);
begin
if not isReady then exit;

if not imVisibleTo(c) then
  addTemporaryVisible(c);

eventDirect:=directTo(c);
//eventDirect.listen;
sendSNAC(4,6, qword_LEasStr(msgID)+#0#2
  +BUIN(c.uin)
  +TLV(5, #0#0+qword_LEasStr(msgID)+capability[1]+TLV($A,#0#2)
    +TLV(5,word_BEasStr(serverPort))+TLV(3,#127#0#0#1)+TLV($F,'')
    +TLV($2711, header2711+char(MSG_FILE)+#0
      +word_LEasStr(word(status2code[myinfo.status]))+#0#0+WNTS('')
      +Z+WNTS('')+dword_LEasStr(0)+Z
      +word_BEasStr(eventDirect.myPort)+#0#0+WNTS('')+dword_LEasStr(0)
      +word_LEasStr(eventDirect.myPort)+#0#0 )
  )
  +TLV(3,'')
);
end; // sendFileOK

procedure TicqSession.sendFileAbort(msgID:TmsgID);
var
  c:Tcontact;
begin
if not isReady then exit;

c:=contactsDB.get(refs[msgID].uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

sendSNAC(4,6, qword_LEasStr(msgID)+#0#2
  +BUIN(refs[msgID].uin)
  +TLV(5, #0#0+qword_LEasStr(msgID)+capability[1]+TLV($B,#0#1) )
);
end; // sendFileAbort

procedure TicqSession.sendFileAck(msgID:TmsgID);
var
  c:Tcontact;
begin
if not isReady then exit;
c:=contactsDB.get(refs[msgID].uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

sendSNAC(4,6, qword_LEasStr(msgID)+#0#2
  +BUIN(refs[msgID].uin)
  +TLV(5, #0#2+qword_LEasStr(msgID)+capability[1] )
);
end; // sendFileAck

procedure TicqSession.sendAuthReq(uin:integer; msg:string);
var
  c:Tcontact;
begin
c:=contactsDB.get(uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

sendSNAC(4,6, qword_LEasStr(SNACref)+#0#4
  +BUIN(uin)
  +TLV(5, myUINle+char(MSG_AUTHREQ)+#0
    +WNTS(myinfo.nick+#$FE+myinfo.first+#$FE+myinfo.last+#$FE+myinfo.email+#$FE#0#$FE+msg)
  )
);
end; // sendAuthReq

function TicqSession.sendMsg(uin,flags:dword; msg:string; var requiredACK:boolean):integer;
var
  flagChar,priorityChar:char;
  c:Tcontact;
begin
result:=-1;
if not isReady then exit;

c:=contactsDB.get(uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

flagChar:=#0;
if IF_multiple and flags>0 then flagChar:=#$80;
priorityChar:=#1;
if IF_urgent and flags>0 then priorityChar:=#2;
if IF_noblink and flags>0 then priorityChar:=#4;

if useMsgType2For(c) then
  begin
  requiredACK:=TRUE;

  sendSNAC(4,6, qword_LEasStr(SNACref)+#0#2
    +BUIN(uin)
    +TLV(5, #0#0+qword_LEasStr(SNACref)+capability[1]
      +TLV($A,#0#1)
      +TLV($F,'')
      +TLV($2711,
        header2711+char(MSG_MSG)+flagChar+
        word_LEasStr(getFullStatusCode)
        +priorityChar+#0
        +WNTS(msg)
        +dword_LEasStr(0)+dword_LEasStr($FFFFFF)
      )
    )
    +TLV(3,'')
  );
  end
else
  begin
  requiredACK:=FALSE;
  sendSNAC(4,6, qword_LEasStr(SNACref)+#0#1
    +BUIN(uin)
    +TLV(2, TLV($0501, #1)+TLV($0101, Z+msg) )   // msg-data-1
//    +TLV(5, myUINle+char(MSG_MSG)+flagChar+WNTS(msg) )   // msg-data-4
    +TLV(6, '')
  );
  end;
result:=addRef(REF_msg,uin);
end; // sendMsg

function TicqSession.sendAutoMsgReq(uin:integer):integer;
var
  c:Tcontact;
  msgtype:byte;
  s:Tstatus;
begin
result:=-1;
c:=contactsDB.get(uin);
if c.status <> SC_ONLINE then s:=c.status
else s:=c.prevStatus;
case s of
  SC_OCCUPIED: msgtype:=MSG_AUTOMSG_OCCUPIED;
  SC_NA: msgtype:=MSG_AUTOMSG_NA;
  SC_DND: msgtype:=MSG_AUTOMSG_DND;
  SC_F4C: msgtype:=MSG_AUTOMSG_F4C;
  else msgtype:=MSG_AUTOMSG_AWAY;
  end;
if not isReady then exit;

sendSNAC(4,6, qword_LEasStr(SNACref)+#0#2
  +BUIN(uin)
  +TLV(5, #0#0+qword_LEasStr(SNACref)+capability[1]
    +TLV($A,#0#1)
    +TLV($F,'')
    +TLV($2711, header2711+char(msgtype)+#3+Z+WNTS('') )
  )
  +TLV(3,'')
);
result:=addRef(REF_msg,uin);
end; // sendAutoMsgReq

procedure TicqSession.sendAddedYou(uin:integer);
var
  c:Tcontact;
begin
if not isReady then exit;

c:=contactsDB.get(uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

sendSNAC(4,6, qword_LEasStr(SNACref)+#0#4
  +BUIN(uin)
  +TLV(5, myUINle+char(MSG_ADDEDYOU)+#$00+WNTS('') )
  +TLV(6,'')
);
end; // sendAddedYou

procedure TicqSession.sendContacts(uin,flags:dword; cl:TcontactList);
var
  s:string;
  c:Tcontact;
begin
if not isReady then exit;
if cl.empty then exit;

c:=contactsDB.get(uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);

s:=intToStr(cl.count)+#$FE;
cl.resetEnumeration;
while cl.hasMore do with cl.getNext do
  s:=s+intToStr(uin)+#$FE+nick+#$FE;

sendSNAC(4,6, qword_LEasStr(SNACref)+#0#4
  +BUIN(uin)
  +TLV(5, myUINle+char(MSG_CONTACTS)+#0+WNTS(s))
  +TLV(6,'')
);
addRef(REF_contacts,uin);
end; // sendContacts

procedure TicqSession.sendAuth(uin:integer);
var
  c:Tcontact;
begin
if not isReady then exit;
c:=contactsDB.get(uin);
if not imVisibleTo(c) then
  addTemporaryVisible(c);
sendSNAC(4,6, qword_LEasStr(SNACref)+#0#4
  +BUIN(uin)
  +TLV(5, myUINle+char(MSG_AUTHGIVEN)+#0+WNTS(''))
  +TLV(6, '')
);
addRef(REF_auth,uin);
end; // sendAuth

procedure TicqSession.sendAuthDenied(uin:integer; msg:string);
begin
if not isReady then exit;
sendSNAC(4,6, qword_LEasStr(SNACref)+#0#4
  +BUIN(uin)
  +TLV(5, myUINle+char(MSG_AUTHDENIED)+#0+WNTS(msg))
  +TLV(6, '')
);
end; // sendAuth

procedure TicqSession.sendSimpleQueryInfo(uin:integer);
begin
if not isReady then exit;
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#0#$1F#5
  +dword_LEasStr(uin)
)));
addRef(REF_simplequery, uin);
end; // sendSimpleQueryInfo

procedure TicqSession.sendQueryInfo(uin:integer);
const
  TAB:array [boolean] of char=(#$B2,#$D0);
begin
if not isReady then exit;
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#0+TAB[myinfo.uin=uin]+#4+dword_LEasStr(uin)
)));
addRef(REF_query,uin);
end; // sendQueryInfo

procedure TicqSession.sendWPsearch(wp:TwpSearch);
begin
if not isReady then exit;

wasUINwp:=wp.uin > 0;
if wasUINwp then
  begin
  sendSimpleQueryInfo(wp.uin);
  exit;
  end;
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#0#$33#5
  +WNTS(wp.first)
  +WNTS(wp.last)
  +WNTS(wp.nick)
  +WNTS(wp.email)
  +dword_BEasStr(wp.age)
  +char(wp.gender)
  +char(wp.lang)
  +WNTS(wp.city)
  +WNTS(wp.state)
  +word_LEasStr(wp.country)
  +WNTS('') // company
  +WNTS('') // dep
  +WNTS('') // position
  +#0       // occupation
  +#0#0+WNTS('')  // past information
  +#0#0+WNTS('')  // interest
  +#0#0+WNTS('')  // affiliation
  +#0#0+WNTS('')  // hompage category
  +char(wp.onlineOnly)
)));
addRef(REF_wp,0);
end; // sendWPsearch

procedure TicqSession.sendReqOfflineMsgs;
begin sendSNAC($15,2, TLV(1, Length_LE( myUINle +#$3C#0#0#0))) end;

procedure TicqSession.sendDeleteOfflineMsgs;
begin sendSNAC($15,2, TLV(1, Length_LE( myUINle +#$3E#0#0#0))) end;

procedure TicqSession.sendDeleteUIN;
begin
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#1#0#$C4#4
  +myUINle
  +WNTS(pwd)
)));
end; // sendDeleteUIN

procedure TicqSession.sendPermissions;
const
  tab1:array [boolean] of char=(#1,#0);
  tab2:array [boolean] of char=(#0,#1);
begin
if not isReady then exit;
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#0#$24#4
  +tab1[authNeeded]+tab2[webaware]+#1#0
)));
end; // sendPermissions

procedure TicqSession.sendSMS(dest, msg:string; ack:boolean);
begin
if not isReady then exit;
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#0#$82#14+#0#1#0#$16+StringOfChar(#0,18)
  +Length_BE( xml_sms(myinfo, dest,msg,ack) )
)));
addRef(REF_sms,0);
end; // sendSMS

procedure TicqSession.sendSaveMyInfoAs(c:Tcontact);
begin
if not isReady then exit;
savingMyInfo.running:=TRUE;
savingMyInfo.ACKcount:=0;
savingMyInfo.c:=c;
saveMyInfo1(c);
saveMyInfo2(c);
saveMyInfo3(c);
saveMyInfo4(c);
end; // sendSaveMyInfoAs

procedure TicqSession.saveMyInfo1(c:Tcontact);  // main-home
begin
if not isReady then exit;
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#1#$EA#3
  +WNTS(c.nick)
  +WNTS(c.first)
  +WNTS(c.last)
  +WNTS(c.email)
  +WNTS(c.city)
  +WNTS(c.state)
  +#0#0
  +#0#0
  +#0#0
  +WNTS(c.cellular)
  +WNTS(c.zip)
  +Word_LEasStr(c.country)
  +char(c.GMThalfs)
  +char(publicEmailTab[publicEmail])
)));
addRef(REF_savemyinfo,1);
end; // saveMyInfo1

procedure TicqSession.saveMyInfo2(c:Tcontact);  // homepage-more
var
  s:string;
begin
if c.birth > 1 then
  c.age:=trunc((now-c.birth)/365);
if c.birth > 1 then
  s:=Word_LEasStr(yearOf(c.birth))
    +char(monthOf(c.birth))
    +char(dayOf(c.birth))
else
  s:=Z;

sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#2#$FD#3
  +char(c.age)
  +#0
  +char(c.gender)
  +WNTS(c.homepage)
  +s
  +char(c.lang[1])
  +char(c.lang[2])
  +char(c.lang[3])
)));
addRef(REF_savemyinfo,2);
end; // savemyinfo2

procedure TicqSession.saveMyInfo3(c:Tcontact);  // about
begin
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#3#6#4
  +WNTS(c.about)
)));
addRef(REF_savemyinfo,3);
end; // savemyinfo3

procedure TicqSession.saveMyInfo4(c:Tcontact);  // more emails
begin
sendSNAC($15,2, TLV(1, Length_LE( myUINle
  +#$D0#7#0#4#$B#4#0
)));
addRef(REF_savemyinfo,4);
end; // savemyinfo4

procedure TicqSession.sendChangePwd(newPwd:string);
begin
if not isReady then exit;
waitingNewPwd:=newPwd;
sendSNAC($15,2, TLV(1, Length_LE( myUINle+#$D0#7#0#0#$2E#4+WNTS(newpwd) ) ) );
end; // sendChangePwd

procedure TicqSession.parseCookie(flap:string);
var
  add:string;
  i:integer;
begin
i:=findTLV(8, flap);
if i > 0 then
  begin
  case getTLVwordBE(@flap[i]) of
    $1D:eventError:=EC_loginDelay;
    $18:eventError:=EC_rateExceeded;
    $01:eventError:=EC_badUIN;
    $05:eventError:=EC_badPwd;
    else eventError:=EC_other;
    end;
  notifyListeners(IE_error);
  disconnect;
  exit;
  end;
add:=getTLV(5, flap);
serviceServerAddr:=copy(add,1,pos(':',add)-1);
serviceServerPort:=copy(add,pos(':',add)+1,10);
cookie:=getTLV(6, flap);
sock.close;
sock.WaitForClose;  // prevent to change properties while the socket is open
if http.enabled then
  begin
  sock.addr:=http.addr;
  sock.port:=http.port;
  end
else
  begin
  sock.addr:=serviceServerAddr;
  sock.port:=serviceServerPort;
  end;
phase:=RECONNECTING_;
notifyListeners(IE_redirecting);
sock.Connect;
end; // parseCookie

procedure TicqSession.parseOncomingUser(snac:string);
var
  s, cap:string;
  ofs, i:integer;
  found:boolean;
begin
eventFlags:=0;
eventTime:=now;
ofs:=1;
eventContact:=contactsDB.get(getBUIN(snac,ofs));
inc(ofs, 4);
eventContact.ip:=getTLVdwordBE($A, snac,ofs);

if existsTLV(3, snac,ofs) then
  eventContact.onlineSince:=UnixToDateTime(getTLVdwordBE(3, snac,ofs))+GMToffset
else
  eventContact.onlineSince:=0;

if existsTLV(2, snac,ofs) then
  eventContact.memberSince:=UnixToDateTime(getTLVdwordBE(2, snac,ofs));

if existsTLV($D, snac,ofs) then
	with eventContact do
	  begin
		s:=getTLV($D, snac,ofs);
	  capabilities:=[];
    extracapabilities:='';
    while s > '' do
      begin
      cap:=copy(s,1,16);
      delete(s,1,16);
      found:=FALSE;
      for i:=1 to length(capability) do
      	if cap = capability[i] then
        	begin
          include(capabilities,i);
          found:=TRUE;
          end;
      if not found then
      	extracapabilities:=extracapabilities+cap;
      end;
	  // temporary fix for icq2go, this prevents from using type-2 messages
	  icq2go:=(5 in capabilities) and not (1 in capabilities);
	  end;

s:=getTLV($C, snac,ofs);
if s > '' then
  begin
  eventContact.internal_ip:=dword_BEat(@s[1]);
  eventContact.port:=dword_BEat(@s[5]);
  eventContact.proto:=word_BEat(@s[10]);
  eventContact.dc_cookie:=dword_BEat(@s[12]);
  eventContact.lastupdate_dw:=dword_BEat(@s[24]);
  eventContact.lastinfoupdate_dw:=dword_BEat(@s[28]);
  eventContact.lastStatusUpdate_dw:=dword_BEat(@s[32]);
  eventContact.lastUpdate:=UnixToDateTime(eventContact.lastupdate_dw)+GMToffset;
  eventContact.lastInfoUpdate:=UnixToDateTime(eventContact.lastinfoupdate_dw)+GMToffset;
  eventContact.lastStatusUpdate:=UnixToDateTime(eventContact.laststatusupdate_dw)+GMToffset;
  end
else
  begin
  eventContact.port:=0;
  eventContact.dc_cookie:=0;
  eventContact.proto:=0;
  eventContact.lastupdate_dw:=0;
  eventContact.lastinfoupdate_dw:=0;
  eventContact.laststatusupdate_dw:=0;
  end;
parseStatus(snac,ofs);
with eventContact do
  if status = SC_OFFLINE then     // there could be no status specified, then SC_ONLINE
    begin
		prevStatus:=status;
    status:=SC_ONLINE;
    notifyListeners(IE_oncoming);
    lastTimeSeenOnline:=eventTime;
    end
end; // parseOncomingUser

procedure TicqSession.parseStatus(snac:string; ofs:integer);
var
  newStatus:Tstatus;
  newInvis:boolean;
  code:integer;
begin
if not existsTLV(6, snac,ofs) then exit;

eventContact.prevStatus:=eventContact.status;
eventOldStatus:=eventContact.status;
eventOldInvisible:=eventContact.invisible;
code:=getTLVdwordBE(6, snac,ofs);
newStatus:=code2status(code);
newInvis:=code and flag_invisible>0;
if eventContact.status = SC_OFFLINE then
  begin
  eventContact.status:=newStatus;
  eventContact.invisible:=newInvis;
  notifyListeners(IE_oncoming);
  eventContact.lastTimeSeenOnline:=eventTime;
  end
else
  if (newStatus <> eventOldStatus) or (newInvis<> eventOldInvisible) then
    begin
    eventContact.status:=newStatus;
    eventContact.invisible:=newInvis;
    notifyListeners(IE_statuschanged);
    end;
end; // parseStatus

procedure TicqSession.parseOffgoingUser(snac:String);
var
  ofs:integer;
begin
eventFlags:=0;
eventTime:=now;
ofs:=1;
eventContact:=contactsDB.get(getBUIN(snac,ofs));
eventContact.prevStatus:=eventContact.status;
eventOldStatus:=eventContact.status;
eventOldInvisible:=eventContact.invisible;
eventContact.status:=SC_OFFLINE;
eventContact.invisible:=FALSE;
eventContact.lastTimeSeenOnline:=eventTime;
notifyListeners(IE_offgoing);
end; // parseoffgoingUser

procedure TicqSession.parseContactsString(s:string);
var
  c:Tcontact;
begin
eventContacts:=TcontactList.create;
chop(#$FE,s);      // skippo il numero dei contatti
while s > '' do
  try
    c:=contactsDB.get(strToInt(chop(#$FE,s)));
    c.nick:=chop(#$FE,s);
    eventContacts.add(c);
  except
  end;
end; // parseContactsString

procedure TicqSession.parseAuthString(s:string);
begin
with eventContact do
  begin
  nick:=ifThen( nick='', chop(#$FE,s), nick);
  first:=ifThen( first='', chop(#$FE,s), first);
  last:=ifThen( last='', chop(#$FE,s), last);
  email:=ifThen( email='', chop(#$FE,s), email);
  end;
chop(#$FE,s);   // skip unknown char
eventMsg:=s;
end; // parseAuthString

procedure TicqSession.notificationForMsg(msgtype:byte; flags:byte; urgent:boolean; msg:string; offline:boolean);
begin
if msgtype in MSG_AUTOMSGS then
  begin
  notifyListeners(IE_automsgreq);
  exit;
  end;
if msg='' then exit;
eventFlags:=0;
if flags and $80 > 0 then inc(eventFlags, IF_multiple);
if urgent then inc(eventFlags, IF_urgent);
if offline then inc(eventFlags, IF_offline);
case msgtype of
  MSG_MSG:
    begin
    eventMsg:=msg;
    notifyListeners(IE_msg);
    end;
  MSG_URL:
    begin
    eventMsg:=chop(#$FE,msg);
    eventAddress:=msg;
    notifyListeners(IE_url);
    end;
  MSG_CONTACTS:
    begin
    parseContactsString(msg);
    notifyListeners(IE_contacts);
    end;
  MSG_ADDEDYOU:
    begin
    parseAuthString(msg);
    notifyListeners(IE_addedYou);
    end;
  MSG_AUTHREQ:
    begin
    parseAuthString(msg);
    notifyListeners(IE_authReq);
    end;
  MSG_EMAIL:
    begin
    parsePagerString(msg);
    notifyListeners(IE_email);
    end;
  MSG_MIRABILIS:
    begin
    parsePagerString(msg);
    notifyListeners(IE_fromMirabilis);
    end;
  MSG_WEBPAGER:
    begin
    parsePagerString(msg);
    notifyListeners(IE_webpager);
    end;
  end;
end; // notificationForMsg

procedure TicqSession.parseGCdata(snac:string; offline:boolean=FALSE);
var
  ofs,v:integer;
  s:string;
begin
ofs:=1;
inc(ofs, 20);
if pos('Greeting Card', getDLS(snac, ofs))=0 then exit;
inc(ofs,3);
v:=ord(snac[ofs]) shl 8+ord(snac[ofs+2]);  // get version
inc(ofs,4);
case v of
  $0100,             // 1.0 not tested
  $0101: inc(ofs,4);
  else inc(ofs,12);    // for version 1.2+
  end;
getDLS(snac,ofs);  // version
getDLS(snac,ofs);  // theme
s:='http://www.icq.americangreetings.com/icqorder.pd?mode=send';
s:=s+'&pre_title='+str2url(getDLS(snac,ofs));
s:=s+'&design='+str2url(getDLS(snac,ofs));
s:=s+'&title='+str2url(getDLS(snac,ofs));
s:=s+'&recipient='+str2url(getDLS(snac,ofs));
s:=s+'&text='+str2url(getDLS(snac,ofs));
s:=s+'&sender='+str2url(getDLS(snac,ofs));
inc(ofs,4); // skip version
if v>=$0102 then
  eventAddress:=getDLS(snac,ofs)
else
  eventAddress:=s;
notifyListeners(IE_gcard);
end; // parseGCdata

procedure TicqSession.parseMsgError(snac:string; ref:integer);
begin
eventMsgID:=ref;
eventInt:=word_BEat(@snac[1]);
notifyListeners(IE_msgError);
end; // parseMsgError

procedure TicqSession.parseServerAck(snac:string; ref:integer);
var
  ofs:integer;
begin
eventMsgID:=qword_LEat(@snac[1]);
ofs:=11;
eventContact:=contactsDB.get(getBUIN(snac,ofs));
notifyListeners(IE_serverAck);
end; // parseServerAck

procedure TicqSession.parseIncomingMsg(snac:string);
var
  msg :string;
  ofs, l :integer;
  priority, msgtype, msgflags :byte;
begin
eventMsgID:=qword_LEat(@snac[1]);
ofs:=11;
eventContact:=contactsDB.get(getBUIN(snac,ofs));
inc(ofs, 4);
eventTime:=now;
if existsTLV(2, snac,ofs) then
  myinfo.memberSince:=UnixToDateTime(getTLVdwordBE(2, snac,ofs));
parseStatus(snac,ofs);
case ord(snac[10]) of // msg format
  1:begin
    ofs:=findTLV(2, snac,ofs)+4+3;
    inc(ofs, 3+ord(snac[ofs]));
    l:=word_BEat(snac,ofs)-4;
    inc(ofs, 2+4);
    msg:=copy(snac,ofs,l);
    notificationForMsg(MSG_MSG,0,TRUE,msg,FALSE);
    end;
  2:begin
    ofs:=findTLV(5, snac,ofs)+4;
    case ord(snac[ofs+1]) of
      1:begin
        notifyListeners(IE_fileabort);
        exit;
        end;
      2:begin
        notifyListeners(IE_fileack);
        exit;
        end;
      end;
    inc(ofs, 2+8+16);
    if existsTLV(4, snac,ofs) then
      eventContact.ip:=getTLVdwordBE(4, snac,ofs);
    if existsTLV(5, snac,ofs) then
      eventContact.port:=getTLVwordBE(5, snac,ofs);
    ofs:=findTLV($2711, snac,ofs)+4;
    inc(ofs, ord(snac[ofs])+2);
    inc(ofs, ord(snac[ofs])+2);

    msgtype:=ord(snac[ofs]);
    msgflags:=ord(snac[ofs+1]);
    priority:=ord(snac[ofs+4]);
    inc(ofs,6);
    msg:=getWNTS(snac, ofs);
    // for now we are not able to manage filetransfers
    if msgtype = MSG_FILE then
      begin
      sendACK(ACK_FILEDENY, 'sorry, i''m not able to receive file at the moment', snac);
      exit;
      end;
    if dontBotherStatus and ((priority = 1) or (msgtype in MSG_AUTOMSGS)) then
      begin
      case getStatus of
        SC_dnd: sendACK(ACK_DND,'',snac);
        SC_occupied: sendACK(ACK_OCCUPIED,'', snac);
        end;
      exit;
      end;
    // here we can be bothered :P
    if msgtype=MSG_FILE then
      begin
      eventcontact.ft_port:=word_BEat(@snac[ofs]);
      inc(ofs, 4);
      eventFilename:=getWNTS(snac, ofs);
      eventInt:=dword_LEat(@snac[ofs]);
      if eventFilename > '' then
        notifyListeners(IE_filereq)
      else
        if refs[eventMsgID].kind = REF_file then
          notifyListeners(IE_fileok);
      exit;
      end;
    if msgtype=MSG_GCARD then
      parseGCdata( copy(snac, ofs, length(snac)) )
    else
      notificationForMsg(msgtype, msgflags, priority=2, msg, FALSE);
    case getStatus of
      SC_away: sendACK(ACK_AWAY,'', snac);
      SC_na: sendACK(ACK_NA,'', snac);
      SC_dnd, SC_occupied: if priority=2 then sendACK(ACK_OK, '', snac) else sendACK(ACK_NOBLINK,'', snac)
      else sendACK(ACK_OK, '', snac)
      end;
    end;
  4:begin
    ofs:=findTLV(5, snac,ofs);
    msg:=ptrWNTS(@snac[ofs+10]);
    msgtype:=ord(snac[ofs+8]);
    if msgtype=MSG_GCARD then
      parseGCdata( copy(snac, ofs+4+6+3+length(msg), length(snac)) )
    else
      notificationForMsg(msgtype, ord(snac[ofs+9]), not dontBotherStatus, msg, FALSE);
    end;
  end; // case
end; // parseincomingMsg

procedure TicqSession.parsePagerString(s:string);
begin
eventName:=chop(#$FE,s);
chop(#$FE,s);
chop(#$FE,s);
eventAddress:=chop(#$FE,s);
chop(#$FE,s);
eventMsg:=s;
end; // parsePagerString

procedure TicqSession.parse1503(snac:string; ref:integer);
var
  ofs:integer;

  function readWNTS:string;
  begin result:=getWNTS(snac, ofs) end;

  function readINT:integer;
  begin result:=dword_LEat(@snac[ofs]); inc(ofs, 4) end;

  function readWORD:word;
  begin result:=word_LEat(@snac[ofs]); inc(ofs, 2) end;

  function readBYTE:byte;
  begin result:=ord(snac[ofs]); inc(ofs) end;

  procedure extractWP;
  var
    next:integer;
  begin
  next:=readWORD;
  inc(next,ofs);
  eventwp.uin:=readINT;
  eventwp.nick:=readWNTS;
  eventwp.first:=readWNTS;
  eventwp.last:=readWNTS;
  eventwp.email:=readWNTS;
  eventwp.authRequired:=readBYTE=0;
  eventwp.status:=readBYTE;

  ofs:=next;

  // request issued from white pages
  if wasUINwp or (refs[ref].kind = REF_wp) then
    begin
    notifyListeners(IE_wpResult);
    exit;
    end;

  // request issued for internal use
  eventContact:=contactsDB.get(eventWP.uin);
  with eventContact do
    begin
    nick:=eventwp.nick;
    first:=eventwp.first;
    last:=eventwp.last;
    email:=eventwp.email;
    notifyListeners(IE_userinfo);
    end;
  end; // extractWP

var
  y:word;
  d,m:byte;
  msgtype,msgflags:byte;
  msg:string;
begin
eventContact:=contactsDB.get(refs[ref].uin);
ofs:=11;
case ord(snac[ofs]) of
  $42: notifyListeners(IE_endOfOfflineMsgs);
  $41:  // offline messages
    begin
    inc(ofs,4);
    eventContact:=contactsDB.get(readINT);
    y:=readWORD;
    m:=readBYTE;
    d:=readBYTE;
    eventTime:=EncodeDate(y,m,d);
    d:=readBYTE; // hours
    m:=readBYTE;
    eventTime:=eventTime+EncodeTime(d,m,0,0)+GMToffset;
    msgtype:=readBYTE;
    msgflags:=readBYTE;
    msg:=readWNTS;
    if msgtype=MSG_GCARD then
      parseGCdata(copy(snac,ofs,length(snac)), TRUE)
    else
      notificationForMsg(msgtype, msgflags, not dontBotherStatus, msg, TRUE);
    end;
  $DA:
    case ord(snac[ofs+4]) of
      $9A, $A4:   // simple query and wp result
        begin
        eventContact.nodb:=FALSE;
        eventContact.infoUpdatedTo:=now;
        if ord(snac[ofs+6])=$A then
          begin
          inc(ofs,7);
          extractWP;
          eventInt:=0;
          if wasUINwp then
            notifyListeners(IE_wpEnd);
          end
        else
          if refs[ref].kind = REF_wp then
            begin
            eventInt:=-1;
            notifyListeners(IE_wpEnd);
            end
          else
            begin
            eventContact.nodb:=TRUE;
            eventError:=EC_badContact;
            notifyListeners(IE_error);
            end;
        end;
      $AE:   // last wp result
        begin
        eventContact.infoUpdatedTo:=now;
        if ord(snac[ofs+6])=$A then
          begin
          inc(ofs,7);
          extractWP;
          eventInt:=readINT;
          end
        else
          eventInt:=-1;
        if refs[ref].kind = REF_wp then
          notifyListeners(IE_wpEnd);
        end;
      $E6:   // query result (about)
        begin
        eventContact.infoUpdatedTo:=now;
        inc(ofs,7);
        eventContact.about:=readWNTS;
        notifyListeners(IE_userinfo);
        end;
      $FA:
        begin
        eventContact.infoUpdatedTo:=now;
        if snac[ofs+6]=#$14 then
          eventContact.nodb:=TRUE;
        notifyListeners(IE_userinfo);
        end;
      $C8:   // query result (main, home)
        begin
        inc(ofs,7);
        with eventContact do
          begin
          noDB:=FALSE;
          infoUpdatedTo:=now;
          nick:=readWNTS;
          first:=readWNTS;
          last:=readWNTS;
          email:=readWNTS;
          city:=readWNTS;
          state:=readWNTS;
          // skip 3
          readWNTS;
          readWNTS;
          readWNTS;
          cellular:=readWNTS;
          SMSable:=pos(' SMS',cellular)>0;
          if SMSable then
            delete(cellular,length(cellular)-3,4);
          zip:=readWNTS;
          country:=readWORD;
          GMThalfs:=readBYTE;
          publicEmail:=not boolean(readBYTE);
          end;
        notifyListeners(IE_userinfo);
        end;
      $DC:   // query result (homepage/more)
        begin
        inc(ofs,7);
        with eventContact do
          begin
          infoUpdatedTo:=now;
          age:=readWORD;
          gender:=readBYTE;
          homepage:=readWNTS;
          y:=readWORD;
          m:=readBYTE;
          d:=readBYTE;
          if y > 0 then birth:=encodeDate(y,m,d)
          else birth:=0;
          lang[1]:=readBYTE;
          lang[2]:=readBYTE;
          lang[3]:=readBYTE;
          end;
        notifyListeners(IE_userinfo);
        end;
      $B4: notifyListeners(IE_uinDeleted);
      $AA:
        if ord(snac[ofs+6])=$A then
          begin
          P_pwd:=waitingNewPwd;
          notifyListeners(IE_pwdChanged);
          end
        else
          begin
          eventError:=EC_cantchangePwd;
          notifyListeners(IE_error);
          end;
      $64,$78,$82,$87:   // acks to save-my-info
        begin
        inc(savingMyinfo.ACKcount);
        if savingMyinfo.ACKcount = 4 then
          begin
          savingMyinfo.running:=FALSE;
          sendStatusCode; // needed(?) for the server to save publicemail
          notifyListeners(IE_myinfoACK);
          end;
        end;
      end;//case
  end;//case
end; // parse1503

procedure TicqSession.parseNewUIN(snac:string);
begin
eventContact:=contactsDB.get(dword_LEat(@snac[47]));
notifyListeners(IE_newUIN);
end; // parseNewUIN

var
  myBeautifulSocketBuffer:string;

procedure TicqSession.received(Sender: TObject; Error: Word);
var
  pkt:string;
  channel,ref:integer;
  service:TsnacService;
begin
pkt:=sock.receiveStr;
if (phase in [CONNECTING_,RECONNECTING_]) and http.enabled then
  begin
  myBeautifulSocketBuffer:=myBeautifulSocketBuffer+pkt;
  if pos(CRLFCRLF, myBeautifulSocketBuffer) = 0 then exit;
  pkt:=chop(CRLFCRLF, myBeautifulSocketBuffer);
  eventData:=pkt+CRLFCRLF;
  notifyListeners(IE_serverSent);
  eventError:=EC_other;
  if AnsiStartsStr('HTTPS/1.0 200', pkt)
  or AnsiStartsStr('HTTPS/1.1 200', pkt)
  or AnsiStartsStr('HTTP/1.0 200', pkt)
  or AnsiStartsStr('HTTP/1.1 200', pkt) then
    case phase of
      CONNECTING_:
        begin
        phase:=LOGIN_;
        notifyListeners(IE_connected);
        end;
      RECONNECTING_:
        begin
        phase:=RELOGIN_;
        notifyListeners(IE_redirected);
        end;
      end
  else
    if ansiStartsStr('HTTP/1.0 407', pkt) then
      eventError:=EC_proxy_badPwd
    else
      begin
      eventError:=EC_proxy_unk;
      eventMsg:=pkt;
      end;

  // pass what follows to the snac cruncher
  pkt:=myBeautifulSocketBuffer;
  myBeautifulSocketBuffer:='';
  
  if eventError <> EC_other then
    begin
    notifyListeners(IE_error);
    disconnect;
    exit;
    end;
  end;

Q.add(pkt);
if Q.error then
  begin
  eventData:=q.popError;
  eventError:=EC_invalidFlap;
  notifyListeners(IE_error);
  disconnect;
  end;
while Q.available do
  begin
  pkt:=Q.pop;
  eventData:=pkt;
  notifyListeners(IE_serverSent);

  channel:=getFlapChannel(pkt);
  if channel = SNAC_CHANNEL then
    begin
    service:=getSnacService(pkt);
    ref:=getSnacRef(pkt);
    delete(pkt,1,16);  // remove header
    end
  else
    begin
    service:=0;
    ref:=0;
    delete(pkt,1,6);  // remove header
    end;
  case phase of
    CREATING_UIN_:
      case channel of
        LOGIN_CHANNEL:
          begin
          sendFLAP(LOGIN_CHANNEL, #0#0#0#1);
          sendCreateUIN;
          notifyListeners(IE_creatingUIN);
          end;
        SNAC_CHANNEL:
          if service = $1705 then
            begin
            parseNewUIN(pkt);
            disconnect;
            end
          else
            begin
            eventError:=EC_cantCreateUin;
            notifyListeners(IE_error);
            end;
        end;
    LOGIN_:
      case channel of
        LOGIN_CHANNEL: if pkt=#0#0#0#1 then sendLogin;
        LOGOUT_CHANNEL: parseCookie(pkt);
        end;
    RELOGIN_:
      begin
      case channel of
        LOGIN_CHANNEL: if pkt=#0#0#0#1 then sendCookie;
        SNAC_CHANNEL:
          if service=$0103 then     // server is ready
            begin
            sendImICQ;
            phase:=SETTINGUP_;
            end;
        end;
      end;
    SETTINGUP_:
      case service of
        $0118:
          begin     // ack to I'm ICQ
          sendSNAC(1,6, '');
          sendSNAC(1,$E, '');
          sendSNAC(2,2, '');
          sendSNAC(3,2, '');
          sendSNAC(4,4, '');
          sendSNAC(9,2, '');
          notifyListeners(IE_almostonline);
          end;
        $0107: sendAckTo107;
        $010F: parse010F(pkt);
        $0903:
          begin
          serverStart;
          sendIMparameter;
          sendCapabilities;
          roaster.setStatus(SC_OFFLINE);
          with roaster do
            begin
            resetEnumeration;
            while hasMore do
              with getNext do
                invisible:=FALSE;
            end;
          myinfo.proto:=7;
          myinfo.status:=startingStatus; 
          myinfo.invisible:=startingInvisible;
          sendSimpleQueryInfo(myinfo.uin);
          sendAddContact(roaster);
          previousInvisible:=not myinfo.invisible;
          sendStatusCode;
          sendPermissions;
          sendClientReady;
          phase:=ONLINE_;
          notifyListeners(IE_online);
          end;
        end;
    ONLINE_:
      if channel <> SNAC_CHANNEL then
        begin
        eventError:=EC_serverDisconnected;
        if existsTLV(9,pkt) then
          case getTLVwordBE(9, pkt) of
            1: eventError:=EC_anotherLogin
            end;
        notifyListeners(IE_error);
        disconnect;
        end
      else
        case service of
          $030B: parseOncomingUser(pkt);
          $030C: parseOffgoingUser(pkt);
          $0401: parseMsgError(pkt,ref);
          $040C: parseServerAck(pkt,ref);
          $0407: parseIncomingMsg(pkt);
          $040B: parse040B(pkt);
          $1503: parse1503(pkt, ref);
          $010A: notifyListeners(IE_toofast);
          $010B: notifyListeners(IE_pause);
          end;
    end;//case
  if Q.error then
    begin
    eventData:=q.popError;
    eventError:=EC_invalidFlap;
    notifyListeners(IE_error);
    end;
  end;

eventData:='';
end; // received

procedure TicqSession.sendIMparameter;
begin sendSNAC(4,2, Z+#$00#$03#$1F#$40#$03#$E7#$03#$E7+Z) end;

procedure TicqSession.sendClientReady;
begin sendSNAC(1,2, #$00#$01#$00#$03#$01#$10#$02#$8A#$00#$02#$00#$01#$01#$01#$02#$8A#$00#$03#$00#$01#$01#$10#$02#$8A#$00#$15#$00#$01#$01#$10#$02#$8A#$00#$04#$00#$01#$01#$10#$02#$8A#$00#$06#$00#$01#$01#$10#$02#$8A#$00#$09#$00#$01#$01#$10#$02#$8A#$00#$0A#$00#$01#$01#$10#$02#$8A) end;

procedure TicqSession.sendCapabilities;
begin sendSNAC(2,4, TLV(5,capability[1]+capability[2]+myInfo.extracapabilities) ) end;

procedure TicqSession.sendImICQ;
begin sendSNAC(1,$17, #$00#$01#$00#$03#$00#$02#$00#$01#$00#$03#$00#$01#$00#$15#$00#$01#$00#$04#$00#$01#$00#$06#$00#$01#$00#$09#$00#$01#$00#$0A#$00#$01) end;

procedure TicqSession.sendCookie;
begin
sendFLAP(LOGIN_CHANNEL, #0#0#0#1+TLV(6,cookie));
cookie:=''; // free mem
end; // sendCookie

procedure TicqSession.sendAckTo107;
begin sendSNAC(1,8, #$00#$01#$00#$02#$00#$03#$00#$04#$00#$05) end;

procedure TicqSession.sendLogin;
begin
sendFLAP( LOGIN_CHANNEL, #0#0#0#1
  +TLV(1, myinfo.uinAsStr)
  +TLV(2, encrypted(pwd))
{  +TLV(3, 'ICQ Inc. - Product of ICQ (TM).2000b.4.63.1.3279.85')
  +TLV($16, word($010A))
  +TLV($17, word($0004))
  +TLV($18, word($003F))
  +TLV($19, word($0001))
  +TLV($1A, word($0CCF))
  +TLV($14, integer($55))}
  +TLV(3, 'ICQ Inc. - Product of ICQ (TM).2001b.5.17.1.3642.85')
  +TLV($16, word($010A))
  +TLV($17, word($0005))
  +TLV($18, word($0011))
  +TLV($19, word($0001))
  +TLV($1A, word($0E3A))
  +TLV($14, integer($55))
  +TLV($E,'us')
  +TLV($F,'en')
);
notifyListeners(IE_loggin);
end; // sendLogin

function TicqSession.removeContact(c:Tcontact):boolean;
begin
result:=roaster.remove(c);
if result then
  begin
  removeFromVisible(c);
  sendRemoveContact(c.buin);
  c.status:=SC_UNK;
  eventInt:=roaster.count;
  notifyListeners(IE_numOfContactsChanged);
  end
end; // removeContact

procedure TicqSession.setStatus(s:Tstatus; inv:boolean);
begin
if s = SC_OFFLINE then
  begin
  disconnect;
  exit;
  end;
if (s = myinfo.status) and (inv = myinfo.invisible) then exit;
startingStatus:=s;
startingInvisible:=inv;
if isReady then
  begin
  if inv <> myinfo.invisible then
    clearTemporaryVisible;
  myinfo.status:=s;
  myinfo.invisible:=inv;
  sendStatusCode;
  eventContact:=myinfo;
  notifyListeners(IE_statuschanged);
  end
else
  connect;
end; // setStatus

function TicqSession.getStatus:Tstatus;
begin
if myinfo=NIL then
  result:=SC_UNK
else
  result:=myinfo.status
end;

procedure TicqSession.addContact(cl:TcontactList);
begin
if cl=NIL then exit;
cl:=cl.clone.remove(roaster);
if isReady then
  cl.setStatus(SC_OFFLINE)
else
  cl.setStatus(SC_UNK);
roaster.add(cl);
if isReady then
  sendAddContact(cl);
eventInt:=roaster.count;
notifyListeners(IE_numOfContactsChanged);
cl.free;
end; // addContact

function TicqSession.addContact(c:Tcontact):boolean;
begin
result:=FALSE;
if c=NIL then exit;
result:=roaster.add(c);
if result then
  begin
  if isReady then
    begin
    c.status:=SC_OFFLINE;
    c.invisible:=FALSE;
    sendAddContact(c.buin);
    end;
  eventInt:=roaster.count;
  notifyListeners(IE_numOfContactsChanged);
  end;
end; // addContact

function TicqSession.readInvisible:TcontactList; begin result:=invisibleList end;
function TicqSession.readRoaster:TcontactList;   begin result:=roaster end;
function TicqSession.readVisible:TcontactList;   begin result:=visibleList end;
function TicqSession.readTemporaryVisible:TcontactList; begin result:=tempvisibleList end;

function TicqSession.add2visible(c:Tcontact):boolean;
begin
result:=FALSE;
if c=NIL then exit;
tempVisibleList.remove(c);
result:=not visibleList.exists(c);
if result then
  begin
  removeFromInvisible(c);
  addContact(c);

  visibleList.add(c);
  if isReady and myinfo.invisible then
    begin
    sendAddVisible(c.buin);
    eventContact:=c;
    notifyListeners(IE_visibilityChanged);
    end;
  end;
end; // add2visible

procedure TicqSession.add2visible(cl:TcontactList);
begin
if cl=NIL then exit;
tempVisibleList.remove(cl);
cl:=cl.clone.remove(visibleList);
removeFromInvisible(cl);
addContact(cl);
visibleList.add(cl);
if isReady and myinfo.invisible then
  begin
  sendAddVisible(cl);
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
cl.free;
end; // add2visible

procedure TicqSession.setVisibleList(cl:TcontactList);
var
  tmp:TcontactList;
begin
removeFromInvisible(cl);
tempVisibleList.remove(cl);
tmp:=TcontactList.create;

// remove visible-cl
tmp.add(visibleList).remove(cl);
if not tmp.empty then sendRemoveVisible(tmp);
// add cl-visible
tmp.clear;
tmp.add(cl).remove(visibleList);
if not tmp.empty then sendAddVisible(tmp);

visibleList.assign(cl);

if isReady and myinfo.invisible then
  begin
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
tmp.free;
end; // setVisibleList

procedure TicqSession.setInvisibleList(cl:TcontactList);
var
  tmp:TcontactList;
begin
removeFromVisible(cl);
tempVisibleList.remove(cl);
tmp:=TcontactList.create;

// remove invisible-cl
tmp.add(invisibleList).remove(cl);
if not tmp.empty then sendRemoveInvisible(tmp); // add cl-invisible
// add cl-invisible
tmp.clear;
tmp.add(cl).remove(invisibleList);
if not tmp.empty then sendAddInvisible(tmp);

tmp.free;
invisibleList.assign(cl);

if isReady and not myinfo.invisible then
  begin
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
end; // setInvisibleList

function TicqSession.removeFromVisible(c:Tcontact):boolean;
begin
result:=FALSE;
if c=NIL then exit;
removeTemporaryVisible(c);
result:=visibleList.remove(c);
if result then
  if isReady and myinfo.invisible then
    begin
    sendRemoveVisible(c.buin);
    eventCOntact:=c;
    notifyListeners(IE_visibilityChanged);
    end;
end; // removeFromVisible

procedure TicqSession.removeFromVisible(cl:tcontactlist);
begin
if cl=NIL then exit;
removeTemporaryVisible(cl);

cl:=cl.clone.intersect(visibleList);
visibleList.remove(cl);
if isReady and myinfo.invisible and not cl.empty then
  begin
  sendRemoveVisible(cl);
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
cl.free;
end; // removeFromVisible

function TicqSession.add2invisible(c:Tcontact):boolean;
begin
result:=FALSE;
if c=NIL then exit;
removeTemporaryVisible(c);
result:=invisibleList.add(c);
if result then
  begin
  removeFromVisible(c);
  if isReady and not myinfo.invisible then
    begin
    sendAddInvisible(c.buin);
    eventContact:=c;
    notifyListeners(IE_visibilityChanged);
    end;
  end;
end; // add2invisible

procedure TicqSession.add2invisible(cl:TcontactList);
begin
if cl=NIL then exit;
removeTemporaryVisible(cl);
cl:=cl.clone.remove(invisibleList);
invisibleList.add(cl);
removeFromVisible(cl);
if isReady and not myinfo.invisible then
  begin
  sendAddInvisible(cl);
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
cl.free;
end; // add2invisible

function TicqSession.removeFromInvisible(c:Tcontact):boolean;
begin
result:=FALSE;
if c=NIL then exit;
removeTemporaryVisible(c);
result:=invisibleList.remove(c);
if result then
  if isReady and not myinfo.invisible then
    begin
    sendRemoveInvisible(c.buin);
    eventContact:=c;
    notifyListeners(IE_visibilityChanged);
    end;
end; // removeFromInvisible

procedure TicqSession.removeFromInvisible(cl:TcontactList);
begin
if cl=NIL then exit;
removeTemporaryVisible(cl);

cl:=cl.clone.intersect(invisibleList);
invisibleList.remove(cl);
if isReady and not myinfo.invisible and not cl.empty then
  begin
  sendRemoveInvisible(cl);
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
cl.free;
end; // removeFromInvisible

function TicqSession.addTemporaryVisible(c:Tcontact):boolean;
begin
result:=FALSE;
if not isReady then exit;
result:=TRUE;
tempvisibleList.add(c);
if myinfo.invisible then
  sendAddVisible(c.buin)
else
  sendRemoveInvisible(c.buin);
eventContact:=c;
notifyListeners(IE_visibilityChanged);
end; // addTemporaryVisible

function TicqSession.addTemporaryVisible(cl:Tcontactlist):boolean;
begin
result:=FALSE;
if CL=NIL then exit;
if not isReady then exit;
result:=TRUE;
tempvisibleList.add(cl);
cl.resetEnumeration();
{ cos non va bene, troppi SNAC inviati. bisogna suddividere CL tra visibleTo
{ e not visibleTo, e inviare solo 2 SNAC. }
while cl.hasMore do
  with cl.getNext() do
    if myinfo.invisible then
      sendAddVisible(buin)
    else
      sendRemoveInvisible(buin);
eventContact:=NIL;
notifyListeners(IE_visibilityChanged);
cl.free;
end; // addTemporaryVisible

function TicqSession.removeTemporaryVisible(c:Tcontact):boolean;
begin
result:=tempvisibleList.remove(c);
if not result or not isReady then exit;
if myinfo.invisible then
  sendRemoveVisible(c.buin)
else
  sendAddInvisible(c.buin);
eventContact:=c;
notifyListeners(IE_visibilityChanged);
end; // removeTemporaryVisible

function TicqSession.removeTemporaryVisible(cl:TcontactList):boolean;
begin
result:=TRUE;
cl:=cl.clone.intersect(tempVisibleList);
if isReady and not cl.empty then
  begin
  tempvisibleList.remove(cl);
  if myinfo.invisible then
    sendRemoveVisible(cl)
  else
    sendAddInvisible(cl);
  eventContact:=NIL;
  notifyListeners(IE_visibilityChanged);
  end;
cl.free;
end; // removeTemporaryVisible

procedure TicqSession.clearTemporaryVisible;
begin removeTemporaryVisible(tempVisibleList) end;

function TicqSession.useMsgType2for(c:Tcontact):boolean;
begin result:=not (c.status in [SC_OFFLINE,SC_UNK]) and not c.invisible and not c.icq2go and (c.proto>=7) and (1 in c.capabilities) end;

procedure TicqSession.sendCreateUIN;
const
  s=#3#$46#0#0;
begin
sendSNAC($17,4, #0#1#0#$3B+Z+#$28#0#3#0+Z+Z+s+s+Z+Z+Z+Z+WNTS(pwd)+s+#0#0#$CF#1);
end; // sendCreateUIN

function TicqSession.maxCharsFor(c:Tcontact):integer;
begin
if c.status in [SC_OFFLINE,SC_UNK] then result:=450
else
  if useMsgType2for(c) then result:=7000
  else result:=2540;
end; // maxCharsFor

function TicqSession.imVisibleTo(c:Tcontact):boolean;
begin
result:=isOnline and (
    tempvisibleList.exists(c)
    or (not myinfo.invisible and not invisibleList.exists(c))
    or (myinfo.invisible and visibleList.exists(c))
    );
end; // imVisibleTo

function TicqSession.getLocalIPstr:string;
begin
Result:=sock.GetXaddr;
if compareText(result,'error')=0 then result:='';
end; // getLocalIPstr

function TicqSession.getLocalIP:integer;
begin result:=WSocketResolveHost(getLocalIPstr).S_addr end;

procedure TicqSession.sendACK(status:integer; msg, snac:string);
var
  s,tlv:string;
  ofs:integer;
begin
ofs:=11;
eventContact:=contactsDB.get(getBUIN(snac,ofs));
if not imVisibleTo(eventContact) then exit;
case status of
  ACK_OCCUPIED,
  ACK_DND,
  ACK_AWAY,
  ACK_NA:
    begin
    eventMsg:=msg;
    notifyListeners(IE_sendingAutomsg);
    msg:=eventMsg;
    end;
  end;
inc(ofs, 4);
tlv:=getTLV(5, snac,ofs);
tlv:=getTLV($2711, tlv,1+2+8+16);
s:=copy(tlv,1,47);  // chunk1+chunk2+msgtype+msgflags
s[27]:=#0;  // zeroes firewall details
s:=s+Dword_LEasStr(status)+WNTS(msg);
case ord(tlv[46]) of
  MSG_MSG: s:=s+ Z+#$FF#$FF#$FF#$FF;
  MSG_GCARD:
    begin
    ofs:=pos('Greeting Card',tlv)-4-20;
    s:=s+ copy(tlv, ofs, 4+20+length('Greeting Card')+7) +Z;
    end;
  MSG_FILE: s:=s+ Z+#$01#$00#$00#$C8#$06#$C9#$00+Z;
  end;
sendSNAC(4,$B, copy(snac, 1, 11+ord(snac[11]))+#0#3+s);
end; // sendACK

procedure TicqSession.setWebaware(value:boolean);
begin
P_webaware:=value;
sendStatusCode;
end; // setWebaware

procedure TicqSession.setAuthNeeded(value:boolean);
begin
P_authNeeded:=value;
sendPermissions;
end; // setAuthNeeded

function TicqSession.addRef(k:TrefKind; uin:integer):integer;
begin
result:=SNACref;
refs[SNACref].kind:=k;
refs[SNACref].uin:=uin;
inc(SNACref);
if SNACref > maxRefs then
  SNACref:=1;
end; // addRef

function TicqSession.dontBotherStatus:boolean;
begin result:=getStatus in [SC_occupied,SC_dnd] end;

procedure TicqSession.parse010F(snac:string);
var
  ofs:integer;
begin
ofs:=ord(snac[1])+5;
if existsTLV(3, snac,ofs) then
  myinfo.onlineSince:=UnixToDateTime(getTLVdwordBE(3, snac,ofs))+GMToffset
else
  myinfo.onlineSince:=0;
if existsTLV(2, snac,ofs) then
  myinfo.memberSince:=UnixToDateTime(getTLVdwordBE(2, snac,ofs));
end; // parse010F

procedure TicqSession.parse040B(snac:string);
var
  ofs:integer;
  accept:byte;
begin
ofs:=1;
eventInt:=dword_LEat(@snac[ofs]);
inc(ofs, 10+1+ord(snac[11])+2+47);
accept:=ord(snac[ofs]);
inc(ofs,4);
eventMsg:=getWNTS(snac,ofs);
eventContact:=contactsDB.get(refs[eventInt].uin);
eventOldStatus:=eventContact.status;
eventOldInvisible:=eventContact.invisible;
case accept of
  $0,$C: eventAccept:=AC_ok;
  $9:
    begin
    eventContact.status:=SC_occupied;
    eventAccept:=AC_denied;
    end;
  $A:
    begin
    eventContact.status:=SC_dnd;
    eventAccept:=AC_denied;
    end;
  $4:
    begin
    eventContact.status:=SC_away;
    eventAccept:=AC_away;
    end;
  $E:
    begin
    eventContact.status:=SC_na;
    eventAccept:=AC_away;
    end;
  end;
if eventOldStatus<>eventContact.status then
	eventContact.prevStatus:=eventContact.status;
if (eventOldStatus<>eventContact.status) or (eventOldInvisible<>eventContact.invisible) then
  begin
  eventFlags:=0;
  eventTime:=now;
  notifyListeners(IE_statuschanged);
  end;
case refs[eventInt].kind of
  REF_file:
    begin
    notifyListeners(IE_fileDenied);
    exit;
    end;
  REF_msg, REF_contacts:
    begin
    refs[eventInt].kind:=REF_null;
    notifyListeners(IE_ack);
    end;
  end;
end; // parse040B

function TicqSession.directTo(c:Tcontact):Tdirect;
begin result:=directs.newFor(c) end;

function TicqSession.serverPort:word;
begin try result:=strToInt(server.getxport) except result:=0 end end;

function TicqSession.serverStart:word;
begin
if DCmode = DC_none then
  begin
  result:=0;
  exit;
  end;
server.port:='0';
server.addr:='0.0.0.0';
server.listen;
result:=serverPort;
end; // serverStart

{function TicqSession.getIPasDword_BE:string;
var
  saddr:TSockAddrIn;
  l:integer;
begin
l:=sizeOf(saddr);
if sock.GetSockName(saddr,l)=0 then
  with saddr.sin_addr.s_un_b do
    result:=s_b1+s_b2+s_b3+s_b4
else
  result:=''
end; // getIPasDword_BE
}

procedure TicqSession.connect;
begin connect(FALSE) end;

procedure TicqSession.connect(createUIN:boolean);
begin
if not isOffline then exit;
if (pwd = '') or not createUIN and (myInfo=NIL) then
  begin
  eventError:=EC_missingLogin;
  notifyListeners(IE_error);
  exit;
  end;
creatingUIN:=createUIN;
sock.proto:='tcp';
if http.enabled then
  begin
  sock.Addr:=http.addr;
  sock.Port:=http.port;
  end
else
  begin
  sock.addr:=loginServerAddr;
  sock.port:=loginServerPort;
  end;
phase:=CONNECTING_;
notifyListeners(IE_connecting);
 try
sock.Connect
 except
eventError:=EC_cantconnect;
eventInt:=WSocket_WSAGetLastError;
notifyListeners(IE_error);
goneOffline;
 end;
end; // connect

procedure TicqSession.setDCmode(v:TicqDCmode);
begin
P_dcmode:=v;
sendStatusCode;
end; // setDCmode

function Ticqsession.getFullStatusCode:dword;
begin
result:=0;
case DCmode of
  DC_roaster: inc(result, flag_dcForRoaster);
  DC_uponauth,DC_none: inc(result, flag_dcByRequest);
  end;
if webaware then inc(result, flag_webaware);
if birthdayFlag then inc(result, flag_birthday);
if myinfo.invisible then inc(result, flag_invisible);
inc(result, status2code[myinfo.status]);
end; // getFullStatusCode

var
  TZinfo:TTimeZoneInformation;

INITIALIZATION

contactsDB:=TcontactList.create;
GetTimeZoneInformation(TZinfo);
case GetTimeZoneInformation(TZInfo) of
  TIME_ZONE_ID_STANDARD: GMToffset:=TZInfo.StandardBias;
  TIME_ZONE_ID_DAYLIGHT: GMToffset:=TZInfo.DaylightBias;
  else GMToffset := 0;
  end;
GMToffset:=-(TZinfo.bias+GMToffset)/(24*60);

FINALIZATION

contactsDB.free;

end.
