{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
{$I-}
unit contacts;

interface

uses
  classes, sysutils, registry, flap, types;

type
  // ordinary statuses start with SC_ONLINE and end with SC_OFFLINE
  Tstatus=(SC_ONLINE, SC_OCCUPIED, SC_DND, SC_NA, SC_AWAY, SC_F4C, SC_OFFLINE, SC_UNK);

  Tlanguages=array [1..3] of byte;

  Tcontact=class;

  TcontactProc=procedure(c:Tcontact);

  Tcontact=class
    uin:integer;
    status:Tstatus;
    prevStatus:Tstatus;
    invisible:boolean;
    gender:integer;
    age:integer;
    ip,internal_ip:dword;
    port, ft_port:integer;
    GMThalfs:Shortint;
    display:string;       // if user want to rename this contact
    nick:string;
    first:string;
    last:string;
    email:string;
    city:string;
    state:string;
    about:string;
    zip:string;
    country:word;
    birth:TdateTime;
    lang:Tlanguages;
    homepage:string;
    cellular:string;
    SMSable:boolean;
    memberSince:Tdatetime;          // GMT
    onlineSince:Tdatetime;          // local time
    lastUpdate_dw:dword;
    lastinfoupdate_dw:dword;
    lastStatusUpdate_dw:dword;
    lastUpdate:Tdatetime;           // local time
    lastInfoUpdate:Tdatetime;       // local time
    lastStatusUpdate:Tdatetime;     // local time
    infoUpdatedTo:TdateTime;        // local time
    lastTimeSeenOnline:TdateTime;   // local time
    group:integer;
    proto:integer;
    dc_cookie:dword;
    nodb:boolean;
    capabilities:set of 1..7;
    extracapabilities:string;
    icq2go:boolean;
    data:pointer;
   public
    constructor create(uin_:integer);
    destructor destroy; override;
    function  buin:string;
    function  displayed:string;
    function  equals(c:Tcontact):boolean;
    procedure clear;
    function  getGMT:TdateTime;
    function  GMTavailable:boolean;
    function  uinAsStr:string;
    end; // Tcontact

  TcontactList=class(Tlist)
   protected
    enumIdx:integer;
   public
    procedure resetEnumeration;
    function  hasMore:boolean;
    function  getNext:Tcontact;
    function  buinlist:string;
    function  get(uin:integer):Tcontact;
    function  getAt(idx:integer):Tcontact;
    function  putAt(idx:integer; c:Tcontact):boolean;
    function  exists(c:Tcontact):boolean; overload;
    function  exists(uin:integer):boolean; overload;
    function  empty:boolean;
    function  add(c:Tcontact):boolean; overload;
    function  add(p:pointer):boolean; overload;
    function  add(cl:TcontactList):TcontactList; overload;
    function  remove(c:Tcontact):boolean; overload;
    function  remove(p:pointer):boolean; overload;
    function  remove(cl:TcontactList):TcontactList; overload;
    function  intersect(cl:TcontactList):TcontactList;
    function  toString:string;
    function  fromString(s:string; db:TcontactList):boolean;
    function  clone:TcontactList;
    procedure assign(cl:TcontactList);
    procedure setStatus(st:Tstatus);
    function  idxOf(uin:integer):integer;
    procedure apply(p:TcontactProc);
    function  toIntArray:TIntegerDynArray;
    function  count(group:integer=-1):integer;
    end; // TcontactList

function validUin(uin:integer):boolean;

var
  onContactCreation, onContactDestroying:TcontactProc;

IMPLEMENTATION

function validUin(uin:integer):boolean;
begin result:=(uin > 0) end;

///////////////////////////////////////////////////////////////////////////

constructor Tcontact.create(uin_:integer);
begin
inherited create;
clear;
uin:=uin_;
if assigned(onContactCreation) then onContactCreation(self);
end; // create

destructor Tcontact.destroy;
begin
if assigned(onContactDestroying) then onContactDestroying(self);
inherited destroy;
end; // destroy

procedure Tcontact.clear;
begin
uin:=0;
status:=SC_UNK;
gender:=0;
age:=0;
ip:=0;
internal_ip:=0;
port:=0;
GMThalfs:=100;
display:='';
nick:='';
first:='';
last:='';
email:='';
city:='';
state:='';
about:='';
zip:='';
country:=0;
group:=0;
birth:=0;
infoUpdatedTo:=0;
lastTimeSeenOnline:=0;
fillChar(lang,sizeOf(lang),0);
homepage:='';
cellular:='';
SMSable:=FALSE;
proto:=0;
nodb:=FALSE;
icq2go:=FALSE;
capabilities:=[];
extracapabilities:='';
end; // clear

function Tcontact.getGMT:TdateTime;
begin result:=-GMThalfs/48 end;

function Tcontact.GMTavailable:boolean;
begin result:=abs(GMThalfs)<>100 end;

function Tcontact.equals(c:Tcontact):boolean;
begin result:=assigned(c) and (c.uin=uin) end;

function Tcontact.buin:string;
begin
result:=intToStr(uin);
result:=char(length(result))+result;
end; // buin

function Tcontact.displayed:string;
begin
if display > '' then result:=display else
if nick > '' then result:=nick else
if first > '' then result:=first else
if last > '' then result:=last else
  result:=intToStr(uin);
end; // displayed

function Tcontact.uinAsStr:string;
begin result:=intToStr(uin) end;

///////////////////////////////////////////////////////////////////

function TcontactList.getAt(idx:integer):Tcontact;
begin
if (idx>=0) and (idx<count) then
  result:=Tcontact(items[idx])
else
  result:=NIL
end; // getAt

function TcontactList.idxOf(uin:integer):integer;
var
  min,max,i:integer;
begin
if count = 0 then
  begin
  result:=-1;
  exit;
  end;
min:=0;
max:=count-1;
  repeat
  result:=(min+max) div 2;
  i:=getAt(result).uin;
  if i = uin then
    exit
  else
    if i > uin then
      max:=result-1
    else
      min:=result+1;
  until min > max;
result:=-1;
end; // idxOf

function TcontactList.exists(c:Tcontact):boolean;
begin result:=(c<>NIL) and (idxOf(c.uin)>=0) end;

function TcontactList.exists(uin:integer):boolean;
begin result:=idxOf(uin)>=0 end;

function TcontactList.add(p:pointer):boolean;
begin result:=Tobject(p) is Tcontact and add(Tcontact(p)) end;

function TcontactList.add(c:Tcontact):boolean;
var
  i:integer;
begin
result:=(c<>NIL) and not exists(c);
if result then
  begin
  i:=0;
  while (i<count) and (c.uin > getAt(i).uin) do
    inc(i);
  insert(i, c);
  end;
end; // add

function TcontactList.putAt(idx:integer; c:Tcontact):boolean;
begin
result:=(c<>NIL) and not exists(c);
if result then
  insert(idx, c);
end; // putAt

function TcontactList.buinlist:string;
var
  i:integer;
begin
result:='';
i:=0;
while i < count do
  begin
  result:=result+Tcontact(items[i]).buin;
  inc(i);
  end;
end; // buinList

function TcontactList.empty:boolean;
begin result:= count=0 end;

function TcontactList.remove(c:Tcontact):boolean;
begin result:=inherited remove(c) >= 0 end;

function TcontactList.remove(p:pointer):boolean;
begin result:=Tobject(p) is Tcontact and remove(Tcontact(p)) end;

function TcontactList.add(cl:TcontactList):TcontactList;
var
  i:integer;
begin
result:=self;
if cl=NIL then exit;
for i:=0 to cl.count-1 do
  add(cl.getAt(i));
end; // add

function TcontactList.remove(cl:TcontactList):TcontactList;
begin
result:=self;
if cl=NIL then exit;
inherited assign(cl, laSrcUnique);
end; // remove

function TcontactList.intersect(cl:TcontactList):TcontactList;
begin
result:=self;
if cl=NIL then
  begin
  clear;
  exit;
  end;
inherited assign(cl, laAnd);
end; // intersect

function TcontactList.toString:string;
var
  i:integer;
begin
result:='';
for i:=0 to count-1 do
  result:=result+intToStr(Tcontact(items[i]).uin)+#13#10;
end;

function TcontactList.fromString(s:string; db:TcontactList):boolean;
var
  i:integer;
begin
result:=TRUE;
clear;
while s>'' do
  begin
  i:=pos(#10,s);
  if (i>1) and (s[i-1]=#13) then dec(i);
  if i=0 then i:=length(s)+1;
  try add(db.get(strToInt( copy(s,1,i-1) ))) except result:=FALSE end;
  if s[i]=#13 then inc(i);
  system.delete(s,1,i);
  end;
end; // fromString

function TcontactList.clone:TcontactList;
var
  i:integer;
begin
result:=TcontactList.create;
for i:=0 to count-1 do
  result.add(getAt(i))
end; // clone

procedure TcontactList.resetEnumeration;
begin enumIdx:=0 end;

function TcontactList.hasMore:boolean;
begin result:=enumIdx<count end;

function  TcontactList.getNext:Tcontact;
begin
result:=getAt(enumIdx);
inc(enumIdx);
end; // getNext

procedure TcontactList.assign(cl:TcontactList);
begin if cl=NIL then clear else inherited assign(cl, laCopy) end;

procedure TcontactList.setStatus(st:Tstatus);
var
  i:integer;
begin
for i:=0 to count-1 do
  getAt(i).status:=st;
end; // setStatus

function TcontactList.get(uin:integer):Tcontact;
var
  i:integer;
begin
i:=idxOf(uin);
if i >= 0 then
  result:=getAt(i)
else
  begin
  result:=Tcontact.create(uin);
  add(result);
  end;
end; // getDB

procedure TcontactList.apply(p:TcontactProc);
var
  i:integer;
begin
i:=0;
while i < count do
  begin
  p(Tcontact(items[i]));
  inc(i);
  end;
end;

function TcontactList.toIntArray:TIntegerDynArray;
var
  i:integer;
begin
setlength(result,count);
for i:=0 to count-1 do
  result[i]:=Tcontact(items[i]).uin;
end; // toIntArray

function TcontactList.count(group:integer=-1):integer;
var
  i:integer;
begin
if group=-1 then
  begin
  result:=inherited count;
  exit;
  end;
result:=0;
for i:=0 to count-1 do
  if Tcontact(items[i]).group = group then
    inc(result);
end; // count

end.
