{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit easterDlg;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, utilLib;

type
  Tpuzzle=class
  public
    img:Timage;
    bak:Tbitmap;
    p:array of array of integer;
    nx,ny,dx,dy,vx,vy:integer;
    bg:Tcolor;
    oldClick:TnotifyEvent;
    procedure init(x_,y_:integer; img_:Timage);
    function  frame(x,y:integer):Trect;
    procedure imgClick(Sender: TObject);
    function  swap(x,y:integer):boolean;
    procedure setVoid(x,y:integer);
    procedure drawBak;
    function  check:boolean;
    procedure shuffle;
    procedure won;
    end;

type
  TeasterFrm = class(TForm)
    img: TImage;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  public
    puzzle:Tpuzzle;
  end;

implementation

uses globalLib;

{$R *.dfm}

function Tpuzzle.check:boolean;
var
  x,y:integer;
begin
result:=TRUE;
for x:=0 to nx-1 do
  for y:=0 to ny-1 do
    if p[x,y]<>x+y*nx then
      result:=FALSE;
end; // check

procedure Tpuzzle.drawBak;
begin
img.canvas.copyrect(img.canvas.cliprect, bak.canvas, bak.canvas.cliprect);
end;

procedure Tpuzzle.init(x_,y_:integer; img_:Timage);
begin
img:=img_;
bg:=clLTgray;
bak:=Tbitmap.create;
bak.Width:=img.width;
bak.height:=img.height;
bak.canvas.copyrect(bak.canvas.cliprect, img.canvas, img.canvas.cliprect);

oldClick:=img.onClick;
img.Onclick:=imgClick;
nx:=x_;
ny:=y_;
setLength(p,nx,ny);
dx:=img.width div nx;
dy:=img.height div ny;
for x_:=0 to nx-1 do
  for y_:=0 to ny-1 do
    p[x_,y_]:=x_+y_*nx;
setVoid(nx-1,ny-1);

with img.canvas do
  begin
  for x_:=0 to nx-1 do
    begin
    pen.color:=clLtGray;
    moveTo(x_*dx,0); lineTo(x_*dx,cliprect.bottom);
    pen.color:=clDkGray;
    moveTo(x_*dx-1,0); lineTo(x_*dx-1,cliprect.bottom);
    end;
  for y_:=0 to nx-1 do
    begin
    pen.color:=clLtGray;
    moveTo(0,y_*dy); lineTo(cliprect.right,y_*dy);
    pen.color:=clDkGray;
    moveTo(0,y_*dy-1); lineTo(cliprect.right,y_*dy-1);
    end;
  end;
end; // init

procedure Tpuzzle.setVoid(x,y:integer);
begin
vx:=x;
vy:=y;
with img.canvas do
  begin
  brush.color:=bg;
  brush.Style:=bsDiagCross;
  fillRect(frame(vx,vy));
  end;
end; //  setVoid

function Tpuzzle.frame(x,y:integer):Trect;
begin
result:=rect(x*dx,y*dy,succ(x)*dx-1,succ(y)*dy-1);
if (x>=nx-1) then result.Right:=img.width;
if (y>=ny-1) then result.bottom:=img.height;
end;

procedure Tpuzzle.won;
begin
beep;
drawBak;
img.onClick:=oldClick;
end; // won

procedure Tpuzzle.imgClick(Sender: TObject);
var
  pt:Tpoint;
  x,y:integer;
begin
if assigned(oldClick) then oldClick(sender);
getCursorPos(pt);
pt:=img.screenToClient(pt);
x:=pt.x div dx;
y:=pt.y div dy;
if (abs(x-vx)<2) and (y=vy) or (abs(y-vy)<2) and (x=vx) then
  if swap(x,y) then
    if check then
      won;
end; // imgClick

function Tpuzzle.swap(x,y:integer):boolean;
var
  i:integer;
begin
result:=FALSE;
if (x<0) or (x>=nx) or (y<0) or (y>=ny) then exit;
img.Canvas.copyrect(frame(vx,vy),img.canvas,frame(x,y));
i:=p[x,y];
p[x,y]:=p[vx,vy];
p[vx,vy]:=i;
setVoid(x,y);
result:=TRUE;
end; // swap

procedure Tpuzzle.shuffle;
var
  i,x,y:integer;
  old:boolean;
begin
old:=img.visible;
img.visible:=FALSE;
randomize;
  repeat
  i:=100;
  while i > 0 do
    begin
    x:=vx;
    y:=vy;
    case random(4) of
      0: inc(x);
      1: dec(x);
      2: inc(y);
      3: dec(y);
      end;
    if swap(x,y) then
      dec(i);
    end;
  until not check;
img.visible:=old;
end; // shuffle

procedure TeasterFrm.FormCreate(Sender: TObject);
begin
puzzle:=Tpuzzle.create;
puzzle.init(4,3,img);
puzzle.shuffle;
end;

procedure TeasterFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
action:=caFree;
destroyHandle;
end;

procedure TeasterFrm.FormShow(Sender: TObject);
begin
applyTaskButton(self);
bringForeground:=handle;
end;

procedure TeasterFrm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
with puzzle do
  begin
  case key of
    VK_UP: swap(vx,vy+1);
    VK_DOWN: swap(vx,vy-1);
    VK_RIGHT: swap(vx-1,vy);
    VK_LEFT: swap(vx+1,vy);
    end;
  if check then won;
  end;
end; // keydown

end.
