{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
{$WARN SYMBOL_PLATFORM OFF}
unit events;

interface

uses
  classes, extctrls, ICQv7, contacts, windows, graphics, hook, sysutils;

const
  EK_null         = 00;
  EK_msg          = 01;
  EK_url          = 02;
  EK_contacts     = 03;
  EK_file         = 04;
  EK_authReq      = 05;
  EK_AddedYou     = 06;
  EK_oncoming     = 07;
  EK_offgoing     = 08;
  EK_auth         = 09;
  EK_authDenied   = 10;
  EK_statuschange = 11;
  EK_automsgreq   = 12;
  EK_gcard        = 13;
  EK_automsg      = 14;
  EK_last         = 14;
// adding events remember to initialize supportedBehactions
const
  event2str:array [0..EK_last] of string=(
    '','msg','url','contacts','file','authreq','addedyou',
    'oncoming','offgoing','auth','authdenied','statuschange','automsgreq','gcard','automsg'
  );
  EI_flags=1;
//  EI_shit=3;

  HI_event=-1;
  HI_hashed=-2;
  HI_cryptMode=-3;

Type
  Thevent=class
    kind       :integer;
    who        :Tcontact;
    when       :TdateTime;
    info       :string;
    flags      :integer;
    fpos       :integer;
    cryptMode  :byte;
    expires    :integer;  // tenths of second, negative if permanent
    cl         :Tcontactlist;

    otherpeer  :Tcontact; // used to keep track of other peer when "who" is us
    class function new(kind_:integer; who_:Tcontact; when_:TdateTime; info_:string; flags_:integer):Thevent;
    destructor destroy; override;
    function  ico:Ticon;
    function  pic:integer;
    function  decrittedInfo:string;
    function  font:Tfont;
    function  clone:Thevent;
    function  toString:string;
    function  urgent:boolean;
    procedure setInfo(info_:string);
    procedure appendToHistoryFile(uin:integer=-1);
    function  getBodyText:string;
    function  getHeaderText:string;
    end; // Thevent

  TeventQ=class(Tlist)
   public
    OnNewTop :procedure of object;

    constructor create;
    destructor destroy; override;
    function  add(kind_:integer; c:Tcontact; when:Tdatetime; flags_:integer):Thevent; overload;
    procedure add(ev:Thevent); overload;
    function  pop:Thevent;
    function  top:Thevent;
    function  empty:boolean;
    function  chop:boolean;
    function  find(kind_:integer; c:Tcontact):integer;
    function  removeAt(i:integer):Boolean;
    function  firstEventFor(c:Tcontact):Thevent;
    function  removeEvent(kind_:integer; c:Tcontact):boolean; overload;
    function  removeEvent(c:Tcontact):boolean; overload;
    procedure clear; override;
    procedure fromString(s:string);
    function  toString:string;
    procedure removeExpiringEvents;
    end; // TeventQ

procedure writeHistorySafely(ev:Thevent; other:Tcontact=NIL);
procedure flushHistoryWritingQ;

implementation

uses
  globalLib, mainDlg, forms, dialogs, utilLib, chatDlg, history, viewinfoDlg,
  prefDlg, outboxDlg, roasterLib, strUtils;

function Thevent.clone:Thevent;
begin
result:=Thevent.create;
result.kind:=kind;
result.who:=who;
result.when:=when;
result.info:=info;
result.flags:=flags;
result.fpos:=fpos;
result.cryptMode:=cryptMode;
  try
    result.cl:=cl.clone;
  except
    result.cl:=NIL
  end;
result.expires:=expires;
end; // clone

destructor Thevent.destroy;
begin
cl.free;
inherited;
end;

procedure Thevent.appendToHistoryFile(uin:integer=-1);
var
  s:string;
begin
if uin<0 then
  uin:=who.uin;
s:=userPath+historyPath+intToStr(uin);
fpos:=sizeOfFile(s);
appendFile(s, toString);
end; // appendToHistoryFile

function Thevent.urgent:boolean;
begin result:=flags and IF_urgent > 0 end;

function Thevent.font:Tfont;
begin
if ICQ.myinfo.equals(who) then
  result:=theme.history.myfont
else
  result:=theme.history.hisfont
end;

function Thevent.ico:Ticon;
begin result:=event2ico(kind) end;

function Thevent.pic:integer;
begin result:=event2imgidx(kind) end;

function Thevent.decrittedInfo:string;
begin
case cryptMode of
  CRYPT_SIMPLE: result:=decritted(info, who.uin);
  CRYPT_KEY1: result:=decritted(info, histcrypt.pwdKey);
  end;
end; // decrittedInfo

procedure Thevent.setInfo(info_:string);
begin
if histcrypt.enabled then
  begin
  cryptMode:=CRYPT_KEY1;
  info:=critted(info_, histcrypt.pwdKey);
  end
else
  begin
  cryptMode:=CRYPT_SIMPLE;
  if who<>NIL then
    info:=critted(info_, who.uin);
  end;
end; // setInfo

class function Thevent.new(kind_:integer; who_:Tcontact; when_:TdateTime; info_:string; flags_:integer):Thevent;
begin
result:=Thevent.create;
result.kind:=kind_;
result.who:=who_;
result.when:=when_;
result.flags:=flags_;
result.fpos:=-1;
result.expires:=-1;
result.cl:=NIL;
result.setInfo(info_);
end; // new

function Thevent.toString:string;

  function extraInfo:string;
  begin
  result:=TLV(EI_flags, int2str(flags));
  result:=int2str(length(result))+result;
  end; // extrainfo

begin
result:=int2str(HI_event)+char(kind)+int2str(who.uin)
       +dt2str(when)+extrainfo+int2str(length(info))+info;
end; // toString

function Thevent.getHeaderText:string;
begin
if kind in [EK_ONCOMING, EK_OFFGOING, EK_STATUSCHANGE] then
  result:=statusName(infoToStatus(info))
else
  result:='';
result:=___('history header '+event2str[kind], [
  formatDatetime(timeformat.chat, when),
  who.displayed,
  ifThen(IF_multiple and flags>0, ___('history header multisend')),
  result
]);
end; // getHeaderText

function Thevent.getBodyText:string;
var
  s:string;
  i:integer;
begin
result:='';
case kind of
  EK_AUTH,
  EK_GCARD,
  EK_ADDEDYOU: result:=___('history body '+event2str[kind]);
  EK_AUTHREQ,
  EK_AUTHDENIED: result:=___('history body '+event2str[kind],[decrittedInfo]);
  EK_AUTOMSG:
    begin
    result:=decrittedInfo;
    delete(result,1,1);
    end;
  EK_CONTACTS:
    begin
    result:=decrittedInfo;
    // backward compatibility (converts old format)
    i:=length(result);
    if i>30 then i:=30;
    while (i>0) and (result[i]<>#2) do dec(i);
    if i <= 0 then exit;
    s:=result; result:='';
    while s > '' do
      begin
      chop(#2,s);
      result:=result+chop(', ',s)+CRLF;
      end;
    end;
  EK_URL,
  EK_MSG: result:=decrittedInfo;
  end;
convertAllNewlinesToCRLF(result);
end; // getBodyText

//////////////////////////////////////////////////////////////

constructor TeventQ.create;
begin
inherited create;
blinking:=TRUE;
end; // create

function TeventQ.find(kind_:integer; c:Tcontact):integer;
begin
result:=count;
while result > 0 do
  begin
  dec(result);
  with Thevent(items[result]) do
    if (kind = kind_) and who.equals(c) then
      exit;
  end;
result:=-1;
end; // find

procedure TeventQ.add(ev:Thevent);
var
  i:integer;
begin
if sortBy = SB_event then
  roasterLib.sort(ev.who);
// contacts and authreq requires distint windows for each event
if ev.kind in [EK_contacts,EK_auth] then
  i:=-1
else
  i:=find(ev.kind, ev.who);

if (i >= 0) and (not ev.urgent or Thevent(items[i]).urgent) then
  ev.free
else
  begin
  if ev.flags and IF_urgent > 0 then
    insert(0, ev)
  else
    inherited add(ev);
  if ev.kind in [EK_oncoming,EK_offgoing] then
    ev.expires:=tempBlinkTime;        // tenth of second
  if count = 1 then
    if assigned(OnNewTop) then OnNewTop;
  saveInboxDelayed:=TRUE;
  end;
end; // add

function TeventQ.add(kind_:integer; c:Tcontact; when:Tdatetime; flags_:integer):Thevent;
begin
result:=Thevent.create;
result.kind:=kind_;
result.who:=c;
result.when:=when;
result.flags:=flags_;
add(result);
end; // add

function TeventQ.pop:Thevent;
begin
result:=top;
removeAt(0);
end; // pop

function TeventQ.top:Thevent;
begin if count=0 then result:=NIL else result:=first end;

procedure TeventQ.clear;
begin
while count > 0 do
  pop.free;
end; // clear

destructor TeventQ.destroy;
begin
clear;
inherited;
end; // destroy

function TeventQ.empty:boolean;
begin result:= count=0 end;

function TeventQ.chop:boolean;
begin
result:=FALSE;
if not empty then
  begin
  pop.free;
  result:=TRUE;
  end;
end; // chop

function TeventQ.removeAt(i:integer):boolean;
var
  c:Tcontact;
begin
result:=(i >= 0) and (i < count);
if result then
  begin
  c:=Thevent(items[i]).who;
  delete(i);
  if i=0 then
    if assigned(OnNewTop) then OnNewTop;
  roasterLib.redraw(c);
  end;
end; // removeAt

function TeventQ.firstEventFor(c:Tcontact):Thevent;
var
  i:integer;
begin
i:=0;
while i < count do
  begin
  result:=Thevent(items[i]);
  if result.who.equals(c) then
    exit;
  inc(i);
  end;
result:=NIL;
end; // firstEventFor

function TeventQ.removeEvent(kind_:integer; c:Tcontact):boolean;
begin result:=removeAt(find(kind_,c)) end;

function TeventQ.removeEvent(c:Tcontact):boolean; 
var
  i:integer;
begin
result:=FALSE;
i:=count;
while i > 0 do
  begin
  dec(i);
  if Thevent(items[i]).who.equals(c) then
    begin
    result:=TRUE;
    removeAt(i)
    end
  end;
end; // removeEvent

const
  FK_KIND    = 00;
  FK_EXPIRES = 01;
  FK_WHO     = 02;
  FK_CL      = 03;
  FK_WHEN    = 04;
//  FK_URGENT  = 05;  OBSOLETE
  FK_INFO    = 06;
  FK_FLAGS   = 07;

procedure TeventQ.fromString(s:string);
var
  t,l:integer;
  e:Thevent;
begin
clear;
e:=NIL;
while length(s) >= 8 do
  begin
  t:=integer((@s[1])^); // 1234
  l:=integer((@s[5])^); // 5678
  if not within(0,l,1000000)
  or not within(0,t,100)
  or (length(s) < 8+l) then break; // corrupted file
  case t of
    FK_KIND: e:=add(integer((@s[9])^), ICQ.myInfo, 0, 0);
    FK_EXPIRES: e.expires:=integer((@s[9])^);
    FK_WHO:
      begin
      e.who:=contactsDB.get(integer((@s[9])^));
      NILifNIL(e.who);
      end;
    FK_WHEN: e.when:=Tdatetime((@s[9])^);
    FK_FLAGS: e.flags:=integer((@s[9])^);
    FK_INFO: e.info:=copy(s,9,l);
    FK_CL:
      if l > 0 then
        begin
        e.cl:=TcontactList.create;
        e.cl.fromString(copy(s,9,l), contactsDB );
        end;
    end;//case
  system.delete(s, 1, 8+l);
  end;
end; // fromString

function TeventQ.toString:string;
var
  i:integer;
  s:string;
begin
result:='';
i:=0;
while i < count do
  with Thevent(items[i]) do
    begin
    s:=TLV(FK_KIND, int2str(kind))
      +TLV(FK_EXPIRES, int2str(expires))
      +TLV(FK_WHO, int2str(who.uin))
      +TLV(FK_WHEN, dt2str(when))
      +TLV(FK_FLAGS, int2str(flags))
      +TLV(FK_INFO, info);
    if assigned(cl) then s:=s+TLV(FK_cl, cl.toString);

    result:=result+s;
    inc(i);
    end;
end; // toString

procedure TeventQ.removeExpiringEvents;
var
  i:integer;
begin
i:=0;
while i < count do
  if Thevent(items[i]).expires >= 0 then
    removeAt(i)
  else
    inc(i);
end; // removeExpiringEvents


var
  writingQ:Tlist;

procedure writeHistorySafely(ev:Thevent; other:Tcontact=NIL);
begin
ev:=ev.clone;
if other<>NIL then
  ev.otherpeer:=other;
if ev.otherpeer=NIL then
  ev.otherpeer:=ev.who;
writingQ.add(ev)
end; // addToHistoryWritingQ

procedure flushHistoryWritingQ;
var
  ev:Thevent;
begin
while writingQ.count > 0 do
  begin
  ev:=writingQ.first;
  writingQ.delete(0);
  ev.appendToHistoryFile(ev.otherpeer.uin);
  end;
end; // flushHistoryWritingQ

INITIALIZATION

writingQ:=Tlist.create;

FINALIZATION

writingQ.free;

end.
