{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit history;

interface

uses
  classes, events, contacts, sysutils, graphics, icqv7, dialogs;

const
  CRYPT_SIMPLE=0;
  CRYPT_KEY1=1;
type
  Thistory=class(Tlist)
    loaded :boolean;
    cryptMode :byte;
    hashed :string;
    function  fromString(s:string):boolean;
    function  toString:string;
    function  getAt(idx:integer):Thevent;
    procedure reset;
    procedure deleteFromTo(fn:string; st,en:integer);
    end; // Thistory

implementation

uses
  globalLib, mainDlg, utilLib;

function Thistory.fromString(s:string):boolean;
var
  ev:Thevent;
  cur:integer;

  function getByte:byte;
  begin result:=ord(s[cur]); inc(cur) end;

  function getDatetime:Tdatetime;
  begin result:=Tdatetime((@s[cur])^); inc(cur,8) end;

  function getInt:integer;
  begin result:=integer((@s[cur])^); inc(cur,4) end;

  function getString:string;
  begin result:=copy(s,cur,getInt); inc(cur,length(result)) end;

  procedure parseExtrainfo;
  var
    code,next,extraEnd:integer;
  begin
  extraEnd:=cur+getInt;
  while cur < extraEnd do
    begin
    code:=getInt;
    next:=cur+getInt;
    case code of
      EI_flags: ev.flags:=getInt;
      end;
    cur:=next;
    end;
  end; // parseExtraInfo

begin
cur:=1;
cryptMode:=CRYPT_SIMPLE;
hashed:='';
while cur <= length(s) do
  begin
  ev:=Thevent.create;
  ev.fpos:=cur-1;
  case getInt of
    HI_event:
      begin
      ev.cryptMode:=cryptMode;
      ev.kind:=getByte;
      ev.who:=contactsDB.get(getInt);
      ev.when:=getDatetime;
      parseExtrainfo;
      ev.info:=getString;
      add(ev);
      end;
    HI_hashed: hashed:=getString;
    HI_cryptMode:
      begin
      getInt; // skip length
      cryptMode:=getByte;
      end;
    else
      begin
      msgDlg(___('corrupted history'),mtError);
      result:=FALSE;
      exit;
      end;
    end;
  end;
loaded:=TRUE;
result:=TRUE;
end; // fromString

function Thistory.toString:string;
var
  i,dim:integer;

  procedure addStr(s:string);
  begin
  while dim+length(s) > length(result) do
    setLength(result, length(result)+10000);
  system.move(s[1], result[dim+1], length(s));
  inc(dim, length(s));
  end; // addStr

begin
result:='';
dim:=0;

if histcrypt.enabled then
  addStr(TLV(HI_cryptMode, char(cryptMode))
       + TLV(HI_hashed, hashed) );
i:=0;
while i < count do
  begin
  addStr( getAt(i).toString );
  inc(i);
  end;
setLength(result, dim);
end; // toString

function Thistory.getAt(idx:integer):Thevent;
begin
if (idx >= 0) and (idx < count) then
  result:=Thevent(items[idx])
else
  result:=NIL
end; // getAt

procedure Thistory.reset;
var
  i:integer;
begin
loaded:=FALSE;
i:=0;
while i < count do
  begin
  Thevent(items[i]).free;
  inc(i);
  end;
clear;
end; // reset

procedure Thistory.deleteFromTo(fn:string; st,en:integer);
var
  i:integer;
begin
i:=st;
while (st>=en) and (getAt(en).fpos < 0) do dec(en);
while i <= en do
  begin
  if getAt(i).fpos < 0 then
    begin
    if i > st then
      utilLib.deleteFromTo(fn, getAt(st).fpos, getAt(i-1).fpos+length(getAt(i-1).toString));
    st:=i+1;
    end;
  inc(i);
  end;
if st > en then exit;
utilLib.deleteFromTo(fn, getAt(st).fpos, getAt(en).fpos+length(getAt(en).toString));
for i:=en downto st do
  begin
  Thevent(items[i]).free;
  delete(i);
  end;
end; // deleteFromTo

end.
