{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit historyVCL;

interface

uses
  windows, controls, history, classes, contacts, sysutils, graphics,
  forms, stdctrls, messages, clipbrd, incapsulate, events, strutils;

type
  TlinkKind=( LK_FTP, LK_EMAIL, LK_WWW, LK_UIN, LK_ED );
const
  linksToUnderline:set of TlinkKind=[LK_FTP, LK_EMAIL, LK_WWW,LK_ED];
type
  TitemKind=( PK_NONE, PK_HEAD, PK_TEXT, PK_ARROWS_UP, PK_ARROWS_DN, PK_LINK,
    PK_SMILE, PK_CRYPTED );
  PhistoryItem=^ThistoryItem;
  ThistoryLink=record
    evIdx:integer;    // -1 for null links
    str:string;
    from,to_:integer;
    kind:TlinkKind;
    id:integer;
    ev:Thevent;
    end;
  ThistoryItem=record
    kind: TitemKind;   // PK_NONE for null items
    ev: Thevent;
    evIdx, ofs, l: integer;
    r: Trect;
    link: ThistoryLink;
    end;
  ThistoryPos=record
    ev: Thevent;    // NIL for null positions
    evIdx: integer; // -1 for void positions
    ofs:integer;    // -1 when the whole event is selected
    end;

  ThistoryBox=class(TcustomControl)
   private
    items:array of ThistoryItem;
    lastTimeClick:TdateTime;
    selecting:boolean;
    justTriggeredAlink, dontTriggerLink, just2clicked:boolean;
    Nrows: integer;
    P_lastEventIsFullyVisible: boolean;
    lastClickedItem, P_pointedSpace, P_pointedItem: ThistoryItem;
    P_topEventNrows, P_bottomEvent: integer;
    linkToUnderline: ThistoryLink;
    avoidErase: boolean;
   public
    topVisible, topOfs: integer;
    offset: integer; // can't show hevents before this
    startSel, endSel: ThistoryPos;
    who: Tcontact;
    history: Thistory;
    margin: Trect;
    startWithLastLine: boolean;
    onPainted: TNotifyEvent;
    property color;
    property canvas;
    property onDragOver;
    property onDragDrop;
    property lastEventIsFullyVisible:boolean read P_lastEventIsFullyVisible;
    property pointedItem:ThistoryItem read P_pointedItem;
    property pointedSpace:ThistoryItem read P_pointedSpace;
    property topEventNrows:integer read P_topEventNrows;
    property bottomEvent:integer read P_bottomEvent;

    constructor create(owner_:Tcomponent); override;
    procedure paint(); override;
    procedure paintOn(cnv:Tcanvas);
    function  getSelText():string;
    function  getSelHtml(smiles:boolean):string;
    procedure mouseDown(Button: TMouseButton; Shift: TShiftState; X,Y:Integer); override;
    procedure mouseUp(Button: TMouseButton; Shift: TShiftState; X,Y:Integer); override;
    procedure mouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure click; override;
    procedure go2end();
    function  offsetPos:integer;
    function  triggerLink(item:ThistoryItem):boolean;
    function  itemAt(pt:Tpoint):ThistoryItem;
    function  spaceAt(pt:Tpoint):ThistoryItem;
    function  wholeEventsAreSelected():boolean;
    function  nothingIsSelected():boolean;
    function  partialTextIsSelected():boolean;
    function  somethingIsSelected():boolean;
    procedure select(from,to_:integer);
    procedure deselect();
    procedure updatePointedItem();
    end; // ThistoryBox

implementation

uses
  globalLib, mainDlg, chatDlg, smilesLib, utilLib, icqv7, Types;

function minor(a,b:ThistoryPos):boolean; overload;
begin result:=(a.evIdx < b.evIdx) or (a.evIdx=b.evIdx) and (a.ofs < b.ofs) end;

function equal(a,b:ThistoryPos):boolean; overload;
begin result:=(a.evIdx = b.evIdx) and (a.ofs = b.ofs) end;

function minor(a,b:ThistoryItem):boolean; overload;
begin result:=(a.evIdx < b.evIdx) or (a.evIdx=b.evIdx) and (a.ofs < b.ofs) end;

function equal(a,b:ThistoryItem):boolean; overload;
begin
result:=(a.evIdx = b.evIdx)
and (a.ofs = b.ofs)
and (a.kind = b.kind)
end;

function equal(a,b:ThistoryLink):boolean; overload;
begin result:=(a.evIdx = b.evIdx) and (a.from = b.from) and (a.to_ = b.to_) end;

function minor(a,b:Tpoint):boolean; overload;
begin result:=(a.y<b.y) or (a.y=b.y) and (a.x<b.y) end;

function isLink(it:ThistoryItem):boolean;
begin result:=it.kind=PK_LINK end;

function historyitem2pos(a:ThistoryItem):ThistoryPos;
begin
if a.kind=PK_NONE then
  begin
  result.ev:=NIL;
  result.evIdx:=-1;
  end
else
  begin
  result.ev:=a.ev;
  result.evIdx:=a.evIdx;
  result.ofs:=a.ofs;
  end
end;

constructor ThistoryBox.create(owner_:Tcomponent);
begin
inherited create(owner_);
avoidErase:=FALSE;
tabStop:=FALSE;
P_lastEventIsFullyVisible:=FALSE;
onPainted:=NIL;
end; // create

procedure ThistoryBox.paintOn(cnv:Tcanvas);
var
  x, y, i, j, evIdx, chunkStart, lastLineStart, quoteCounter, lineHeight,
  skippedLines, bodySkipCounter, rightLimit, bottomLimit, foundSmileIdx,
  Nitems: integer;
  linkTheWholeBody, s: string;
  SOS, EOS: ThistoryPos;
  ev: Thevent;
  c: Tcontact;
  selectedClr: Tcolor;
  hls: Thls;
  pic: Tbitmap;
  whatFound: ( _nothing, _wrap, _return, _smile, _bold, _underline );
  foundSmile: string;
  firstCharactersForSmiles: set of char;  // for faster smile recognition
  foundLink: ThistoryLink;
  mouse: Tpoint;
  eventFullyPainted, firstEvent,
  nowBold, nowUnderline, pleaseDontDrawUpwardArrows: boolean;

  procedure newLine();
  begin
  if bodySkipCounter<=0 then inc(y, lineHeight)
  else inc(skippedLines);
  x:=margin.left;
  lineHeight:=0;
  inc(Nrows);
  dec(bodySkipCounter);
  end; // newLine

  function findSmile():boolean;
  var
    j:integer;
    smile:string;
  begin
  result:=FALSE;
  if not (s[i] in firstCharactersForSmiles) then exit;
  foundSmileIdx:=-1;
  for j:=0 to theme.smiles.n-1 do
    begin
    smile:=theme.smiles.ascii[j];
    if matches(s,i,smile) and (theme.smiles.pics[j]<>NIL)
    and ((foundSmileIdx<0) or (length(smile) > length(foundSmile))) then
      begin
      if (length(s) >= i+length(smile))
      and (smile[1]=':')
      and (s[i+length(smile)] in ['a'..'z','0'..'9','A'..'Z',#128..#255])
      and (smile[length(smile)]<>s[i+length(smile)])
      then continue;      
      foundSmile:=smile;
      foundSmileIdx:=j;
      result:=TRUE;
      end;
    end;
  end; // findSmile

  function isEmailAddress(s:string; start:integer; var end_:integer):boolean;
  var
    j:integer;
    existsDot:boolean;
  begin
  result:=FALSE;
  j:=start;
  // try to find the @
  while (j <= length(s)) and (s[j] in EMAILCHARS) and (j-start<30) do
    inc(j);
  if s[j]<>'@' then exit;
  // @ found, now skip the @ and search for .
  inc(j);
  existsDot:=FALSE;
  while (j < length(s)) and (s[j+1] in EMAILCHARS) do
    begin
    if s[j]='.' then
      begin
      existsDot:=TRUE;
      break;
      end;
    inc(j);
    end;
  // at least a valid char after the . must exists
  if not existsDot or not (s[j] in EMAILCHARS) then exit;
  // go forth till we're out or we meet an invalid char
  repeat inc(j)
  until (j > length(s)) or not (s[j] in EMAILCHARS);
  end_:=j-1;
  if s[end_] = '.' then dec(end_);
  //while s[start-1] in emailChar do dec(start);
  result:=TRUE;
  end; // isEmailAddress

  function isUIN(s:string; start:integer; var end_:integer):boolean;
  var
    i:integer;
  begin
  result:=FALSE;
  i:=start;
  while (i <= length(s)) and (s[i] in ['0'..'9']) and (i-start<10) do
    inc(i);
  if i-start > 6 then
    begin
    end_:=i-1;
    result:=TRUE;
    end;
  end; // isUIN

  function findLink():boolean;

    procedure setResult(lk:TlinkKind; end_:integer=0);
    const
		  allowedChars: array [TlinkKind] of set of char=( FTPURLCHARS, EMAILCHARS,
		  	WEBURLCHARS, ['0'..'9'], EDURLCHARS );
    begin
    if end_ = 0 then
    	begin
	    end_:=i;
	    while (end_ < length(s)) and (s[end_+1] in allowedChars[lk]) do
	      inc(end_);
      end;
    if (end_>0) and (end_<=length(s)) then
      while s[end_] in ['?',')','.',','] do dec(end_);
    foundLink.str:=copy(s,i,end_-i+1);
    foundLink.from:=i;
    foundLink.to_:=end_;
    foundLink.kind:=lk;
    inc(foundlink.id);
    result:=TRUE;
    end;

  var
    e:integer;
  begin
  result:=FALSE;
  if linkTheWholeBody > '' then
    begin
    setResult(LK_WWW, length(s));
    foundLink.str:=linkTheWholeBody;
    exit;
    end;
  if isEmailAddress(s,i,e) then
    setResult(LK_EMAIL, e)
  else
    case upcase(s[i]) of
      'H':if Imatches(s,i,'http://') or Imatches(s,i,'https://') then
      			setResult(LK_WWW);
      'W':if Imatches(s,i,'www.') or Imatches(s,i,'web.') then
            setResult(LK_WWW);
      'F':if Imatches(s,i,'ftp://') or Imatches(s,i,'ftp.') then
            setResult(LK_FTP);
      'E':if Imatches(s,i,'ed2k://') then
      			setResult(LK_ED);
      '1'..'9': if isUIN(s,i,e) then setResult(LK_UIN, e);
      end;
  end; // findLink

  function findFidonet(sym:char):boolean;
  var
    j:integer;
  begin
  result:=FALSE;
  if (s[i] <> sym)
  or ((i>1) and not (s[i-1] in WHITESPACES)) // word begin
  or (i+2 > length(s)) then exit;
  j:=i+1;
  while (j<length(s)) and (s[j] in ALPHANUMERIC) do inc(j);
  if (s[j]<>sym) or (j-i=1) then exit; // ends with sym, and no 2 sym bound
  result:=TRUE; 
  end; // findFidonet

  procedure newLineHeight(v:integer); overload;
  begin if lineHeight < v then lineHeight:=v end;

  procedure newLineHeight(s:string); overload;
  begin newLineHeight(cnv.textHeight(s)) end;

  function addItem(k:TitemKind; o,l:integer; r:Trect):PhistoryItem;
  begin
  inc(Nitems);
  if length(items) < Nitems then
    setLength(items, length(items)+20);
  result:=@items[Nitems-1];
  result.kind:=k;
  result.ev:=ev;
  result.evIdx:=evIdx;
  result.ofs:=o;
  result.l:=l;
  result.r:=r;
  if k=PK_LINK then result.link:=foundLink;
  end; // addItem

  function withinTheLink(i:integer):boolean;
  begin
  result:=(foundLink.from > 0)
    and (i >= foundLink.from)
    and (i <= foundLink.to_)
  end; // withinTheLink

  procedure drawBody(cnv:Tcanvas);
  var
    lastSmileChar: char;
    len, smileCount: integer;
    size: Tsize;
    first, bool, wasInsideSelection: boolean;
    r,intersect:Trect;
    nowPos: ThistoryPos;
    quoteCounting:boolean;

    function withinTheSelection(i:integer):boolean;
    begin
    nowPos.evIdx:=evIdx;
    nowPos.ev:=ev;
    nowPos.ofs:=i;
    result:=(SOS.ev<>NIL) and (EOS.ev<>NIL) and (
      (SOS.ofs<0) and within(SOS.evIdx,evIdx,EOS.evIdx)
      or not minor(EOS,nowPos) and minor(SOS,nowpos)
    );
    end;

    procedure applyFont();
    begin
    if quoteCounter=0 then
      cnv.font.assign(ev.font)
    else
      if odd(quoteCounter) xor ICQ.myinfo.equals(who) then
        cnv.font.assign(theme.history.myquoted)
      else
        cnv.font.assign(theme.history.hisquoted);

    applyUserCharset(font);

    if nowBold then
      with cnv.Font do style:=style+[fsBold];
    if nowUnderline then
      with cnv.Font do style:=style+[fsUnderline];
    end; // applyFont

  begin
  if (s='') or (y >= bottomLimit) then exit;
  // draw upward arrows
  if firstEvent and (topOfs > 0) then
    begin
    bodySkipCounter:=topOfs;
    if not pleaseDontDrawUpwardArrows then
      begin
      pic:=getPic(PIC_SCROLL_UP);
      x:=margin.left;
      r:=rect(x,y, rightLimit, y+pic.height);
      addItem(PK_ARROWS_UP, 0,0, r);
      while x < rightLimit do
        inc(x, drawPic(cnv, x,y, pic).cx);
      inc(y, pic.height );
      end;
    end
  else
    bodySkipCounter:=0;

  quoteCounter:=0;
  whatFound:=_nothing;
  i:=1;
  foundLink.from:=0;
  lastLineStart:=1;
  lineHeight:=0;
  x:=margin.left;
  cnv.brush.color:=theme.history.bgcolor;
  cnv.font.assign(ev.font);
  wasInsideSelection:=FALSE;
  len:=length(s);
  quoteCounting:=TRUE;
  nowBold:=FALSE;
  nowUnderline:=FALSE;
  // loop until there's text to be painted
  while (y < bottomLimit) and (i <= len) do
    begin
    chunkStart:=i;
    case whatFound of
      _nothing:
        begin
        applyFont();
        j:=x;
        // go forth, until sth special is found
        while i <= len do
          begin
          // reached the end of the link, stop, we must paint it underlined
          if (foundLink.from > 0) and (i > foundLink.to_) then break;
          // reached a selection edge, stop, we must paint it selected
          bool:=withinTheSelection(i);
          if wasInsideSelection <> bool then
            begin
            wasInsideSelection:=bool;
            break;
            end;
          // things to consider only outside a link
          if foundLink.from = 0 then
            begin
            if s[i] in [#10,#13] then begin whatFound:=_return; break end;
            if useSmiles and findSmile() then begin whatFound:=_smile; break end;
            if s[i] < #32 then s[i]:=#32; // convert control chars
            if findLink() then break;
            if quoteCounting then
              if s[i]='>' then inc(quoteCounter)
              else
              	{ quoting sequence terminates where a non-">" char is found
                { or a non-single blankspace is found or a non-">"-preceeded
                { blankspace is found }
                if (s[i]<>' ') or (quoteCounter=0) or (i=1) or (s[i-1]<>'>') then
                  quoteCounting:=FALSE;
            if fontstylecodes.enabled then
              begin
              if (s[i]='*') and (nowBold or findFidonet('*')) then
                 begin whatFound:=_bold; break end;
              if (s[i]='_') and (nowUnderline or findFidonet('_')) then
                 begin whatFound:=_underline; break end;
              end;
            end;
          applyFont();
          size:=txtSizeL(cnv.handle, @s[i], 1);
          inc(j, size.cx);
          if j > rightLimit then // no more room
            begin
            // search backward for a good place where to split
            j:=i;
            repeat dec(j) until (j=lastLineStart) or (s[j] in ['-',' ']);
            // found. choose it
            if j>chunkStart then i:=j+1;
            whatFound:=_wrap;
            break
            end;
          r:=rect(j-size.cx,y,j,y+size.cy);
          if bodySkipCounter<=0 then
            if withinTheLink(i) then addItem(PK_LINK, i,1, r)
            else addItem(PK_TEXT, i,1, r);
          inc(i);
          end;//while
        // no text, suddenly a break comes
        if i = chunkStart then continue;

        j:=i-chunkStart; // = length of text
        applyFont();
        size:=txtSizeL(cnv.handle, @s[chunkStart], j); // size on screen
        // is it a link?
        if withinTheLink(chunkStart)
        and (evIdx = linkToUnderline.evIdx)
        and within(linkToUnderline.from,chunkStart,linkToUnderline.to_) then
          with cnv.font do style:=style+[fsUnderline];
        newLineHeight('I');
        if withinTheSelection(chunkStart) then cnv.brush.color:=selectedClr
        else cnv.brush.color:=theme.history.bgcolor;
        // finally paint the text
        if bodySkipCounter<=0 then textOut(cnv.handle, x,y, @s[chunkStart], j);
        inc(x, size.cx);
        if (i >= foundLink.to_) then foundLink.from:=0;
        continue;
        end;
      _underline:
        begin
        inc(i);
        nowUnderline:=not nowUnderline;
        end;
      _bold:
        begin
        inc(i);
        nowBold:=not nowBold;
        end;
      _return:
        begin
        case s[i] of
          #10: inc(i);
          #13:
            begin
            inc(i);
            if (i<=len) and (s[i]=#10) then inc(i);
            end;
          end;
        nowBold:=FALSE;
        nowUnderline:=FALSE;
        quoteCounting:=TRUE;
        quoteCounter:=0;
        newLineHeight('I');
        lastLineStart:=i;
        newLine();
        end;
      _smile:
        begin
        // count times smile has to be repeated by last character
        smileCount:=1;
        inc(i, length(foundSmile));
        lastSmileChar:=foundSmile[length(foundSmile)];
        while (i<=len) and (s[i]=lastSmileChar) do
          begin
          inc(i);
          inc(smileCount);
          end;
        pic:=theme.smiles.pics[foundSmileIdx];
        newLineHeight(pic.height+1);
        // paint
        size.cx:=pic.width+1;
        size.cy:=pic.height;
        cnv.brush.color:=selectedClr;
        first:=TRUE;
        while smileCount > 0 do
          begin
          if x+size.cx > rightLimit then
            begin
            newLine();
            newLineHeight(pic.height+1);
            end;
          // only the first one has full length
          if first then j:=length(foundSmile) else j:=1;
          r:=rect(x,y,x+size.cx,y+size.cy);
          if bodySkipCounter<=0 then
            begin
            addItem( PK_SMILE, chunkStart,j, r);
            if withinTheSelection(chunkStart) then
              cnv.fillRect(r);
            cnv.draw(x,y+(lineHeight-size.cy) div 2, pic);
            end;
          inc(chunkStart);
          inc(x, size.cx);
          first:=FALSE;
          dec(smileCount);
          end;
        end;
      _wrap: newLine();
      end;//case
    whatFound:=_nothing;
    end;//while
  newLine();
  if (i > len) and (y <= bottomLimit) then
    begin
    eventFullyPainted:=TRUE;
    exit;
    end;
  // downward arrows
  pic:=getPic(PIC_SCROLL_DOWN);
  x:=margin.left;
  y:=bottomLimit-pic.height;
  r:=rect(x,y, rightLimit, y+pic.height);
  while intersectRect(intersect, items[Nitems-1].r, r) do
    dec(Nitems);
  addItem(PK_ARROWS_DN, 0,0, r);
  // clear
  cnv.brush.color:=theme.history.bgcolor;
  r.Bottom:=cnv.ClipRect.bottom;
  cnv.fillRect(r);
  // draw
  while x < rightLimit do
    inc(x, drawPic(cnv, x,y, pic).cx);
  inc(y, pic.height );
  end; // drawBody

  procedure drawHeader(cnv:Tcanvas);
  begin
  lineHeight:=0;
  x:=margin.left;
  cnv.brush.color:=theme.history.bgcolor;
  c:=ev.who;
  // shall we paint the header as selected?
  if wholeEventsAreSelected and within(SOS.evIdx,evIdx,EOS.evIdx) then
    cnv.brush.color:=selectedClr;
  /// draw header
  inc(x, drawPic(cnv, x,y,ev.pic).cx+3);
  cnv.font.assign(ev.font);
  applyUserCharset(cnv.font);
  newLineHeight('I');
  newLineHeight(getPic(ev.pic).height);
  cnv.textOut(x,y+1,s);
  x:=cnv.penpos.x;

  // some events draws an extra icon on the right
  pic:=NIL;
  case ev.kind of
    EK_ONCOMING,
    EK_STATUSCHANGE:
      begin
      s:=ev.info;
      if length(s)>=4 then
        pic:=getPic(status2imgIdx(Tstatus(str2int(s)), (length(s)>4) and boolean(s[5])));
      end;
    EK_OFFGOING: pic:=getPic(status2imgIdx(SC_OFFLINE));
    end;
  if pic<>NIL then drawPic(cnv, x+2,y, pic);

  addItem(PK_HEAD, -1,0, rect(margin.left,y,x,y+lineHeight));
  j:=x;
  inc(lineHeight);
  newLine();
  // underline
  cnv.pen.color:=selectedClr;
  cnv.moveTo(0, y-1); cnv.lineTo(j, y-1);
  inc(y,1);
  end; // drawHeader

begin
Nitems:=0;
setLength(items, 0);
P_lastEventIsFullyVisible:=TRUE;
P_bottomEvent:=-1;
if topVisible < offset then exit;

// sort startsel and endSel
if minor(startSel, endSel) then
  begin
  SOS:=startsel;
  EOS:=endSel;
  end
else
  begin
  SOS:=endSel;
  EOS:=startsel;
  end;
// calculates a darker/brighter color
hls:=color2hls(theme.history.bgcolor);
with hls do if l>0.5 then l:=l-0.2 else l:=l+0.2;
selectedClr:=hls2color(hls);
margin:=rect(2,2,2,2);
// finds all first characters of all smiles
firstCharactersForSmiles:=[];
for i:=0 to theme.smiles.n-1 do
  if theme.smiles.pics[i]<>NIL then
    include(firstCharactersForSmiles, theme.smiles.ascii[i][1]);

mouse:=screenToClient(mousePos);
cnv.brush.color:=theme.history.bgcolor;
if not avoidErase then cnv.fillRect(clientRect);
rightLimit:=clientWidth-margin.right;
bottomLimit:=clientHeight-margin.bottom;
y:=margin.Top;
evIdx:=topVisible;
foundlink.id:=0;
Nrows:=0;
P_topEventNrows:=0;
firstEvent:=TRUE;
skippedLines:=0;
pleaseDontDrawUpwardArrows:=FALSE;
while (y < bottomLimit) and (evIdx < history.count) do
  begin
  ev:=history.getAt(evIdx);
  foundlink.ev:=ev;
  foundlink.evIdx:=evIdx;
  eventFullyPainted:=FALSE;
  bodySkipCounter:=0;
  s:=ev.getHeaderText;
  if ev.kind = EK_GCARD then linkTheWholeBody:=ev.decrittedInfo
  else linkTHeWholeBody:='';
  drawHeader(cnv);
  // if there is enough space for the body
  if y < bottomLimit then
    begin
    // gets the text to be painted
    s:=ev.getBodyText;
    eventFullyPainted:= s='';
    if startWithLastLine and firstEvent then
      begin
      pleaseDontDrawUpwardArrows:=TRUE;
      topOfs:=maxInt;
      drawBody(cnv);
      pleaseDontDrawUpwardArrows:=FALSE;
      topOfs:=skippedLines-1;
      end;
    if s = '' then eventFullyPainted:=y < bottomLimit
    else drawBody(cnv);
    inc(y);
    end;
  inc(evIdx);
  if firstEvent then P_topEventNrows:=Nrows-1;
  firstEvent:=FALSE;
  end;//while
P_bottomEvent:=evIdx-1;
P_lastEventIsFullyVisible:= eventFullyPainted and (evIdx=history.count);
setLength(items, Nitems);
updatePointedItem();
if assigned(onPainted) and (cnv=canvas) then onPainted(self);
end; // paintOn

procedure ThistoryBox.Paint();
begin paintOn(canvas) end;

function ThistoryBox.getSelText():string;
var
  SOS,EOS:ThistoryPos;
  i:integer;
  dim:integer;
  ev:Thevent;

  procedure addStr(s:string);
  begin
  while dim+length(s) > length(result) do
    setLength(result, length(result)+10000);
  system.move(s[1], result[dim+1], length(s));
  inc(dim, length(s));
  end; // addStr

begin
result:='';
dim:=0;

if startSel.ev=NIL then exit;

if minor(startSel,endSel) then
  begin
  SOS:=startsel;
  EOS:=endSel;
  end
else
  begin
  SOS:=endSel;
  EOS:=startsel;
  end;

if startSel.ofs<0 then
  begin
  for i:=SOS.evIdx to EOS.evIdx do
    begin
    ev:=history.getAt(i);
    addStr(ev.getHeaderText+CRLF+ev.getBodyText);
    if dim > 0 then
      if result[dim] = #10 then
        addStr(CRLF)
      else
        addStr(CRLF+CRLF);
    end;
  setLength(result, dim);
  exit;
  end;

if SOS.evIdx = EOS.evIdx then
  addStr( copy(history.getAt(SOS.evIdx).getBodyText, SOS.ofs+1, EOS.ofs-SOS.ofs))
else
  begin
  addStr( copy(history.getAt(SOS.evIdx).getBodyText, SOS.ofs+1, 99999)+CRLF);
  i:=SOS.evIdx+1;
  while i < EOS.evIdx do
    begin
    addStr(history.getAt(i).getBodyText+CRLF);
    inc(i);
    end;
  addstr(copy(history.getAt(EOS.evIdx).getBodyText, 1, EOS.ofs));
  end;

setLength(result, dim);
end; // getSelText

function applyHtmlFont(fnt:Tfont; s:string):string;
var
  h,q:string;
begin
h:='<font size=2 face="'+fnt.name+'" color=#'+color2str(fnt.color)+'>';
q:='</font>';
if fsItalic in fnt.style then
  begin
  h:=h+'<i>';
  q:='</i>'+q;
  end;
if fsBold in fnt.style then
  begin
  h:=h+'<b>';
  q:='</b>'+q;
  end;
result:=h+s+q;
end; // applyHtmlFont

function ThistoryBox.getSelHtml(smiles:boolean):string;
var
  SOS,EOS:ThistoryPos;
  i, dim:integer;
  ev:Thevent;

  procedure addStr(s:string);
  begin
  while dim+length(s) > length(result) do
    setLength(result, length(result)+10000);
  system.move(s[1], result[dim+1], length(s));
  inc(dim, length(s));
  end; // addStr

begin
result:='';
dim:=0;

if startSel.ev=NIL then exit;
if minor(startSel,endSel) then
  begin
  SOS:=startsel;
  EOS:=endSel;
  end
else
  begin
  SOS:=endSel;
  EOS:=startsel;
  end;
addStr('<html><head></head><body bgcolor=#'+color2str(theme.history.bgcolor)+'>');
for i:=SOS.evIdx to EOS.evIdx do
  begin
  ev:=history.getAt(i);
  addStr( CRLF+
    applyHtmlFont(ev.font,
      '<u>['+eventName(ev.kind)+'] '+datetimeToStr(ev.when)+', '
      +ev.who.displayed+'</u>'+'<br>'+str2html(ev.getBodyText)+'<br><br>' )
  );
  end;
addStr(CRLF+'</body></html>');

setLength(result, dim);
end; // getSelHtml

procedure ThistoryBox.mouseDown(Button: TMouseButton; Shift: TShiftState; X,Y:Integer);

  function doubleClick:boolean;
  begin
  result:=just2clicked;
  if just2clicked and not justTriggeredAlink and equal(lastClickedItem,pointedItem) then
    viewHeventWindow(history.getAt(endSel.evIdx));
  end; // doubleClick

begin
dontTriggerLink:=FALSE;
just2clicked:=now-lastTimeClick < dblClickTime;
if shift=[ssRight] then
  begin
  updatePointedItem();
  chatFrm.poppedup:=point(x,y);
  chatFrm.delete1.enabled:= wholeEventsAreSelected;
  chatFrm.saveas1.enabled:= somethingIsSelected;
  chatFrm.copytoclipboard1.visible:= somethingIsSelected;
  chatFrm.copylinktoclipboard1.visible:=pointedItem.kind=PK_LINK;
  chatFrm.addlinktofavorites1.visible:=(pointedItem.kind=PK_LINK)
    and (pointedItem.link.kind in [LK_WWW, LK_FTP]);
  chatFrm.addtoroaster1.visible:=(pointedItem.kind=PK_LINK)
    and (pointedItem.link.kind=LK_UIN);
  if chatFrm.addtoroaster1.visible then
    try
      chatFrm.selectedUIN:=strToInt(pointedItem.link.str);
      addGroupsToMenu(self, chatFrm.addtoroaster1, chatFrm.addcontactAction);
    except
      chatFrm.addtoroaster1.visible:=FALSE;
    end;
  chatFrm.viewmessageinwindow1.enabled:= chatFrm.thisChat.historyNowCount>0;
  chatFrm.selectall1.enabled:= chatFrm.thisChat.historyNowCount>0;
  with clientToScreen(point(x,y)) do
    chatFrm.histMenu.popup(x,y);

  exit;
  end;
if ssShift in shift then
  begin
  if pointedSpace.kind<>PK_NONE then
    begin
    endSel:=historyitem2pos(pointedSpace);
    repaint();
    end;
  inherited;
  exit;
  end;
deselect();
case pointedSpace.kind of
  PK_NONE: selecting:=TRUE;
  PK_CRYPTED:
    if enterPwdDlg(histcrypt.pwd) then
      histcrypt.pwdkey:=calculate_KEY1(histcrypt.pwd);
  PK_HEAD, PK_TEXT, PK_LINK, PK_SMILE:
    if ((pointedSpace.kind<>PK_HEAD) or (pointedItem.kind = PK_HEAD))
    and not DoubleClick then
      begin
      selecting:=TRUE;
      startsel:=historyitem2pos(pointedSpace);
      endSel:=startsel;
      end;
  PK_ARROWS_UP:
    begin
    chatFrm.histScrollLine(-1);
    exit; // prevent redundant repaint
    end;
  PK_ARROWS_DN:
    begin
    chatFrm.histScrollLine(+1);
    exit; // prevent redundant repaint
    end;
  else
    exit;
  end;
lastTimeClick:=now;
repaint;
if pointedItem.kind = PK_NONE then lastClickedItem:=pointedSpace
else lastClickedItem:=pointedItem;
updatePointedItem();
inherited;
end; // mouseDown

procedure ThistoryBox.click();
begin
if not dontTriggerLink
and (pointedItem.kind=lastClickedItem.kind)
and (pointedItem.link.id=lastClickedItem.link.id) then
  triggerLink(pointedItem)
end;

procedure ThistoryBox.mouseUp(Button: TMouseButton; Shift: TShiftState; X,Y:Integer);
var
  s:string;
begin
selecting:=FALSE;
if somethingIsSelected then
  begin
  if autocopyHist then
    begin
    s:=getSelText;
    if s > '' then
      clipboard.asText:=s
    end
  end;
if autoDeselect then
  begin
  deselect();
  repaint;
  end;
inherited;
end; // mouseUp

function ThistoryBox.spaceAt(pt:Tpoint):ThistoryItem;
var
  i,j,m,l:integer;
  r:Trect;
begin
result.kind:=PK_NONE;
i:=0;
l:=length(items)-1;
if l=0 then exit;
j:=l;
m:=-1;
// search for an item on the same row
while i<=j do
  begin
  m:=(i+j) div 2;
  r:=items[m].r;
  if within(r.Top, pt.Y, r.bottom) then
    break;
  if (pt.y < r.top) or (pt.y < r.bottom) and (pt.x < r.left) then j:=m-1
  else i:=m+1;
  end;
// if no item is matching the Y, move backward until the item is behind PT
while m > 0 do
  begin
  if pt.y < items[m-1].r.top then dec(m)
  else break
  end;
// without leaving this row, move backward searching for a better matching item
while m>0 do
  begin
  r:=items[m-1].r;
  if within(r.Top, pt.Y, r.bottom) and (pt.X <= r.right) then dec(m)
  else break
  end;
// same, but move forward
while m<l do
  begin
  r:=items[m+1].r;
  if within(r.Top, pt.Y, r.bottom) and (pt.X >= r.left) then inc(m)
  else break
  end;
// do we have a valid item?
if m>=0 then
  begin
  result:=items[m];
  // if pt is on the first half of the item then move behind 
  if pt.X < centerPoint(result.r).x then dec(result.ofs);
  end;
end; // spaceAt

function ThistoryBox.itemAt(pt:Tpoint):ThistoryItem;
var
  i,j,m:integer;
  r:Trect;
begin
result.kind:=PK_NONE;
i:=0;
j:=length(items)-1;
while i<=j do
  begin
  m:=(i+j) div 2;
  r:=items[m].r;
  if ptInRect( r, pt ) then
    begin
    result:=items[m];
    break;
    end;
  if (pt.y < r.top) or (pt.y < r.bottom) and (pt.x < r.left) then j:=m-1
  else i:=m+1;
  end;
end; // itemAt

procedure ThistoryBox.mouseMove(Shift: TShiftState; X, Y: Integer);
begin updatePointedItem() end; // mouseMove

procedure ThistoryBox.go2end();
var
  bmp:Tbitmap;
begin
bmp:=createBitmap(width, height);
topVisible:=history.count-1;
  repeat
  dec(topVisible);
  paintOn(bmp.canvas);
  until (topVisible < offset) or not lastEventIsFullyVisible;
inc(topVisible);
repaint();
bmp.free;
end; // go2end

function ThistoryBox.offsetPos():integer;
begin result:=topVisible-offset end;

function ThistoryBox.wholeEventsAreSelected():boolean;
begin result:=(startSel.ev<>NIL) and (startSel.ofs<0) end;

function ThistoryBox.nothingIsSelected():boolean;
begin result:=startSel.ev=NIL end;

function ThistoryBox.somethingIsSelected():boolean;
begin result:=startSel.ev<>NIL end;

function ThistoryBox.partialTextIsSelected():boolean;
begin result:=(startSel.ev<>NIL) and (startSel.ofs>=0) end;

procedure ThistoryBox.select(from, to_:integer);
begin
startSel.ofs:=-1;
startSel.evIdx:=from;
startSel.ev:=history.getAt(from);
endSel.ofs:=-1;
endSel.evIdx:=history.count-1;
endSel.ev:=history.getAt(to_);
end; // select

procedure ThistoryBox.deselect();
begin
startSel.ev:=NIL;
startSel.evIdx:=-1;
end; // deselect

procedure ThistoryBox.updatePointedItem();
var
  p:ThistoryPos;
  oldIt, oldSp:ThistoryItem;
  oldLink:ThistoryLink;
  pt:Tpoint;
begin
pt:=screenToClient(mousePos);
oldIt:=P_pointedItem;
oldSp:=P_pointedSpace;
oldLink:=linkToUnderline;
P_pointedItem:=itemAt(pt);
P_pointedSpace:=spaceAt(pt);
// no interesting movement
if equal(oldIt, P_pointedItem) and equal(oldSp, P_pointedSpace) then exit;
// link underlining, mouse cursor shape
if isLink(pointedItem) and (pointedItem.link.kind in linksToUnderline) then
  begin
  linkToUnderline:=pointedItem.link;
  cursor:=crHandPoint
  end
else
  begin
  linkToUnderline.evIdx:=-1;
  cursor:=crDefault;
  end;
// repaint necessary?
if not equal(linkToUnderline, oldLink) then
  begin
  avoidErase:=TRUE;
  paint();
  avoidErase:=FALSE;
  end;
// here the selecting management section begins
if not selecting then exit;
// selecting, no link has to be triggered
dontTriggerLink:=TRUE;
// updating the selection end point
if pointedSpace.kind=PK_NONE then exit;
p:=historyitem2pos(pointedSpace);
if minor(startSel,p) then inc(p.ofs, pointedSpace.l-1);
if equal(endSel,p) then exit; // no change?
endSel:=p;
// some adjustment could be needed
if nothingIsSelected() then startSel:=historyitem2pos(pointedItem)
else
  if startSel.ofs < 0 then endSel.ofs:=-1
  else if endSel.ofs < 0 then endSel.ofs:=0;
repaint();
end; // updatePointedItem

function ThistoryBox.triggerLink(item:ThistoryItem):boolean;
var
  s:string;
begin
result:=FALSE;
if item.kind<>PK_LINK then exit;
s:=item.link.str;
case item.link.kind of
  LK_WWW:
    begin
    if not Imatches(s,1,'http://') and not Imatches(s,1,'https://') then s:='http://'+s;
    openURL(s);
    end;
  LK_FTP:
    begin
    if not Imatches(s,1,'ftp://') then s:='ftp://'+s;
    openURL(s);
    end;
  LK_EMAIL:
    begin
    if not Imatches(s,1,'mailto:') then s:='mailto:'+s;
    exec(s);
    end;
  end;
end; // triggerLink

end.


