{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit langLib;

interface

uses
  sysUtils, strUtils, forms, controls, stdctrls, menus, classes, ExtCtrls,
  Buttons, ComCtrls, CheckLst, ColorPickerButton;

function getTranslation(const key:string; const args:array of const):string; overload;

procedure resetLanguage;
procedure loadLanguage(l:string);
procedure translateWindows;
procedure translateWindow(w:Tform);

implementation

uses
  utilLib, globalLib, chatDlg, mainDlg;
var
  lang:array of record
    key,text:string;
    end;
  alreadyLoaded:array of string;  // keep track of loaded modules

procedure resetLanguage;
begin
setlength(alreadyLoaded, 0);
setlength(lang, 0)
end; // resetLanguage

function idxOf(const key:string):integer;
var
  l,u,c:integer;
begin
result:=-1;
if length(lang)=0 then exit;
l:=0;
u:=length(lang)-1;
  repeat
  result:=(l+u) div 2;
  c:=compareStr(lang[result].key, key);
  if c=0 then
    exit
  else
    if c > 0 then
      u:=result-1
    else
      l:=result+1;
  until l > u;
result:=-1;
end; // idxOf

procedure add(k,v:string);
var
  i:integer;
begin
i:=idxOf(k);
if i>=0 then
  begin
  lang[i].text:=v;
  exit;
  end;
setlength(lang, length(lang)+1);
i:=length(lang)-2;
while (i>=0) and (k < lang[i].key) do
  begin
  lang[i+1]:=lang[i];
  dec(i);
  end;
lang[i+1].key:=k;
lang[i+1].text:=v;
end; // add

procedure loadLanguageFile(fn:string);
var
  k,v:string;
  f:text;
  i:integer;
begin
if findInStrings(fn, alreadyLoaded) >= 0 then exit;
setlength(alreadyLoaded, length(alreadyLoaded)+1);
alreadyLoaded[length(alreadyLoaded)-1]:=fn;
IOresult;
assignfile(f, fn);
reset(f);
while not eof(f) do
  begin
  readln(f,k);
  if k='' then continue;
  if AnsiStartsStr('___',k) then
    begin
    delete(k,1,3);
    i:=pos(';',k);
    if i>0 then delete(k,i,length(k));
    k:=trim(k);
    readln(f,v);
    add(k,v);
    end;
  if k[1]='\' then
    begin
    delete(k,1,1);
    v:=chop('\',k);
    if v='extends' then
      loadLanguage(chop('\',k));
    if v='include' then
      loadLanguageFile(ExtractFilePath(fn)+chop('\',k));
    end;
  end;
closefile(f);
end; // loadLanguageFile

procedure loadLanguage(l:string);
begin
loggaEvt('loading language: '+l);
loadLanguageFile(myPath+langsPath+l+'\'+langFilename);
loggaEvt('language loaded');
end;

function getTranslation(const key:string; const args:array of const):string;
var
  i:integer;
begin
result:='___'+key;
i:=idxOf(key);
if i<0 then exit;
result:=format(lang[i].text, args);
result:=ansiReplaceStr(result,'\n',#13);
//result:=ansiReplaceStr(result,'\s',' ');
end; // getTranslation

function trans(s:string):string;
begin
if AnsiStartsStr('___',s) then
  result:=getTranslation(copy(s,4,9999),[])
else
  result:=s;
end; // trans

procedure translateComponent(c:Tcomponent; window:Tform);

  procedure recurMenu(it:Tmenuitem);
  var
    i:integer;
  begin
  it.caption:=trans(it.caption);
  it.hint:=trans(it.hint);
  with it do
    for i:=0 to count-1 do
      recurMenu(items[i]);
  end; // recurMenu

  procedure recurTree(t:Ttreenode); overload;
  var
    i:integer;
  begin
  t.text:=trans(t.text);
  for i:=0 to t.count-1 do
    recurTree(t.item[i]);
  end; // recurTree

  procedure recurTree(t:Ttreenodes); overload;
  var
    i:integer;
  begin
  for i:=0 to t.count-1 do
    recurTree(t.item[i]);
  end; // recurTree

  procedure tstrings_trans(s:Tstrings);
  var
    i:integer;
  begin
  for i:=0 to s.count-1 do
    s[i]:=trans(s[i]);
  end; // tstrings_trans

var
  i:integer;
begin
if c is Tcustomform then with c as Tcustomform do
  caption:=trans(caption);
if c is Tmenu then with c as Tmenu do
  recurMenu(items);
if c is Tcontrol then with c as Tcontrol do
  hint:=trans(hint);
if c is Tlabel then with c as Tlabel do
  caption:=trans(caption);
if c is Ttabsheet then with c as Ttabsheet do
  caption:=trans(caption);
if c is Tlabelededit then with c as Tlabelededit do with editlabel do
  caption:=trans(caption);
if c is Tgroupbox then with c as Tgroupbox do
  caption:=trans(caption);
if c is Tbitbtn then with c as Tbitbtn do
  caption:=trans(caption);
if c is Tpanel then with c as Tpanel do
  caption:=trans(caption);
if c is Ttreeview then with c as Ttreeview do
  recurTree(items);
if c is Tbutton then with c as Tbutton do
  caption:=trans(caption);
if c is Ttoolbutton then with c as Ttoolbutton do
  caption:=trans(caption);
if c is Tspeedbutton then with c as Tspeedbutton do
  caption:=trans(caption);
if c is Tradiogroup then with c as Tradiogroup do
  begin
  caption:=trans(caption);
  tstrings_trans(items);
  end;
if c is TcomboBox then with c as TcomboBox do
  begin  // itemindex is lost during translation
  i:=itemIndex;
  tstrings_trans(items);
  itemIndex:=i;
  end;
if c is TColorPickerButton then with c as TColorPickerButton do
  begin
  caption:=trans(caption);
  customText:=trans(customText);
  end;
if c is Tcheckbox then with c as Tcheckbox do
  if caption > '' then
    begin
    caption:=trans(caption);
    width:=22+txtSize(window.canvas.handle, caption).cx;
    end;
if c is Tradiobutton then with c as Tradiobutton do
  if caption > '' then
    begin
    caption:=trans(caption);
    width:=22+txtSize(window.canvas.handle, caption).cx;
    end;
if c is Tchecklistbox then with c as Tchecklistbox do
  tstrings_trans(items);
if c is Tlistbox then with c as Tlistbox do
  tstrings_trans(items);

for i:=c.componentCount-1 downto 0 do
  translateComponent(c.components[i], window);
end; // translateComponent

procedure translateWindow(w:Tform);
begin translateComponent(w,w) end;

procedure translateWindows;
var
  i:integer;
begin
chatFrm.Caption:=mainFrm.Caption+' - '+___('chat title');
i:=0;
while i < screen.formCount do
  begin
  translateWindow(screen.forms[i]);
  inc(i);
  end;
end; // translateWindows

end.
