{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit outboxDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, ComCtrls, outboxLib, contacts,
  Menus, JvMenus;

type
  ToutboxFrm = class(TForm)
    list: TListBox;
    memo: TMemo;
    Splitter: TSplitter;
    deleteBtn: TSpeedButton;
    saveBtn: TSpeedButton;
    groupbox: TGroupBox;
    charsLbl: TLabel;
    infoLbl: TLabel;
    processChk: TCheckBox;
    menu: TJvPopupMenu;
    Viewinfo1: TMenuItem;
    Sendmsg1: TMenuItem;
    procedure FormShow(Sender: TObject);
    procedure listDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
    procedure listClick(Sender: TObject);
    procedure deleteBtnClick(Sender: TObject);
    procedure SplitterMoved(Sender: TObject);
    procedure memoChange(Sender: TObject);
    procedure saveBtnClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Sendmsg1Click(Sender: TObject);
    procedure Viewinfo1Click(Sender: TObject);
    procedure listMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure listDblClick(Sender: TObject);
    procedure menuPopup(Sender: TObject);
  public
    lastSelected:TOevent;
    function  selectedContact:Tcontact;
    procedure updateList;
    procedure updateMemo;
    procedure updateChars;
    function  validIdx(i:integer):boolean;
    procedure open(c:Tcontact=NIL);
  public procedure destroyHandle;
  end;

var
  outboxFrm: ToutboxFrm;

implementation

{$R *.DFM}

uses
  globalLib, utilLib, ICQv7, chatDlg, strutils;

function ToutboxFrm.validIdx(i:integer):boolean;
begin result:=(i>=0) and (i<list.items.count) end;

procedure ToutboxFrm.updateList;
var
  i:integer;
begin
list.clear;
for i:=0 to outbox.count-1 do
  list.items.addObject('', outbox.list[i]);
updatememo;
end; // updateList

procedure ToutboxFrm.updatememo;
var
  s:string;
begin
memo.readonly:=TRUE;
if not validIdx(list.itemIndex) then
  begin
  infoLbl.caption:='';
  memo.clear;
  updateChars;
  saveBtn.enabled:=FALSE;
  exit;
  end;
lastSelected:=list.Items.Objects[list.itemIndex] as TOevent;
with lastSelected do
  begin
  memo.text:=info;
  saveBtn.enabled:=FALSE;
  case kind of
    OE_msg:
      begin
      s:=s+___('msg');
      memo.readonly:=FALSE;
      end;
    OE_contacts: s:=s+___('contacts');
    OE_addedYou: s:=s+___('addedyou');
    OE_auth: s:=s+___('auth');
    OE_authDenied: s:=s+___('authdenied');
    else s:='';
    end;
  infoLbl.caption:=___('outbox info',[
    ifThen(flags and IF_multiple >0, '('+___('multisend')+') '),
    s,
    contactsDB.get(uin).displayed,
    datetimeToStr(wrote),
    ifThen(lastmodify>0, datetimeToStr(lastmodify), datetimeToStr(wrote))
  ]);
  end;

updateChars;
end;

procedure ToutboxFrm.updateChars;
begin charsLbl.caption:=___('chars')+' '+intToStr(length(memo.text)) end;

procedure ToutboxFrm.FormShow(Sender: TObject);
begin
icon:=pic2ico(PIC_OUTBOX);
deleteBtn.glyph.assign(getPic(PIC_DELETE));
saveBtn.glyph.assign(getPic(PIC_SAVE));
applyTaskButton(self);
end;

procedure ToutboxFrm.listDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  oe:TOevent;
  bmp:integer;
  cnv:Tcanvas;
  x,y:integer;
  c:Tcontact;
  msg:string;
  oldClr:Tcolor;

  procedure outText(s:string);
  begin
  cnv.textOut(x,y,s);
  inc(x, cnv.textWidth(s));
  end;

begin
cnv:=list.canvas;
cnv.fillRect(rect);

oe:=list.Items.Objects[index] as TOevent;
x:=rect.Left+2;
y:=rect.top;
cnv.font.size:=-10;
if oe.kind <> OE_email then
  c:=contactsDB.get(oe.uin)
else
  c:=NIL;

case oe.kind of
  OE_msg: bmp:=PIC_MSG;
  OE_contacts: bmp:=PIC_CONTACTS;
  OE_addedYou: bmp:=PIC_ADD_CONTACT;
  OE_auth:
    begin
    bmp:=-1;
    msg:=___('yes');
    end;
  OE_authDenied:
    begin
    bmp:=-1;
    msg:=___('no');
    end;
  else exit;
  end;
if bmp>0 then
  inc(x, drawPic(cnv,x,y,bmp).cx+2)
else
  begin
  cnv.Font.style:=[fsBold];
  if odSelected	in state then
    outText(msg)
  else
    begin
    oldClr:=cnv.brush.color;
    cnv.brush.color:=clBtnFace;
    cnv.Pen.color:=oldClr;
    cnv.Ellipse(x,y,x+cnv.textWidth(msg),y+cnv.textHeight(msg));
    cnv.brush.color:=oldClr;
    SetBkMode(cnv.handle, TRANSPARENT);
    outText(msg);
    SetBkMode(cnv.handle, OPAQUE);
    end;
  cnv.Font.style:=[];
  inc(x,4);
  end;
if oe.kind=OE_email then
  outText(oe.email)
else
  outText(c.displayed);
end; // list drawitem

procedure ToutboxFrm.listClick(Sender: TObject);
begin
updatememo;
deleteBtn.enabled:= list.itemIndex>=0;
saveBtn.enabled:=FALSE;
end;

procedure ToutboxFrm.deleteBtnClick(Sender: TObject);
var
  ev:TOevent;
  i:integer;
begin
for i:=0 to list.items.count-1 do
  if list.Selected[i] then
    begin
    ev:=list.Items.Objects[i] as TOevent;
    outbox.remove(ev);
    ev.free;
    end;
updateList;
end;

procedure ToutboxFrm.SplitterMoved(Sender: TObject);
begin
list.height:=splitter.top;
memo.height:=clientHeight-splitter.height-list.height
end;

procedure ToutboxFrm.memoChange(Sender: TObject);
begin
saveBtn.enabled:=TRUE;
updateChars;
end;

procedure ToutboxFrm.saveBtnClick(Sender: TObject);
begin
with list.Items.Objects[list.itemIndex] as TOevent do
  begin
  info:=memo.text;
  lastmodify:=now;
  end;
saveBtn.enabled:=FALSE;
end;

procedure ToutboxFrm.FormResize(Sender: TObject);
var
  i:integer;
begin
if memo.boundsrect.Bottom > clientHeight then
  begin
  i:=clientHeight-splitter.boundsrect.bottom;
  if i > 10 then
    memo.height:=i;
  end;
end;

procedure ToutboxFrm.open(c:Tcontact=NIL);
var
  i:integer;
begin
lastSelected:=NIL;
updateList;
if c=NIL then
  i:=list.items.indexOfObject(lastSelected)
else
  begin
  i:=list.items.count-1;
  while (i > 0) and ((list.Items.Objects[i] as TOevent).uin <> c.uin) do
    dec(i);
  end;
if not validIdx(i) then
  if list.items.count > 0 then
    i:=0
  else
    i:=-1;
if i>=0 then
  begin
  list.itemIndex:=i;
  list.selected[i]:=TRUE;
  end;
updateMemo;
deleteBtn.enabled:= validIdx(i);
show;
end; // open

procedure ToutboxFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin destroyHandle end;

function ToutboxFrm.selectedContact:Tcontact;
begin
result:=NIL;
if (lastSelected<>NIL) then
  with lastSelected do
    if kind <> OE_email then
      result:=contactsDB.get(uin);
end; // selectedContact

procedure ToutboxFrm.Sendmsg1Click(Sender: TObject);
begin
if lastSelected = NIL then exit;
outbox.remove(lastSelected);
processOevent(lastSelected);
lastSelected.Free;
updateList;
end;

procedure ToutboxFrm.Viewinfo1Click(Sender: TObject);
begin if selectedContact<>NIL then viewinfoAbout(selectedContact) end;

procedure ToutboxFrm.listMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
if button=mbRight then
  if selectedContact<>NIL then
    with mousePos do
      menu.popup(x,y);
end;

procedure ToutboxFrm.listDblClick(Sender: TObject);
begin if selectedContact<>NIL then chatFrm.openOn(selectedContact) end;

procedure ToutboxFrm.destroyHandle; begin inherited end;

procedure ToutboxFrm.menuPopup(Sender: TObject);
begin
Sendmsg1.Enabled:=ICQ.isOnline;
end;

end.
