{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit outboxLib;

interface

uses
  classes, contacts, sysutils;

const
  OE_msg        = 1;
  OE_contacts   = 2;
  OE_addedYou   = 3;
  OE_auth       = 4;
  OE_authDenied = 5;
  OE_file       = 6;
  OE_email      = 7;
  OE_automsgreq = 8;

type
  TOEvent=class
    kind:integer;
    flags:integer;
    uin:integer;
    email:string;
    info:string;
    cl:Tcontactlist;
    wrote,lastmodify:Tdatetime;
    // ack fields
    ID:integer;
    timeSent:TdateTime;
    filepos:integer;
    destructor destroy; override;
    function toString:string;
    procedure fromString(s:string);
    end; // TOEvent

  Toutbox=class(Tlist)
    destructor destroy; override;

    function toString:string;
    procedure fromString(s:string);

    function empty:boolean;
    function pop:TOevent;
    function popVisible:TOevent;

    procedure clear; override;
    function add(kind,dest:integer; flags:integer=0; info:string=''):Toevent; overload;
    function add(kind,dest,flags:integer; cl:Tcontactlist):Toevent; overload;
    function getAt(idx:integer):TOevent;
    function remove(ev:TOevent):boolean; overload;
    function stFor(who:Tcontact):boolean;
    function findID(id:Integer):integer;

    procedure updateScreenFor(uin:integer);
    end; // Toutbox

implementation

uses
  globalLib, mainDlg, utilLib, chatDlg, icqv7;

destructor Toutbox.destroy;
begin
clear;
inherited
end; // destroy

function Toutbox.toString:string;
var
  s:string;
  i:integer;
begin
result:='';  // file version
for i:=0 to count-1 do
  begin
  s:=getAt(i).toString;
  result:=result+int2str(length(s))+s;
  end;
critt(result, ICQ.myinfo.uin);
result:='VER'+int2str(1)+result;
end; // toString

procedure Toutbox.fromString(s:string);
var
  i,l:integer;
  ev:Toevent;
begin
s:=decritted(copy(s,8,length(s)), ICQ.myinfo.uin);
clear;
i:=1;
while i < length(s) do
  begin
  l:=integer((@s[i])^);
  inc(i,4);
  ev:=TOevent.create;
  ev.fromString( copy(s,i,l) );
  inc(i,l);
  add(ev);
  end;
end; // fromString

procedure Toutbox.clear;
var
  i:integer;
begin
for i:=0 to count-1 do
  with getAt(i) do
    begin
    updateScreenFor(uin);
    free;
    end;
inherited;
saveOutboxDelayed:=TRUE;
end; // clear

function Toutbox.add(kind,dest,flags:integer; cl:Tcontactlist):TOevent;
begin
result:=add(kind,dest,flags);
result.cl:=Tcontactlist.create;
result.cl.assign(cl);
end; // add

function Toutbox.add(kind,dest:integer; flags:integer=0; info:string=''):TOevent;
var
  found:boolean;
  i:integer;
begin
result:=NIL;
found:=FALSE;
for i:=0 to count-1 do
  begin
  result:=getAt(i);
  if (kind in [OE_addedyou,OE_auth,OE_authDenied]) and (kind=result.kind) and (result.uin=dest) then
    begin
    found:=TRUE;
    break;
    end;
  end;
if not found then
  begin
  result:=TOevent.create;
  add(result);
  end;
result.kind:=kind;
result.flags:=flags;
result.uin:=dest;
result.info:=info;
result.wrote:=now;
result.lastmodify:=now;
updateScreenFor(result.uin);
saveOutboxDelayed:=TRUE;
end; // add

function Toutbox.getAt(idx:integer):TOevent;
begin
if (idx>=0) and (idx<count) then
  result:=list[idx]
else
  result:=NIL;
end; // getAt

function Toutbox.remove(ev:TOevent):boolean;
begin
result:=inherited remove(ev)>=0;
updateScreenFor(ev.uin);
saveOutboxDelayed:=TRUE;
end; // remove

function Toutbox.popVisible:TOevent;
var
  i:integer;
begin
i:=0;
while i < count do
  begin
  result:=getAt(i);
  if (result.flags and IF_sendWhenImVisible=0)
  or ICQ.imVisibleTo(contactsDB.get(result.uin)) then
    begin
    delete(i);
    saveOutboxDelayed:=TRUE;
    updateScreenFor(result.uin);
    exit;
    end;
  inc(i);
  end;
result:=NIL;
end; // popVisible

function Toutbox.pop:TOevent;
begin
result:=NIL;
if count>0 then
  begin
  result:=getAt(0);
  delete(0);
  updateScreenFor(result.uin);
  end;
saveOutboxDelayed:=TRUE;
end; // pop

function Toutbox.empty:boolean;
begin result:=count=0 end;

function Toutbox.stFor(who:Tcontact):boolean;
var
  i:integer;
begin
result:=FALSE;
for i:=0 to count-1 do
  if getAt(i).uin = who.uin then
    begin
    result:=TRUE;
    exit;
    end;
end; // stFor

procedure Toutbox.updateScreenFor(uin:integer);
begin
if uin = 0 then exit;
redrawUIN(uin);
if chatFrm<>NIL then
  with chatFrm do
    if (thischat<>NIL) and (thischat.who.uin=uin) then
      sbar.repaint;
mainFrm.sbar.Repaint;
end; // updateScreenFor

function Toutbox.findID(id:Integer):integer;
begin
for result:=count-1 downto 0 do
  if getAt(result).id = id then
    exit;
result:=-1;
end; // findID

////////////////////////////////////////////////////////////////////////

const
  OEK_kind    = 1;
  OEK_flags   = 2;
  OEK_email   = 3;
  OEK_uin     = 4;
  OEK_info    = 5;
  OEK_wrote   = 6;
  OEK_cl      = 7;

function TOevent.toString:string;

  procedure writeDown(code:integer; data:string);
  begin result:=result+int2str(length(data))+int2str(code)+data end;

begin
result:='';
writeDown(OEK_kind, int2str(kind));
writeDown(OEK_flags, int2str(flags));
if kind=OE_email then
  writeDown(OEK_email, email)
else
  writeDown(OEK_uin, int2str(uin));
writeDown(OEK_info, info);
writeDown(OEK_wrote, dt2str(wrote));
if assigned(cl) then
  writeDown(OEK_cl, cl.tostring);
end; // toString

procedure TOevent.fromString(s:string);
var
  i,L,code,next:integer;
begin
i:=1;
while i < length(s) do
  begin
  L:=integer((@s[i])^); inc(i,4);
  code:=integer((@s[i])^); inc(i,4);
  next:=i+L;
  case code of
    OEK_kind: kind:=integer((@s[i])^);
    OEK_flags: flags:=integer((@s[i])^);
    OEK_wrote: wrote:=Tdatetime((@s[i])^);
    OEK_info: info:=copy(s,i,L);
    OEK_email: email:=copy(s,i,L);
    OEK_uin: uin:=integer((@s[i])^);
    OEK_cl:
      begin
      if cl=NIL then cl:=Tcontactlist.create;
      cl.fromstring(copy(s,i,L), contactsDB);
      end;
    end;
  i:=next;
  end;
end; // fromString

destructor TOevent.destroy;
begin
cl.free;
inherited;
end;

end.
