{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit roasterLib;

interface

uses
  classes, icqv7, windows, contacts, stdctrls, chatDlg, dialogs, graphics, VirtualTrees;

type
  Tdivisor=(d_online,d_offline,d_contacts,d_nil);
const
  divsWithGroups=[d_online,d_contacts,d_offline];
  divisor2str:array [Tdivisor] of string=('online','offline','contacts','not in list');
  NODE_ROOT=0;
  NODE_DIV=1;
  NODE_GROUP=2;
  NODE_CONTACT=3;

type
  Tnode=class
    kind       :integer;
    contact    :Tcontact;
    divisor    :Tdivisor;
    groupId    :integer;
    textOfs    :integer;
    outboxRect :Trect;
    treenode   :PVirtualNode;
    order      :integer;
    constructor create(divisor_:Tdivisor); overload;
    constructor create(groupId_:integer; divisor_:Tdivisor);  overload;
    constructor create(contact_:Tcontact); overload;
    destructor destroy; override;
    procedure setExpanded(val:boolean; recur:boolean=FALSE);
    function  isVisible:boolean;
    function  parent:Tnode;
    function  childrenCount:integer;
    function  firstChild:Tnode;
    function  expanded:boolean;
    function  level:integer;
    function  rect:Trect;
    function  next:Tnode;
    function  prev:Tnode;
    end;

function  contactsUnder(n:Tnode):integer;
procedure updateHiddenNodes;
procedure popup(x:integer=-1; y:integer=-1);
procedure popupOn(n:Tnode; x:integer=-1; y:integer=-1);
function  compareNodes(Node1,Node2:Tnode):integer;
function  str2divisor(s:string):Tdivisor;
procedure rebuild;
procedure redraw; overload;
function  redraw(c:Tcontact):boolean; overload;
function  redraw(n:Tnode):boolean; overload;
function  update(c:Tcontact):boolean;
function  exists(c:Tcontact):boolean;
function  focus(c:Tcontact):boolean; overload;
function  focus(n:Tnode):boolean; overload;
function  focus(tn:Pvirtualnode):boolean; overload;
function  focusTemp(n:Tnode):boolean;
function  remove(c:Tcontact):boolean;
procedure sort(c:Tcontact); overload;
procedure sort(n:Tnode); overload;
procedure sort(tn:Pvirtualnode); overload;
procedure formresized;
function  onlineMaxY:integer;
function  fullMaxY:integer;
function  addGroup(name:string):integer;
function  removeGroup(id:integer):boolean;
procedure edit(n:Tnode);
function  focused:Tnode;
function  focusedContact:Tcontact;
procedure expand(n:Tnode);
procedure collapse(n:Tnode);
function  nodeAt(x,y:integer):Tnode;
procedure focusPrevious;
procedure clear;
procedure setOnlyOnline(v:boolean);
function  getNode(tn:Pvirtualnode):Tnode;
procedure setNewGroupFor(c:Tcontact; grp:integer);
function  isUnderDiv(n:Tnode):Tdivisor;

var
  building:boolean=FALSE;
  dragging:boolean=FALSE; // roaster.dragging doesn't work
  inplace:record
    edit    :Tedit;
    what    :integer;
    contact :Tcontact;
    groupId :integer;
    node    :Tnode;
    end;
  contactsPool :Tlist;
  expandedByTempFocus :Tnode;

implementation

uses
  mainDlg, sysutils, globalLib, utilLib, groupsLib;
var
  // declared globally to speedup the compare callback functions
  tmpT1,tmpT2:Tdatetime;
  tmpC1,tmpC2:Tcontact;
  tmpB1,tmpB2:boolean;
  tmpI1,tmpI2:integer;

  divs:array [Tdivisor] of Tnode;
  buildingOnline:boolean;

function getNode(tn:Pvirtualnode):Tnode;
begin
if tn=NIL then
  result:=NIL
else
  begin
  result:=mainfrm.roaster.getnodedata(tn);
  if result<>NIL then
    result:=Tnode(pointer(result)^);
  end;
end; // getNode

function str2divisor(s:string):Tdivisor;
begin
for result:=low(result) to high(result) do
  if s = divisor2str[result] then
    exit;
raise Exception.Create('str2divisor');
end; // str2divisor

procedure updateHiddenNode(n:Tnode);
begin
if (n=NIL) or (n.treenode=NIL) then exit;
mainfrm.roaster.isVisible[n.treenode]:=not showOnlyImVisibleTo or ICQ.imVisibleTo(n.contact);
end; // updateHiddenNode

procedure updateHiddenNodes;
var
  i:integer;
begin
if (divs[d_offline]<>NIL) then
  mainfrm.roaster.isVisible[divs[d_offline].treenode]:=not showOnlyOnline;
for i:=0 to contactsPool.count-1 do
  updateHiddenNode(contactsPool[i]);
end; // updateHiddenNodes

function compareContacts(c1,c2:Tcontact):integer;
begin
result:=0;
if (c1=c2) or (c1=NIL) or (c2=NIL) then exit;
tmpT1:=TCE(c1.data^).lastMsgTime;
tmpT2:=TCE(c2.data^).lastMsgTime;
case sortBy of
  SB_alpha: result:=compareText(c1.displayed, c2.displayed);
  SB_event:
    begin
    tmpB1:=TRUE;
    if (tmpT1 < startTime) and (tmpT2 < startTime) then
      if (c1.status=SC_OFFLINE) and (c2.status=SC_OFFLINE) then
        begin
        tmpT1:=TCE(c1.data^).lastEventTime;
        tmpT2:=TCE(c2.data^).lastEventTime;
        if (tmpT1 < startTime) and (tmpT2 < startTime) then
          tmpB1:=FALSE
        end
      else
        tmpB1:=FALSE;

    if not tmpB1 then
      result:=compareText(c1.displayed, c2.displayed)
    else
      if tmpT1 > tmpT2 then result:=-1
      else if tmpT1 < tmpT2 then result:=+1;
    end;
  else
    begin
    tmpB1:=ICQ.imVisibleTo(c1);
    tmpB2:=ICQ.imVisibleTo(c2);
    if tmpb1 and not tmpb2 then result:=-1;
    if tmpb2 and not tmpb1 then result:=+1;
    end;
  end;
end; // compareContacts

// this sort criteria is deeper than compareContacts one, sit has to think about groups and online status
function compareContacts4build(item1,item2:pointer):integer;
begin
result:=0;
if (item1=item2) or (item1=NIL) or (item2=NIL) then exit;
tmpC1:=Tcontact(item1);
tmpC2:=Tcontact(item2);

// different sections?
if buildingOnline then
  case tmpC1.status of
    SC_UNK:
      case tmpC2.status of
        SC_OFFLINE: result:=+1;
        SC_UNK: ;
        else result:=+1;
        end;
    SC_OFFLINE:
      case tmpC2.status of
        SC_OFFLINE: ;
        SC_UNK: result:=-1;
        else result:=+1;
        end;
    else
      case tmpC2.status of
        SC_OFFLINE: result:=-1;
        SC_UNK: result:=-1;
        end;
    end;
if result<>0 then
  exit;
// different groups?
if showGroups and (tmpC1.group <> tmpC2.group) then
  begin
  if tmpC1.group = 0 then
    result:=-1
  else
    if tmpC2.group = 0 then
      result:=+1
    else
      begin
      tmpI1:=groups.get(tmpC1.group).order;
      tmpI2:=groups.get(tmpC2.group).order;
      if tmpI1=tmpI2 then
        result:=compareText(groups.id2name(tmpC1.group), groups.id2name(tmpC2.group))
      else
        if tmpI1<tmpI2 then result:=-1
        else result:=+1;
      end;
  exit;
  end;
result:=compareContacts(tmpC1,tmpC2);
end; // compareContacts4build

function compareNodes(Node1,Node2:Tnode):integer;
begin
result:=0;
if (node1=NIL) or (node2=NIL) or (node1=node2) then exit;
if node1.level < node2.level then result:=-1;
if node1.level > node2.level then result:=+1;
if result<>0 then exit;
if (node1.kind=node2.kind) and (node1.order<>node2.order) then
  begin
  if node1.order<node2.order then result:=-1
  else result:=+1;
  exit;
  end;
case node1.kind of
  NODE_DIV:
    if node2.kind <> NODE_DIV then result:=-1
    else result:=compareInt(ord(node1.divisor),ord(node2.divisor));
  NODE_GROUP:
    case node2.kind of
      NODE_DIV: result:=+1;
      NODE_CONTACT: result:=+1;
      NODE_GROUP:
        result:=compareText( groups.id2name(node1.groupId),
          groups.id2name(node2.groupId) );
      end;
  NODE_CONTACT:
    case node2.kind of
      NODE_DIV: result:=+1;
      NODE_GROUP: result:=-1;
      NODE_CONTACT: result:=compareContacts(node1.contact,node2.contact);
      end;
  end;
end; // compareNodes

constructor Tnode.create(divisor_:Tdivisor);
begin
inherited create;
kind:=NODE_DIV;
divisor:=divisor_;
end; // create

constructor Tnode.create(groupId_:integer; divisor_:Tdivisor);
begin
inherited create;
kind:=NODE_GROUP;
groupId:=groupId_;
divisor:=divisor_;
end; // create

constructor Tnode.create(contact_:Tcontact);
begin
inherited create;
kind:=NODE_CONTACT;
contact:=contact_;
TCE(contact_.data^).node:=self;
contactsPool.add(self);
end; // create

destructor Tnode.destroy;
begin
case kind of
  NODE_CONTACT:
    begin
    contactsPool.remove(self);
    TCE(contact.data^).node:=NIL;
    end;
  NODE_GROUP: groups.a[groups.idxOf(groupId)].node[divisor]:=NIL;
  NODE_DIV: divs[divisor]:=NIL;
  end;
inherited;
end; // destroy

function Tnode.parent:Tnode;
begin
if treenode.parent=NIL then
  result:=NIL
else
  result:=getNode(treenode.parent)
end;

function Tnode.childrenCount:integer;
begin result:=treenode.childcount end;

function Tnode.firstChild:Tnode;
begin result:=getNode(treenode.FirstChild) end;

function Tnode.next:Tnode;
begin result:=getNode(treenode.NextSibling) end;

function Tnode.prev:Tnode;
begin result:=getNode(treenode.PrevSibling) end;

function Tnode.expanded:boolean;
begin result:=vsExpanded in treenode.states end;

function Tnode.level:integer;
begin result:=mainfrm.roaster.getnodelevel(treenode) end;

function Tnode.isVisible:boolean;
var
  n,root:Pvirtualnode;
begin
n:=treenode;
root:=mainfrm.roaster.rootnode;

result:=vsVisible in n.states;
  repeat
  n:=n.parent;
  if n<>root then
    result:=(vsVisible in n.states) and (vsExpanded in n.states); 
  until not result or (n=root);
end; // isVisible

procedure Tnode.setExpanded(val:boolean; recur:boolean=FALSE);
begin
if not recur then
  mainfrm.roaster.expanded[treenode]:=val
else
  if val then
    mainfrm.roaster.FullExpand(treenode)
  else
    mainfrm.roaster.FullCollapse(treenode);
end; // setexpanded

function Tnode.rect:Trect;
begin result:=mainfrm.roaster.GetDisplayRect(treenode, -1, FALSE) end;

/////////////////////////////////////////////////////////////

function insertNode(d:Tdivisor):Tnode; overload;
var
  a:Tnode;
begin
result:=divs[d];
if assigned(result) then
  exit;

a:=Tnode.create(d);
result:=a;
result.treenode:=mainfrm.roaster.addchild(NIL, result);
divs[d]:=result;
if not building then sort(result);
end; // insertNode

function insertNode(id:integer; d:Tdivisor):Tnode; overload;
var
  idx:integer;
begin
idx:=groups.idxOf(id);
result:=groups.a[idx].node[d];
if assigned(result) then exit;
insertNode(d); // ensure divisor existence
result:=Tnode.create(id, d);
result.order:=groups.a[idx].order;
result.treenode:=mainFrm.roaster.addChild(divs[d].treenode, result);
groups.a[idx].node[d]:=result;
if not building then sort(result);
end; // insertNode

function shouldBeUnder(c:Tcontact; d:Tdivisor):Tnode;
begin
result:=insertNode(d);
if (c.group = 0) or not showGroups or not (d in divsWithGroups) then exit;
if not groups.exists(c.group) then
  c.group:=0
else
  result:=insertNode(c.group, d);
end; // shouldBeUnder

function insertNode(c:Tcontact; under:Tnode):Tnode; overload;

  procedure checkExpansion;
  var
    n:Tnode;
    d:Tdivisor;
  begin
  n:=result.parent;
    repeat
    case n.kind of
      NODE_GROUP:
        begin
        d:=n.parent.divisor;
        n.setexpanded(groups.a[groups.idxOf(n.groupId)].expanded[d]);
        end;
      NODE_DIV:
        begin
        n.setexpanded(TRUE);
        break;
        end;
      end;
    n:=n.parent;
    until n=NIL;
  end; // checkExpansion

begin
result:=Tnode.create(c);

result.treenode:=mainFrm.roaster.addChild(under.treenode, result);
updateHiddenNode(result);
if not building then
  checkExpansion;
autosizeDelayed:=TRUE;
if not building then sort(result);
end; // insertNode

function insertNode(c:Tcontact; d:Tdivisor):Tnode; overload;
begin result:=insertNode(c, shouldBeUnder(c,d)) end;

function insertNode(c:Tcontact):Tnode; overload;
begin
if notinlist.exists(c) then
  result:=insertNode(c, d_nil)
else
  if not ICQ.readroaster.exists(c) then
    result:=NIL
  else
    if icq.isonline then
      if c.status = SC_OFFLINE then
        result:=insertNode(c, d_offline)
      else
        result:=insertNode(c, d_online)
    else
      result:=insertNode(c, d_contacts)
end; // insertNode

function removeNode(n:Tnode):boolean; overload;
var
  parent:Tnode;
begin
result:=assigned(n);
if not result then exit;

while n.childrenCount > 0 do
  removeNode(n.firstChild);

parent:=n.parent;
mainfrm.roaster.deleteNode(n.treenode);
n.free;

if parent=NIL then exit;
if parent.childrenCount=0 then
  case parent.kind of
    NODE_GROUP: removeNode(parent);
    NODE_DIV: if parent.divisor in [d_nil,d_contacts] then removeNode(parent);
    end;
autosizeDelayed:=TRUE;
end; // removeNode

function removeNode(c:Tcontact):boolean; overload;
begin result:=removeNode(TCE(c.data^).node) end;

procedure rebuild;
var
  i:integer;
  roaster:TcontactList;
  d:Tdivisor;
  c:Tcontact;
  oldFocusedContact:Tcontact;
begin
oldFocusedContact:=focusedContact;
try
building:=TRUE;
//with mainfrm.roaster.treeoptions do autooptions:=autoOptions-[toAutosort];
// reset
clear;

fillChar(divs,sizeOf(divs),0);
if groups=NIL then exit;
for i:=0 to groups.count-1 do
  with groups.a[i] do
    fillchar(node, sizeOf(node), 0);
// roaster section
roaster:=icq.readroaster.clone;
roaster.sort(compareContacts4build);

buildingOnline:=icq.isOnline;
if buildingOnline then
  begin
  insertNode(d_online);
  insertNode(d_offline);
  end;
roaster.resetEnumeration;
while roaster.hasMore do
  begin
  c:=roaster.getNext;
  if not buildingOnline then
    insertNode(c, d_contacts)
  else
    if c.status = SC_OFFLINE then
      insertNode(c, d_offline)
    else
      insertNode(c, d_online)
  end;

roaster.free;
// NIL section
if assigned(notInList) then
  with notInList do
    begin
    resetEnumeration;
    while hasMore do
      insertNode(getNext, d_nil);
    end;

// expands all divs
for d:=low(d) to high(d) do
  if assigned(divs[d]) then
    begin
    divs[d].setExpanded(TRUE);
    for i:=0 to groups.count-1 do
      with groups.a[i] do
        if node[d]<>NIL then
          node[d].setexpanded(expanded[d])
    end;

updateHiddenNodes;
building:=FALSE;
//with mainfrm.roaster.treeoptions do autooptions:=autoOptions+[toAutosort];
autosizeDelayed:=TRUE;
finally
//sort(mainfrm.roaster.rootnode);
with mainfrm.roaster do
  begin
  show;
  if mainfrm.visible then
    setFocus;
  if oldFocusedContact<>NIL then
    focus(oldFocusedContact)
  else
    if totalcount > 0 then
      begin
      focusedNode:=RootNode.firstchild;
      topnode:=RootNode.firstchild;
      end;
  end;
end;
end; // rebuild

procedure redraw;
begin mainfrm.roaster.repaint end;

function redraw(c:Tcontact):boolean;
begin
result:=FALSE;
if c=NIL then exit;
result:=redraw(TCE(c.data^).node);
updateHiddenNode(TCE(c.data^).node);
if chatFrm<>NIL then
  chatFrm.redrawTab(c);
end; // redraw

function redraw(n:Tnode):boolean;
var
  r:Trect;
begin
result:=FALSE;
if (n=NIL) or (n.treenode=NIL) then exit;
R := mainfrm.roaster.GetDisplayRect(n.treenode, -1, FALSE);
InvalidateRect(mainfrm.roaster.handle, @R, True);
end; // redraw

function update(c:Tcontact):boolean;
var
  d:Tdivisor;
  wasFocused:boolean;
  n:Tnode;
begin
result:=FALSE;
if c=NIL then exit;
wasFocused:=c=focusedContact;
removeNode(c);
n:=insertNode(c);
if n=NIL then exit;
d:=isUnderDiv(n);
if assigned(divs[d]) then
  divs[d].setExpanded(TRUE);
sort(c);
if wasFocused then
  focus(c);
end; // update

function remove(c:Tcontact):boolean;
begin
result:=ICQ.removeContact(c) or notInList.remove(c);
visibleList.remove(c);
saveListsDelayed:=TRUE;
removeNode(c);
if chatFrm<>NIL then chatFrm.userChanged(c);
end; // remove

function exists(c:Tcontact):boolean;
begin result:=(c<>NIL) and (c.data<>NIL) and (TCE(c.data^).node<>NIL) end;

function focus(tn:Pvirtualnode):boolean; overload;
begin result:=focus(getnode(tn)) end;

function focus(n:Tnode):boolean; overload;
begin
result:=assigned(n) and assigned(n.treenode);
if result then
  mainfrm.roaster.focusedNode:=n.treenode
else
  mainfrm.roaster.focusednode:=NIL;
clickedNode:=n;
clickedContact:=focusedContact;
clickedGroup:=-1;
if clickedNode=NIL then exit;
if clickedNode.kind=NODE_GROUP then clickedGroup:=clickedNode.groupId;
end; // focus

function focus(c:Tcontact):boolean; overload;
begin result:=(c<>NIL) and focus(TCE(c.data^).node) end;

procedure formresized;
var
  d:Tdivisor;
begin
if not mainFrm.visible then exit;
for d:=low(d) to high(d) do
  redraw(divs[d]);
end;

procedure clear;
begin
mainfrm.roaster.clear;
while contactsPool.count > 0 do
  Tnode(contactsPool.last).free; 
end; // clear

function lowestVisibleNodeFrom(n:PVirtualNode):PvirtualNode;
begin
while mainfrm.roaster.expanded[n] and (n.childcount>0) do
  n:=mainfrm.roaster.getLastVisibleChild(n);
result:=n;
end;

function nodeBottomSide(n:PvirtualNode):integer;
begin
result:=mainfrm.roaster.GetDisplayRect(n, -1, FALSE).bottom
  -mainfrm.roaster.GetDisplayRect(mainfrm.roaster.rootnode.FirstChild, -1, FALSE).top;
end;

function onlineMaxY:integer;
begin
if divs[d_online]=NIL then result:=-1
else result:=nodeBottomSide( lowestVisibleNodeFrom(divs[d_online].treenode) );
end; // onlineMaxY

function fullMaxY:integer;
begin
if mainfrm.roaster.RootNodeCount = 0 then result:=-1
else result:=nodeBottomSide( lowestVisibleNodeFrom(mainfrm.roaster.rootnode) )
end;

function addGroup(name:string):integer;
var
  d:Tdivisor;
  first:Tnode;
begin
// create the new group
result:=groups.add;
groups.a[groups.idxOf(result)].name:=name;
// add it to divisors, and focus on the first one
first:=NIL;
for d:=high(d) downto low(d) do
  if (d in divsWithGroups) and assigned(divs[d]) then
    first:=insertNode(result, d);
focus(first);
saveGroupsDelayed:=TRUE;
end; // addGroup

function removeGroup(id:integer):boolean;
var
  cl:TcontactList;
  c:Tcontact;
  d:Tdivisor;
begin
result:=groups.exists(id);
if not result then exit;
cl:=icq.readroaster.clone;
cl.resetEnumeration;
while cl.hasMore do
  begin
  c:=cl.getNext;
  if c.group = id then
    remove(c);
  end;
cl.free;
with groups.a[groups.idxOf(id)] do
  for d:=low(d) to high(d) do
    removeNode(node[d]);
groups.delete(id);
saveGroupsDelayed:=TRUE;
end; // removeGroup

procedure edit(n:Tnode);
begin
inplace.edit.free;
if (n=NIL) or (n.treenode=NIL) then exit;
if not (n.kind in [NODE_GROUP, NODE_CONTACT]) then exit;
if not formVisible(mainfrm) then mainfrm.toggleVisible();
focus(n);
inplace.edit:=Tedit.create(mainFrm.roaster);
inplace.edit.hide;
inplace.edit.parentFont:=FALSE;
inplace.edit.parent:=mainFrm.roaster;
inplace.node:=n;
with n.rect do
  inplace.edit.SetBounds(n.textOfs,top,right-n.textofs-1,bottom-top);
case n.kind of
  NODE_CONTACT:
    begin
    inplace.what:=NODE_CONTACT;
    inplace.contact:=n.contact;
    inplace.edit.text:=inplace.contact.displayed;
    end;
  NODE_GROUP:
    begin
    with inplace.edit.font do style:=style+[fsBold];
    inplace.what:=NODE_GROUP;
    inplace.groupId:=n.groupId;
    inplace.edit.text:=groups.id2name(n.groupId);
    end;
  end;
inplace.edit.show;
inplace.edit.SetFocus;
inplace.edit.onExit:=mainfrm.roasterStopEditing;
inplace.edit.onKeyPress:=mainfrm.roasterKeyEditing;
end; // edit

function focused:Tnode;
begin
with mainFrm.roaster do
  if focusedNode=NIL then
    result:=NIL
  else
    result:=getNode(focusednode)
end; // focused

function focusedContact:Tcontact;
var
  n:Tnode;
begin
result:=NIL;
n:=focused;
if (n<>NIL) and (n.kind = NODE_CONTACT) then
  result:=n.contact
end; // focusedContact

procedure expand(n:Tnode);
begin if n<>NIL then n.setExpanded(TRUE) end;

procedure collapse(n:Tnode);
begin if n<>NIL then n.setExpanded(FALSE) end;

function nodeAt(x,y:integer):Tnode;
var
  n:Pvirtualnode;
begin
with mainFrm.roaster do
  begin
  n:=getNodeAt(x,y);
  if n=NIL then
    result:=NIL
  else
    result:=getNode(n)
  end;
end; // nodeAt

procedure focusPrevious;
begin with mainfrm.roaster do focusedNode:=GetPreviousVisible(focusedNode) end;

procedure setOnlyOnline(v:boolean);
begin
showOnlyOnline:=v;
updateHiddenNodes;
end; // setOnlyOnline

procedure sort(c:Tcontact);
begin if c<>NIL then sort(TCE(c.data^).node) end;

procedure sort(n:Tnode);
begin if n<>NIL then sort(n.treenode.parent) end;

procedure sort(tn:Pvirtualnode);
begin if tn<>NIL then mainfrm.roaster.sort(tn, 0, sdAscending, FALSE) end;

procedure setNewGroupFor(c:Tcontact; grp:integer);
begin
c.group:=grp;
update(c);
dbUpdateDelayed:=TRUE;
end; // setNewGroupFor

function isUnderDiv(n:Tnode):Tdivisor;
begin
result:=d_contacts;
while n<>NIL do
  if n.kind = NODE_DIV then
    begin
    result:=n.divisor;
    exit;
    end
  else
    n:=n.parent;
enD; // isUnderDiv

function focusTemp(n:Tnode):boolean;
var
  p:Tnode;
  bak:boolean;
begin
bak:=setRoasterAnimation(FALSE);
if (expandedByTempFocus<>NIL) and (n<>expandedByTempFocus) then
  begin
  p:=expandedByTempFocus.parent;
  if p.kind = NODE_GROUP then
    p.setExpanded(FALSE);
  end;
if not n.isVisible then
  expandedByTempFocus:=n;
result:=focus(n);
setRoasterAnimation(bak);
end; // focusTemp

procedure popup(x:integer=-1; y:integer=-1);
begin popupOn(clickedNode) end;

procedure popupOn(n:Tnode; x:integer=-1; y:integer=-1);
begin
if x<0 then
  begin
  x:=mousePos.x;
  y:=mousePos.y;
  end;
if n=NIL then
  mainfrm.divisorMenu.popup(x,y)
else
  case n.kind of
    NODE_GROUP: mainfrm.groupMenu.popup(x,y);
    NODE_CONTACT: mainfrm.contactMenu.popup(x,y);
    else mainfrm.divisorMenu.popup(x,y);
    end;
end; // popupOn

function contactsUnder(n:Tnode):integer;
begin
result:=0;
n:=n.firstChild;
while n<>NIL do
  begin
  if n.childrenCount > 0 then inc(result, contactsUnder(n));
  if n.kind = NODE_CONTACT then inc(result);
  n:=n.next;
  enD;
end; // contactsUnder

INITIALIZATION

contactsPool:=Tlist.create;
sortBy:=SB_event;

FINALIZATION

contactsPool.free;

end.
