unit themedit_fr;

interface

uses 
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, JVSpin, ColorPickerButton, ExtCtrls, MPlayer, Mask, ToolEdit,
  Buttons, JvColorCombo, JvEdit, JvCombobox;

type
  TthemeditFr = class(TFrame)
    colorBtn: TColorPickerButton;
    fontBox: TJvFontComboBox;
    sizeSpin: TJvSpinEdit;
    sizeLbl: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    img: TImage;
    stringBox: TMemo;
    fileBox: TFilenameEdit;
    mplayer: TMediaPlayer;
    addBtn: TBitBtn;
    propsBox: TComboBox;
    textBox: TMemo;
    fontLbl: TLabel;
    procedure propsBoxChange(Sender: TObject);
    procedure addBtnClick(Sender: TObject);
    procedure fileBoxChange(Sender: TObject);
    procedure setupGUI;
    procedure updateGUI;
    procedure fontBoxChange(Sender: TObject);
  end;

implementation

uses
  globalLib, utilLib, Types;

{$R *.dfm}

procedure TthemeditFr.setupGUI;
const
  gap=15;
var
  p:^Tthemeproperty;
  y:integer;
begin
if propsBox.itemIndex<0 then exit;
p:=@themeprops[propsBox.itemIndex];
case p.kind of
  TP_color: colorBtn.selectionColor:=Tcolor(p.ptr^);
  TP_font:
    with Tfont(p.ptr^) do
      begin
      colorBtn.selectionColor:=color;
      fontBox.ItemIndex:=findInStrings(name, fontBox.items);
      sizeSpin.Value:=size;
      end;
  TP_pic: assignImgBmp(img,Tbitmap(p.ptr^));
  TP_ico: assignImgIco(img,Ticon(p.ptr^));
  TP_string: stringBox.text:=string(p.ptr^);
  TP_sound:
    begin
    mplayer.FileName:=string(p.ptr^);
    fileBox.Text:=string(p.ptr^);
    try mplayer.open except end;
    end;
  end;
y:=propsBox.BoundsRect.Bottom+gap;
colorBtn.visible:= p.kind in [TP_color,TP_font];
if colorBtn.visible then
  begin
  colorBtn.top:=y;
  inc(y, colorBtn.height+gap);
  end;
fontBox.visible:= p.kind=TP_font;
sizeSpin.visible:= p.kind=TP_font;
sizeLbl.visible:= p.kind=TP_font;
fontLbl.visible:= p.kind=TP_font;
if fontBox.visible then
  begin
  fontBox.top:=y;
  sizeSpin.top:=y;
  sizeLbl.top:=y;
  inc(y, fontBox.height+gap);
  fontLbl.Top:=y;
  inc(y, 150+gap);
  end;
img.visible:= p.kind in [TP_pic,TP_ico];
if img.visible then
  begin
  img.top:=y;
  inc(y, img.height+gap);
  end;
stringBox.visible:= p.kind=TP_string;
if stringBox.visible then
  begin
  stringBox.top:=y;
  inc(y, stringBox.height+gap);
  end;
fileBox.visible:= p.kind=TP_sound;
if filebox.visible then
  begin
  fileBox.top:=y;
  inc(y, filebox.height+gap);
  end;
mplayer.visible:= p.kind=TP_sound;
if mplayer.visible then
  begin
  mplayer.top:=y;
  //inc(y, mplayer.height+gap);
  end;
updateGUI;
end; // setupGUI

procedure TthemeditFr.updateGUI;
begin
fontLbl.Font.name:=fontBox.fontname;
fontLbl.font.size:=round(sizeSpin.value);
fontLbl.font.color:=colorBtn.SelectionColor;
end; // updateGUI

procedure TthemeditFr.propsBoxChange(Sender: TObject);
begin setupGUI end;

procedure TthemeditFr.addBtnClick(Sender: TObject);
var
  p:^Tthemeproperty;
  s,line:string;
  lastsection:string;
  color:Tcolor;
  bak:Tpoint;
begin
bak:=textBox.caretpos;
// find the last section
s:=textBox.text;
while s>'' do
  begin
  line:=trim(chopline(s));
  if (line>'') and (line[1]='[') and (line[length(line)]=']') then
    lastsection:=line;
  end;
lastsection:=copy(lastsection,2,length(lastsection)-2);

s:=textBox.text+CRLF;
p:=@themeprops[propsBox.itemIndex];
if lastsection<>p.section then
  s:=s+'['+p.section+']'+CRLF;
color:=colorBtn.selectionColor;
case p.kind of
  TP_color: s:=s+p.name+'='+color2str(color);
  TP_font: s:=s+
    p.name+'.color='+color2str(color)+CRLF+
    p.name+'.size='+intToStr(trunc(sizeSpin.value))+CRLF+
    p.name+'.name='+fontbox.text;
  TP_sound: s:=s+p.name+'='+fileBox.text;
  TP_string: s:=s+p.name+'='+newline2slashn(stringBox.text);
  end;
textBox.text:=s;
textBox.caretpos:=bak;
textBox.Modified:=TRUE;
end;

procedure TthemeditFr.fileBoxChange(Sender: TObject);
begin
mplayer.FileName:=filebox.text;
try mplayer.open except end;
end;

procedure TthemeditFr.fontBoxChange(Sender: TObject);
begin updateGUI end;

end.
