{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit themesLib;

interface

procedure loadTheme(fn:string);
procedure loadThemeScript(fn:string);
procedure unloadTheme;
procedure resetTheme;
procedure applyTheme;

procedure refreshThemelist;

type
  Pthemeinfo=^Tthemeinfo;
  Tthemeinfo=record
    fn,title,desc,logo:string;
    end;
var
  themelist:array of Tthemeinfo;

implementation

uses
  utilLib, sysutils, mainDlg, graphics, globalLib, contacts, types, forms,
  virtualtrees, events, controls, chatDlg, buttons, strUtils, ComCtrls, prefDlg;
var
  cacheLoaded, cacheValid:boolean;
  cacheTime:integer;

procedure refreshThemelist;
var
  sr:TSearchRec;
  fn,s,line,k,v,section:string;
  n:integer;
begin
n:=0;
setlength(themelist,0);
if findFirst(myPath+themesPath+'*.theme.ini', faAnyFile, sr) = 0 then
  repeat
  if sr.name[1]<>'.' then
    begin
    fn:=sr.name;
    s:=loadFile(myPath+themesPath+fn);
    if trim(chopline(s))='&RQ theme file version 1' then
      begin
      setlength(themelist, n+1);
      themelist[n].fn:=fn;
      section:='';
      while s>'' do
        begin
        line:=chopline(s);
        if (line>'') and (line[1]='[') then
          begin
          line:=trim(line);
          if line[length(line)]=']' then
            section:=copy(line,2,length(line)-2);
          continue;
          end;
        v:=line;
        k:=trim(chop('=',v));
        v:=trim(v);
        if section='' then
          begin
          if k='logo' then themelist[n].logo:=v;
          if k='title' then themelist[n].title:=v;
          if k='desc' then themelist[n].desc:=ansiReplaceStr(v,'\n',CRLF);
          end;
        if section='desc' then
          with themelist[n] do desc:=desc+line+CRLF;
        end;
      with themelist[n] do desc:=trimright(desc);
      inc(n);
      end;
    end;
  until findNext(sr) <> 0;
findClose(sr);
end; // refreshThemelist

procedure loadThemeScript(fn:string);
type
  Tsection=(_null,_roaster,_tip,_pics,_icons,_history,_smiles,_sounds,_menu);
const
  sectionLabels:array [Tsection] of string=('','roaster','tip','pics','icons',
    'history','smiles','sounds','menu');
var
  path,k,v:string;
  loadedpic:Tbitmap;

  procedure crop(bmp:Tbitmap; x,y,dx,dy:integer);
  begin
  if dy < 0 then exit;
  if (x<>0) or (y<>0) then
    bmp.canvas.copyRect(rect(0,0,dx,dy), bmp.canvas, rect(x,y,x+dx,y+dy));
  bmp.Width:=dx;
  bmp.height:=dy;
  end; // crop

  function fullpath(fn:string):string;
  begin if ansipos(':',fn)=0 then result:=path+fn else result:=fn end;

  procedure parsePic(var bmp:Tbitmap);
  var
    s,fn:string;
    x,y,dx,dy,i:integer;
  begin
  s:=v;
  fn:=chop(';',s);
  if fn='' then exit;
  if bmp=NIL then bmp:=Tbitmap.create;
  x:=str2valor(chop(';',s));
  y:=str2valor(chop(';',s));
  dx:=str2valor(chop(';',s));
  dy:=str2valor(chop(';',s));
  if fn = '@' then
    bmp.assign(loadedpic)
  else
    if AnsiStartsText('@pics.', fn) then
      begin
      i:=findInStrings(copy(fn,6,length(fn)),pic2str);
      if i>0 then bmp.assign(theme.pics[i]);
      end
    else
      begin
      if not loadPic(fullpath(fn), loadedpic) then loadedpic.width:=0;
      bmp.assign(loadedpic);
      end;
  crop(bmp,x,y,dx,dy);
  end; // parsePic

  function fontAvailable(list:string):string;
  begin
  result:=chop(';',list);
  while (list>'') and (screen.fonts.IndexOf(result) < 0) do
    result:=chop(';',list);
  end; // fontAvailable

  procedure parseFont(prefix:string; font:Tfont);
  var
    i:integer;
  begin
  if k=prefix+'.name' then font.name:=fontAvailable(v);
  if k=prefix+'.size' then font.size:=strToInt(v);
  if k=prefix+'.color' then font.color:=str2color(v);
  if k=prefix+'.charset' then
    if isOnlyDigits(v) then
      font.charset:=strToInt(v)
    else
      if IdentToCharset(v,i) then
        font.charset:=i
      else
        if IdentToCharset(v+'_CHARSET',i) then
          font.charset:=i;
  if k=prefix+'.style' then
    font.style:=str2fontstyle(v);
  end; // parseFont

var
  section:Tsection;
  txt,line:string;
  i:integer;
  pic,smilePic:Tbitmap;

  procedure parseIco(var ico:Ticon);
  begin
  if ico <> NIL then ico.free;
	parsePic(pic);
  ico:=pic2ico(pic);
  end; // parseIco

begin
path:=ExtractFilePath(fn);
section:=_null;
txt:=loadfile(fn);
if fileage(fn) > cacheTime then cacheValid:=FALSE;
loadedPic:=Tbitmap.create;
pic:=Tbitmap.create;
smilePic:=NIL;
while txt>'' do
  try
  line:=chopline(txt);
  line:=trim(chop('#',line));
  if line='' then continue;
  if (line[1]='[') and (line[length(line)]=']') then
    begin
    i:=findInStrings(lowercase(copy(line,2,length(line)-2)),sectionLabels);
    if i<0 then section:=_null
    else section:=Tsection(i);
    continue;
    end;
  v:=line;
  k:=trim(chop('=',v));
  v:=trim(v);
  if k='include' then loadThemeScript(fullpath(v));

	if sectionLabels[section] = 'smiles' then
  	begin
    if isSupportedPicFile(line) then
      begin
      loadpic(fullpath(line), loadedpic);
      smilePic:=dupePic(loadedPic);
      end
    else
      theme.smiles.add(line, smilePic);
    continue;
  	end;

  for i:=0 to length(themeprops)-1 do
    if themeprops[i].section = sectionLabels[section] then
      // multi-properties objects (like fonts) can't be compared to k
      if themeprops[i].kind = TP_FONT then
        parseFont(themeprops[i].name, Tfont(themeprops[i].ptr^))
      else
        if themeprops[i].name = k then
          case themeprops[i].kind of
            TP_STRING: string(themeprops[i].ptr^):=ansiReplaceStr(v,'\n',CRLF);
            TP_COLOR: Tcolor(themeprops[i].ptr^):=str2color(v);
            TP_SOUND: string(themeprops[i].ptr^):=fullpath(v);
            TP_PIC: parsePic(Tbitmap(themeprops[i].ptr^));
            TP_ICO: parseIco(Ticon(themeprops[i].ptr^));
            end;
  except
  end;
loadedpic.free;
pic.free;
end; // loadThemeScript

procedure applyTheme;
begin
if (chatFrm<>NIL) then
  begin
  setupChatButtons;
  chatFrm.updateGraphics;
  end;
with mainfrm.roaster.treeoptions do
  if theme.roaster.bgpic.width > 0 then
    paintoptions:=paintoptions+[toShowBackground]
  else
    paintoptions:=paintoptions-[toShowBackground];
mainfrm.roaster.color:=theme.roaster.bgcolor;
mainfrm.bar.color:=theme.roaster.barcolor;
mainfrm.sbar.color:=theme.roaster.barcolor;
mainFrm.roaster.background.bitmap.assign(theme.roaster.bgpic);
mainfrm.roaster.colors.focusedselectioncolor:=theme.roaster.selectioncolor;
mainfrm.roaster.colors.focusedselectionbordercolor:=theme.roaster.selectioncolor;
mainfrm.roaster.colors.unfocusedselectioncolor:=theme.roaster.selectioncolor;
mainfrm.roaster.colors.unfocusedselectionbordercolor:=theme.roaster.selectioncolor;
mainfrm.bar.Height:=getPic(PIC_RATMAN).height+7;
mainfrm.menuBtn.height:=getPic(PIC_RATMAN).height+5;
mainfrm.menuBtn.width:=getPic(PIC_RATMAN).width+5;
mainfrm.statusBtn.height:=getPic(PIC_STATUS_ONLINE).height+5;
mainfrm.statusBtn.width:=getPic(PIC_STATUS_ONLINE).width+5;
mainfrm.visibilityBtn.height:=getPic(PIC_VISIBILITY_ALL).height+5;
mainfrm.visibilityBtn.width:=getPic(PIC_VISIBILITY_ALL).width+5;
mainfrm.menuBtn.glyph.assign(getPic(PIC_RATMAN));
mainfrm.FormResize(NIL);
mainFrm.menu.leftmargin:=getPic(PIC_MENUBAR).width;
mainfrm.updateStatusGlyphs;
mainfrm.menu.LeftMargin:=getPic(PIC_MENUBAR).width+3;
mainfrm.roaster.DefaultNodeHeight:=getPic(PIC_STATUS_ONLINE).height+2;
if assigned(prefFrm) then prefFrm.setPics();
repaintAllWindows;
end; // applyTheme

procedure loadTheme(fn:string);
begin
unloadTheme;
if not FileExists(fn) then fn:=myPath+themesPath+'default.theme.ini';
theme.fn:=ExtractFileName(fn);
cacheLoaded:=loadCachedTheme(fn+'.cache');
if cacheLoaded then cacheTime:=FileAge(fn+'.cache');
cacheValid:=cacheLoaded and (cacheTime>fileAge(fn));
loadThemeScript(fn);
if not cacheValid then saveCachedTheme(fn+'.cache');
end; // loadTheme

procedure unloadTheme;
var
  i:integer;
begin
theme.roaster.bgpic.width:=0;
theme.roaster.barcolor:=clBtnFace;
for i:=1 to PIC_LAST do
  freePic(i);
for i:=1 to ICO_LAST do
  freeAndNIL(theme.icons[i]);
freeAndNIL(theme.icons[ICO_EMPTY]);
theme.icons[ICO_EMPTY]:=Ticon.create;
theme.smiles.clear;
end; // unloadTheme

procedure resetTheme;
begin
theme.roaster.font.assign(mainfrm.font);
theme.history.myfont.assign(mainfrm.font);
theme.history.hisfont.assign(mainfrm.font);
end; // resetTheme

end.

