{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit traylib;

interface

uses
  forms, ShellAPI, Messages, windows, controls, contacts, events, extCtrls, graphics, icqv7;

const
  WM_TRAY = WM_USER+1;
type
  TtrayIcon=class
    private
      data:TNotifyIconData;
      shown:boolean;
    public
      constructor create(handle:HWND);
      procedure minimize;
      procedure update;
      procedure hide;
      procedure show;
      procedure setIcon(icon:Ticon);
      procedure setTip(s:string);
      procedure setIconFile(fn:string);
      procedure updateHandle(handle:HWND);
    end; // TtrayIcon

type
  TstatusIcon=class
   public
    trayIcon:TtrayIcon;
    lastIco:Ticon;
    lastTip:string;
    constructor create;
    destructor destroy; override;
    procedure update;
    procedure empty;
    procedure handleChanged;
    end; // TstatusIcon

implementation

uses
  globalLib, mainDlg, sysutils, utilLib;

constructor TstatusIcon.create;
begin
trayIcon:=TtrayIcon.create(0);
trayIcon.setTip('&&&RQ');
lastIco:=NIL;
lastTip:='';
end; // create

destructor TstatusIcon.destroy;
begin
trayIcon.hide;
trayIcon.destroy;
end; // destroy

procedure TstatusIcon.update;
var
  e:Thevent;
  newIco:Ticon;
  s:string;
begin
if (eventQ=NIL) or eventQ.empty then
  begin
  if ICQ.isOnline then
    begin
    s:=statusName(status);
    newIco:=getIco(status, invisibleMode);
    end
  else
    if ICQ.isConnecting or resolving then
      begin
      s:=___('connecting');
      if invisibleMode then
        newIco:=getIco(ICO_INVISIBLE_CONNECTING)
      else
        newIco:=getIco(ICO_CONNECTING)
      end
    else
      begin
      s:=statusName(SC_OFFLINE);
      newIco:=getIco(SC_OFFLINE, invisibleMode);
      end;
  if assigned(ICQ.myinfo) then
    s:=___('tray status',[s,ICQ.myinfo.uin])
  else
    s:='&&RQ';
  end
else
  begin
  e:=eventQ.top;
  if blinking or onStatusDisable[icq.myinfo.status].blinking then
    newIco:=e.ico
  else
    newIco:=getIco(ICO_EMPTY);
  if e.who=NIL then
    s:=___('error')
  else
    s:=___('tray '+event2str[e.kind],[e.who.displayed]);
  end;

if newIco <> lastIco then
  begin
  lastIco:=newIco;
  trayIcon.setIcon(lastIco);
  end;
if s <> lastTip then
  begin
  lastTip:=s;
  trayIcon.setTip(s);
  end;
end; // update

procedure TstatusIcon.empty;
begin
lastIco:=getIco(ICO_EMPTY);
trayIcon.setIcon(lastIco);
end;

procedure TstatusIcon.handleChanged;
begin trayIcon.updateHandle(mainFrm.handle) end;

constructor TtrayIcon.create(handle:HWND);
begin
with data do
  begin
  uCallbackMessage := WM_TRAY;
  cbSize := sizeof(data);
  Wnd := Handle;
  uID := 100;
  uFlags := NIF_MESSAGE + NIF_ICON + NIF_TIP;
  end;
setIcon(application.icon);
setTip(application.title);
end; // create

procedure TtrayIcon.updateHandle(handle:HWND);
begin
if not shown then
  begin
  data.wnd:=handle;
  exit;
  end;
hide;
data.wnd:=handle;
Shell_NotifyIcon(NIM_ADD, @data)
end;

procedure TtrayIcon.update;
begin
if shown then
  if not Shell_NotifyIcon(NIM_MODIFY, @data) then
    Shell_NotifyIcon(NIM_ADD, @data);
end; { update }

procedure TtrayIcon.setIcon(icon:Ticon);
begin
if icon=NIL then exit;
data.hIcon:=icon.handle;
update;
end; { setIcon }

procedure TtrayIcon.setIconFile(fn:string);
var
  ico:Ticon;
begin
ico:=Ticon.create;
ico.loadFromFile(fn);
setIcon(ico);
end; // setIconFile

procedure TtrayIcon.setTip(s:string);
begin
strPCopy(data.szTip, s);
update;
end; // setTip

procedure TtrayIcon.minimize;
begin
show;
Application.ShowMainForm := False;
// Toolwindows dont have a TaskIcon. (Remove if TaskIcon is to be show when form is visible)
SetWindowLong(Application.Handle, GWL_EXSTYLE, WS_EX_TOOLWINDOW);
end; // minimizeToTray

procedure TtrayIcon.show;
begin
shown:=true;
Shell_NotifyIcon(NIM_ADD, @data)
end; { show }

procedure TtrayIcon.hide;
begin Shell_NotifyIcon(NIM_DELETE, @data) end;

end.
