{
Copyright (C) 2002-2004  Massimo Melina (www.rejetto.com)

This file is part of &RQ.

    &RQ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    &RQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with &RQ; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit uinlistLib;

interface

uses
  contacts, classes;

type
  PuinList=^TuinList;
  TuinList=record
    name:string;
    desc:string;
    cl:TcontactList;
    end;

  TuinLists=class(Tlist)
   private
    enumidx :integer;
   public
    destructor destroy; override;
    function  exists(name:string):boolean;
    function  idxOf(name:string):integer;
    function  getAt(idx:integer):PuinList;
    function  put(name:string):PuinList;
		function  remove(ul:PuinList):boolean; overload;
    procedure fromString(s:string);
    function  toString:string;
    procedure clear; override;
    function  names:string;
    function  get(name:string):PuinList;
    procedure resetEnumeration;
    function  hasMore:boolean;
    function  getNext:PuinList;
    end;

implementation

uses
  sysutils, icqv7, utilLib;

function TuinLists.exists(name:string):boolean;
begin result:=idxOf(name)>=0 end;

function TuinLists.idxOf(name:string):integer;
begin
result:=count-1;
while (result>=0) and (compareText(getAt(result).name,name)<>0) do
  dec(result);
end; // idxof

function TuinLists.getAt(idx:integer):PuinList;
begin result:=PuinList(items[idx]) end;

function TuinLists.put(name:string):PuinList;
var
  idx:integer;
begin
idx:=idxOf(name);
if idx>=0 then
  begin
  result:=getAt(idx);
  exit;
  end;
new(result);
result.name:=name;
result.desc:='';
result.cl:=TcontactList.create;
add(result);
end;

function TuinLists.remove(ul:PuinList):boolean;
var
  i:integer;
begin
result:=FALSE;
for i:=0 to count-1 do
  if items[i] = ul then
    begin
    dispose(ul);
    inherited remove(ul);
    result:=TRUE;
    exit;
    end;
end; // remove

procedure Tuinlists.clear;
var
  i:integer;
begin
for i:=0 to count-1 do
  dispose(PuinList(items[i]));
inherited;
end;

destructor Tuinlists.destroy;
begin
clear;
inherited;
end; // destroy

const
  FK_NAME=1;
  FK_DESC=2;
  FK_UIN=3;

procedure Tuinlists.fromString(s:string);
var
  l,t:integer;
begin
clear;
while s > '' do
  begin
  l:=integer((@s[1])^);
  t:=integer((@s[5])^);
  case t of
    FK_NAME: put(copy(s,9,l));
    FK_DESC: PuinList(last)^.desc:=copy(s,9,l);
    FK_UIN: PuinList(last)^.cl.add(contactsDB.get(integer((@s[9])^)));
    end;
  system.delete(s,1,8+l);
  end;
end; // fromstring

function Tuinlists.toString:string;

  procedure writeDown(code:integer; data:string);
  begin result:=result+int2str(length(data))+int2str(code)+data end;

var
  i,j:integer;
begin
result:='';
for i:=0 to count-1 do
  with getAt(i)^ do
    begin
    writedown(FK_NAME, name);
    writedown(FK_DESC, desc);
    for j:=0 to cl.count-1 do
      writedown(FK_UIN, int2str(cl.getAt(j).uin));
    end;
end; // tostring

function Tuinlists.names:string;
var
  i:integer;
begin
result:='';
for i:=0 to count-1 do
  result:=result+getAt(i)^.name+#13;
if result > '' then
  setLength(result,length(result)-1); 
end; // names

function Tuinlists.get(name:string):PuinList;
var
  i:integer;
begin
i:=idxOf(name);
if i<0 then
  result:=NIL
else
  result:=getAt(i);
end; // get

procedure Tuinlists.resetEnumeration;
begin enumIdx:=0 end;

function Tuinlists.hasMore:boolean;
begin result:=enumIdx<count end;

function Tuinlists.getNext:PuinList;
begin
result:=getAt(enumIdx);
inc(enumIdx);
end; // getNext

end.
